Package: binxr
Type: Package
Title: 'Binance' REST API Client
Version: 0.1.2
Authors@R: c(
    person("Oliver", "Zhou", email = "oliver.yxzhou@gmail.com", role = c("aut", "cre")),
    person("Lily", "Li", email = "lilyli.qc.hk@gmail.com", role = "aut")
    )
Author: Oliver Zhou [aut, cre],
  Lily Li [aut]
Maintainer: Oliver Zhou <oliver.yxzhou@gmail.com>
Description: Client for the 'Binance' <https://www.binance.com/> Spot,
    USD-M Futures, and Options REST APIs. Provides helper functions for signed
    and unsigned requests, market data retrieval, account access, and order
    management with 'data.table' output by default. COIN-M Futures, Portfolio
    Margin, WebSocket, SBE, and FIX APIs are not included.
License: MIT + file LICENSE
URL: https://github.com/OliverLDS/binxr,
        https://oliverlds.github.io/binxr/
BugReports: https://github.com/OliverLDS/binxr/issues
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: httr2, jsonlite, digest, data.table, rlang
Suggests: testthat (>= 3.0.0), waldo
Config/testthat/edition: 3
RoxygenNote: 7.2.3
NeedsCompilation: no
Packaged: 2026-08-01 15:01:06 UTC; oliver
Repository: CRAN
Date/Publication: 2026-08-02 18:40:02 UTC
Built: R 4.5.2; ; 2026-08-02 23:52:22 UTC; unix

Package: binxr
Type: Package
Title: 'Binance' REST API Client
Version: 0.1.1
Authors@R: c(
    person("Oliver", "Zhou", email = "oliver.yxzhou@gmail.com", role = c("aut", "cre")),
    person("Lily", "Li", email = "lilyli.qc.hk@gmail.com", role = "aut")
    )
Author: Oliver Zhou [aut, cre],
  Lily Li [aut]
Maintainer: Oliver Zhou <oliver.yxzhou@gmail.com>
Description: Client for the 'Binance' <https://www.binance.com/> Spot,
    Futures, and Options REST APIs. Provides helper functions for signed
    and unsigned requests, market data retrieval, account access, and order
    management with 'data.table' output by default.
License: MIT + file LICENSE
URL: https://github.com/OliverLDS/binxr,
        https://oliverlds.github.io/binxr/
BugReports: https://github.com/OliverLDS/binxr/issues
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: httr2, jsonlite, digest, data.table, rlang
Suggests: testthat (>= 3.0.0), waldo
Config/testthat/edition: 3
RoxygenNote: 7.2.3
NeedsCompilation: no
Packaged: 2026-05-13 07:49:36 UTC; oliver
Repository: CRAN
Date/Publication: 2026-05-18 18:30:20 UTC
Built: R 4.5.2; ; 2026-05-18 19:40:54 UTC; unix

# Generated by roxygen2: do not edit by hand

export(binxr_config_futures)
export(binxr_config_options)
export(binxr_config_spot)
export(cancel_fapi_trade_order)
export(cancel_fapi_trade_orders_all)
export(coerce_account_frames)
export(config_futures)
export(config_options)
export(config_spot)
export(futures_cancel_algo_order)
export(futures_cancel_all_algo_orders)
export(futures_cancel_all_orders)
export(futures_cancel_order)
export(futures_countdown_cancel_all)
export(futures_get_24hr_ticker)
export(futures_get_account)
export(futures_get_account_summary)
export(futures_get_account_trades)
export(futures_get_aggregate_trades)
export(futures_get_algo_order)
export(futures_get_algo_orders)
export(futures_get_balance)
export(futures_get_book_ticker)
export(futures_get_commission_rate)
export(futures_get_continuous_klines)
export(futures_get_exchange_info)
export(futures_get_force_orders)
export(futures_get_funding_info)
export(futures_get_funding_rate_history)
export(futures_get_historical_trades)
export(futures_get_index_constituents)
export(futures_get_index_price_klines)
export(futures_get_insurance_balance)
export(futures_get_klines)
export(futures_get_mark_price)
export(futures_get_mark_price_klines)
export(futures_get_multi_assets_mode)
export(futures_get_open_algo_orders)
export(futures_get_open_interest)
export(futures_get_open_orders)
export(futures_get_order)
export(futures_get_order_amendments)
export(futures_get_order_book)
export(futures_get_order_rate_limit)
export(futures_get_orders)
export(futures_get_position_mode)
export(futures_get_position_risk)
export(futures_get_positions)
export(futures_get_premium_index_klines)
export(futures_get_recent_trades)
export(futures_get_rpi_depth)
export(futures_get_server_time)
export(futures_get_symbol_adl_risk)
export(futures_get_ticker_price)
export(futures_get_trading_schedule)
export(futures_modify_order)
export(futures_ping)
export(futures_place_algo_order)
export(futures_place_order)
export(futures_set_leverage)
export(futures_set_margin_type)
export(futures_set_multi_assets_mode)
export(futures_set_position_mode)
export(futures_test_order)
export(get_fapi_account)
export(get_fapi_account_position_risk)
export(get_fapi_account_positions)
export(get_fapi_account_summary)
export(get_fapi_exchange_info)
export(get_fapi_klines)
export(get_fapi_mark_price)
export(get_fapi_system_time)
export(get_fapi_trade_open_orders)
export(get_fapi_trade_order)
export(get_fapi_trade_orders)
export(get_spot_mark_price)
export(options_cancel_all_orders)
export(options_cancel_all_orders_by_underlying)
export(options_cancel_order)
export(options_get_24hr_ticker)
export(options_get_account_trades)
export(options_get_commission)
export(options_get_exchange_info)
export(options_get_exercise_history)
export(options_get_exercise_records)
export(options_get_funding_flow)
export(options_get_index_price)
export(options_get_klines)
export(options_get_margin_account)
export(options_get_mark_price)
export(options_get_open_interest)
export(options_get_open_orders)
export(options_get_order)
export(options_get_order_book)
export(options_get_order_history)
export(options_get_positions)
export(options_get_recent_block_trades)
export(options_get_recent_trades)
export(options_get_server_time)
export(options_ping)
export(options_place_order)
export(place_fapi_trade_order)
export(round_price_qty)
export(set_fapi_account_leverage)
export(set_fapi_account_margin_type)
export(set_fapi_account_position_side)
export(spot_amend_order_keep_priority)
export(spot_cancel_all_orders)
export(spot_cancel_order)
export(spot_cancel_order_list)
export(spot_cancel_replace_order)
export(spot_get_24hr_ticker)
export(spot_get_account)
export(spot_get_account_trades)
export(spot_get_aggregate_trades)
export(spot_get_all_order_lists)
export(spot_get_allocations)
export(spot_get_average_price)
export(spot_get_book_ticker)
export(spot_get_commission_rates)
export(spot_get_exchange_info)
export(spot_get_execution_rules)
export(spot_get_historical_block_trades)
export(spot_get_historical_trades)
export(spot_get_klines)
export(spot_get_open_order_lists)
export(spot_get_open_orders)
export(spot_get_order)
export(spot_get_order_amendments)
export(spot_get_order_book)
export(spot_get_order_list)
export(spot_get_orders)
export(spot_get_prevented_matches)
export(spot_get_recent_trades)
export(spot_get_reference_price)
export(spot_get_reference_price_calculation)
export(spot_get_relevant_filters)
export(spot_get_rolling_window_ticker)
export(spot_get_server_time)
export(spot_get_ticker_price)
export(spot_get_trading_day_ticker)
export(spot_get_ui_klines)
export(spot_get_unfilled_order_count)
export(spot_ping)
export(spot_place_oco_order)
export(spot_place_order)
export(spot_place_order_list_oco)
export(spot_place_order_list_opo)
export(spot_place_order_list_opoco)
export(spot_place_order_list_oto)
export(spot_place_order_list_otoco)
export(spot_place_sor_order)
export(spot_test_order)
export(spot_test_sor_order)
export(util_round_price_qty)
import(data.table)
importFrom(rlang,"%||%")

# binxr

[![Documentation](https://img.shields.io/badge/docs-pkgdown-blue.svg)](https://oliverlds.github.io/binxr/)

An R client for the **Binance REST API** (Spot, Futures, and Options).

Documentation site: <https://oliverlds.github.io/binxr/>

## Overview
`binxr` provides lightweight helpers for:
- Public market data (ticker prices, klines, exchange info)
- Authenticated endpoints (account info, positions, and order management)
- Spot, USD-M futures, and options product areas
- Data returned as `data.table` for easy downstream analysis

## Installation
```r
# From GitHub (using remotes or devtools)
remotes::install_github("OliverLDS/binxr")
```

## Quick Start
```r
library(binxr)

# Create a public config without credentials.
cfg <- config_futures(api_key = NULL, secret_key = NULL)

# Prepare values using local helper logic, without making a network request.
exchange_info <- list(
  symbols = list(
    ETHUSDT = list(
      filters = list(
        list(filterType = "PRICE_FILTER", tickSize = "0.01"),
        list(filterType = "LOT_SIZE", stepSize = "0.001")
      )
    )
  )
)

round_price_qty(exchange_info, "ETHUSDT", price = 1800.123, quantity = 0.12345)
```

Live API requests require network access. Authenticated trading endpoints also
require Binance API credentials supplied explicitly or via `BINX_API_KEY` and
`BINX_SECRET_KEY`.

```r
library(binxr)

cfg <- config_futures()

# Public futures market data.
futures_get_server_time(config = cfg)
futures_get_mark_price(symbol = "ETHUSDT", config = cfg)

# Public spot and options market data.
spot_get_ticker_price(symbol = "BTCUSDT", config = config_spot())
options_get_mark_price(config = config_options())
```

Order-management helpers such as `futures_place_order()` and
`spot_place_order()` are intentionally not shown as copy-paste quick-start code:
review Binance permissions, testnet settings, and order parameters before using
them with real credentials.

## Requirements
- R >= 4.1.0
- Packages: `httr2`, `jsonlite`, `digest`, `data.table`, `rlang`

## License
MIT License (see [LICENSE](LICENSE) file).

---
Developed by **Oliver Zhou**, 2025.

#' Get Binance Futures server time
#'
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `POSIXct` timestamp.
#' @export
futures_get_server_time <- function(config = config_futures()) {
  .ms_to_datetime(.binance_server_time(config))
}

#' @rdname futures_get_server_time
#' @export
get_fapi_system_time <- function(config = binxr_config_futures()) {
  .binxr_deprecate_alias("get_fapi_system_time", "futures_get_server_time")
  futures_get_server_time(config = config)
}

#' Get Binance Futures exchange info
#'
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list. Symbol and asset lists are named when present.
#' @export
futures_get_exchange_info <- function(config = config_futures()) {
  payload <- .request_public(config, "/fapi/v1/exchangeInfo")
  if (!is.null(payload$symbols)) {
    names(payload$symbols) <- vapply(payload$symbols, `[[`, "", "symbol")
  }
  if (!is.null(payload$assets)) {
    names(payload$assets) <- vapply(payload$assets, `[[`, "", "asset")
  }
  payload
}

#' @rdname futures_get_exchange_info
#' @export
get_fapi_exchange_info <- function(config = binxr_config_futures()) {
  .binxr_deprecate_alias("get_fapi_exchange_info", "futures_get_exchange_info")
  futures_get_exchange_info(config = config)
}

#' Get Binance Futures mark price
#'
#' @param symbol Optional trading pair symbol such as `"ETHUSDT"`. If `NULL`,
#'   returns all symbols.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_get_mark_price <- function(symbol = NULL, config = config_futures()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  .request_public(config, "/fapi/v1/premiumIndex", query = list(symbol = symbol))
}

#' @rdname futures_get_mark_price
#' @export
get_fapi_mark_price <- function(symbol = NULL, config = binxr_config_futures()) {
  .binxr_deprecate_alias("get_fapi_mark_price", "futures_get_mark_price")
  futures_get_mark_price(symbol = symbol, config = config)
}

#' Get Binance Futures klines
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param interval Kline interval string.
#' @param startTime Optional start time in milliseconds since Unix epoch.
#' @param endTime Optional end time in milliseconds since Unix epoch.
#' @param limit Maximum number of rows to return.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_klines <- function(
    symbol,
    interval,
    startTime = NULL,
    endTime = NULL,
    limit = 500,
    json_list = FALSE,
    config = config_futures()) {
  .validate_symbol(symbol)
  .validate_scalar_character(interval, "interval")
  .validate_positive_integerish(limit, "limit")
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")

  payload <- .request_public(
    config,
    "/fapi/v1/klines",
    query = list(
      symbol = symbol,
      interval = interval,
      limit = limit,
      startTime = startTime,
      endTime = endTime
    )
  )

  if (isTRUE(json_list)) {
    return(payload)
  }

  kline_dt <- data.table::rbindlist(payload)
  names(kline_dt) <- c(
    "datetime", "open", "high", "low", "close", "volume",
    "datetime_close", "quote_asset_volume", "num_trades",
    "taker_buy_base_vol", "taker_buy_quote_vol", "ignore"
  )
  kline_dt <- .normalize_time_cols(kline_dt, c("datetime", "datetime_close"))
  .coerce_numeric_cols(
    kline_dt,
    c("open", "high", "low", "close", "volume", "quote_asset_volume",
      "num_trades", "taker_buy_base_vol", "taker_buy_quote_vol", "ignore")
  )
}

#' @rdname futures_get_klines
#' @export
get_fapi_klines <- function(
    symbol,
    interval,
    startTime = NULL,
    endTime = NULL,
    limit = 500,
    config = binxr_config_futures()) {
  .binxr_deprecate_alias("get_fapi_klines", "futures_get_klines")
  futures_get_klines(
    symbol = symbol,
    interval = interval,
    startTime = startTime,
    endTime = endTime,
    limit = limit,
    json_list = FALSE,
    config = config
  )
}

#' Test Binance Futures connectivity
#'
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_ping <- function(config = config_futures()) {
  .request_public(config, "/fapi/v1/ping")
}

#' Get Binance Futures order book
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param limit Optional depth limit. One of `5`, `10`, `20`, `50`, `100`,
#'   `500`, or `1000`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_get_order_book <- function(symbol, limit = NULL, config = config_futures()) {
  .validate_symbol(symbol)
  if (!is.null(limit)) {
    .validate_positive_integerish(limit, "limit")
    if (!(limit %in% c(5L, 10L, 20L, 50L, 100L, 500L, 1000L))) {
      stop("`limit` must be one of: 5, 10, 20, 50, 100, 500, 1000.", call. = FALSE)
    }
  }

  .request_public(config, "/fapi/v1/depth", query = list(symbol = symbol, limit = limit))
}

#' Get Binance Futures recent trades
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param limit Maximum number of trades to return. Must not exceed `1000`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_recent_trades <- function(symbol, limit = 500, json_list = FALSE, config = config_futures()) {
  .validate_symbol(symbol)
  .futures_validate_limit(limit)
  .validate_json_list_flag(json_list)

  payload <- .request_public(config, "/fapi/v1/trades", query = list(symbol = symbol, limit = limit))
  if (isTRUE(json_list)) {
    return(payload)
  }

  trades_dt <- .maybe_as_dt(payload)
  trades_dt <- .coerce_numeric_cols(trades_dt, c("id", "price", "qty", "quoteQty"))
  .normalize_time_cols(trades_dt, "time")
}

#' Get Binance Futures aggregate trades
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param fromId Optional aggregate trade ID.
#' @param startTime Optional start time in milliseconds since Unix epoch.
#' @param endTime Optional end time in milliseconds since Unix epoch.
#' @param limit Maximum number of trades to return. Must not exceed `1000`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_aggregate_trades <- function(
    symbol,
    fromId = NULL,
    startTime = NULL,
    endTime = NULL,
    limit = 500,
    json_list = FALSE,
    config = config_futures()) {
  .validate_symbol(symbol)
  .validate_optional_scalar_numeric(fromId, "fromId")
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .futures_validate_limit(limit)
  .validate_json_list_flag(json_list)

  if (!is.null(fromId) && (!is.null(startTime) || !is.null(endTime))) {
    stop("`fromId` must not be supplied with `startTime` or `endTime`.", call. = FALSE)
  }
  .futures_validate_time_window(startTime, endTime, 60L * 60L * 1000L, "1 hour")

  payload <- .request_public(
    config,
    "/fapi/v1/aggTrades",
    query = list(symbol = symbol, fromId = fromId, startTime = startTime, endTime = endTime, limit = limit)
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  trades_dt <- .maybe_as_dt(payload)
  trades_dt <- .coerce_numeric_cols(trades_dt, c("a", "p", "q", "f", "l"))
  .normalize_time_cols(trades_dt, "T")
}

#' Get Binance Futures 24hr ticker statistics
#'
#' @param symbol Optional trading pair symbol.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_get_24hr_ticker <- function(symbol = NULL, config = config_futures()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  .request_public(config, "/fapi/v1/ticker/24hr", query = list(symbol = symbol))
}

#' Get Binance Futures ticker price
#'
#' @param symbol Optional trading pair symbol.
#' @param version One of `"v2"` or `"deprecated"`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_get_ticker_price <- function(
    symbol = NULL,
    version = c("v2", "deprecated"),
    config = config_futures()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  version <- match.arg(version)
  path <- if (identical(version, "v2")) "/fapi/v2/ticker/price" else "/fapi/v1/ticker/price"
  .request_public(config, path, query = list(symbol = symbol))
}

#' Get Binance Futures order book ticker
#'
#' @param symbol Optional trading pair symbol.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_get_book_ticker <- function(symbol = NULL, config = config_futures()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  .request_public(config, "/fapi/v1/ticker/bookTicker", query = list(symbol = symbol))
}

#' Get Binance Futures open interest
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_get_open_interest <- function(symbol, config = config_futures()) {
  .validate_symbol(symbol)
  .request_public(config, "/fapi/v1/openInterest", query = list(symbol = symbol))
}

#' Get Binance Futures funding rate history
#'
#' @param symbol Optional trading pair symbol.
#' @param startTime Optional start time in milliseconds since Unix epoch.
#' @param endTime Optional end time in milliseconds since Unix epoch.
#' @param limit Maximum number of rows to return. Must not exceed `1000`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_funding_rate_history <- function(
    symbol = NULL,
    startTime = NULL,
    endTime = NULL,
    limit = 100,
    json_list = FALSE,
    config = config_futures()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .futures_validate_limit(limit)
  .validate_json_list_flag(json_list)

  payload <- .request_public(
    config,
    "/fapi/v1/fundingRate",
    query = list(symbol = symbol, startTime = startTime, endTime = endTime, limit = limit)
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  rate_dt <- .maybe_as_dt(payload)
  rate_dt <- .coerce_numeric_cols(rate_dt, c("fundingRate", "markPrice"))
  .normalize_time_cols(rate_dt, "fundingTime")
}

#' Get Binance Futures funding info
#'
#' @param symbol Optional trading pair symbol.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_funding_info <- function(symbol = NULL, json_list = FALSE, config = config_futures()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  .validate_json_list_flag(json_list)

  payload <- .request_public(config, "/fapi/v1/fundingInfo", query = list(symbol = symbol))
  if (isTRUE(json_list)) {
    return(payload)
  }

  .coerce_numeric_cols(.maybe_as_dt(payload), c("adjustedFundingRateCap", "adjustedFundingRateFloor", "fundingIntervalHours"))
}

#' Get Binance Futures mark price klines
#'
#' @inheritParams futures_get_klines
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_mark_price_klines <- function(
    symbol, interval, startTime = NULL, endTime = NULL, limit = 500, json_list = FALSE, config = config_futures()) {
  .futures_get_derived_klines("/fapi/v1/markPriceKlines", symbol, interval, startTime, endTime, limit, json_list, config)
}

#' Get Binance Futures index price klines
#'
#' @inheritParams futures_get_klines
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_index_price_klines <- function(
    symbol, interval, startTime = NULL, endTime = NULL, limit = 500, json_list = FALSE, config = config_futures()) {
  .futures_get_derived_klines("/fapi/v1/indexPriceKlines", symbol, interval, startTime, endTime, limit, json_list, config)
}

#' Get Binance Futures premium index klines
#'
#' @inheritParams futures_get_klines
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_premium_index_klines <- function(
    symbol, interval, startTime = NULL, endTime = NULL, limit = 500, json_list = FALSE, config = config_futures()) {
  .futures_get_derived_klines("/fapi/v1/premiumIndexKlines", symbol, interval, startTime, endTime, limit, json_list, config)
}

#' Get Binance Futures continuous contract klines
#'
#' @param pair Futures pair, for example `"BTCUSDT"`.
#' @param contract_type One of `"PERPETUAL"`, `"CURRENT_MONTH"`,
#'   `"NEXT_MONTH"`, `"CURRENT_QUARTER"`, or `"NEXT_QUARTER"`.
#' @param interval Kline interval string.
#' @param startTime Optional start time in milliseconds since Unix epoch.
#' @param endTime Optional end time in milliseconds since Unix epoch.
#' @param limit Maximum number of rows to return. Must not exceed `1500`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_continuous_klines <- function(
    pair,
    contract_type = c("PERPETUAL", "CURRENT_MONTH", "NEXT_MONTH", "CURRENT_QUARTER", "NEXT_QUARTER"),
    interval,
    startTime = NULL,
    endTime = NULL,
    limit = 500,
    json_list = FALSE,
    config = config_futures()) {
  .validate_symbol(pair, "pair")
  contract_type <- match.arg(contract_type)
  .validate_scalar_character(interval, "interval")
  .futures_validate_limit(limit, max = 1500)
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .validate_json_list_flag(json_list)

  payload <- .request_public(
    config,
    "/fapi/v1/continuousKlines",
    query = list(pair = pair, contractType = contract_type, interval = interval, startTime = startTime, endTime = endTime, limit = limit)
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  .futures_as_klines_dt(payload)
}

#' @noRd
.futures_validate_limit <- function(limit, max = 1000L, arg = "limit") {
  .validate_positive_integerish(limit, arg)
  if (limit > max) {
    stop(sprintf("`%s` must be less than or equal to %d.", arg, max), call. = FALSE)
  }
  invisible(limit)
}

#' @noRd
.futures_validate_time_window <- function(start_time, end_time, max_diff_ms, label) {
  if (is.null(start_time) || is.null(end_time)) {
    return(invisible(TRUE))
  }
  if (end_time < start_time) {
    stop("`endTime` must be greater than or equal to `startTime`.", call. = FALSE)
  }
  if ((end_time - start_time) > max_diff_ms) {
    stop(sprintf("The time between `startTime` and `endTime` must not exceed %s.", label), call. = FALSE)
  }
  invisible(TRUE)
}

#' @noRd
.futures_as_klines_dt <- function(payload) {
  kline_dt <- data.table::rbindlist(payload)
  names(kline_dt) <- c(
    "datetime", "open", "high", "low", "close", "volume",
    "datetime_close", "quote_asset_volume", "num_trades",
    "taker_buy_base_vol", "taker_buy_quote_vol", "ignore"
  )
  kline_dt <- .normalize_time_cols(kline_dt, c("datetime", "datetime_close"))
  .coerce_numeric_cols(
    kline_dt,
    c("open", "high", "low", "close", "volume", "quote_asset_volume",
      "num_trades", "taker_buy_base_vol", "taker_buy_quote_vol", "ignore")
  )
}

#' @noRd
.futures_get_derived_klines <- function(path, symbol, interval, startTime, endTime, limit, json_list, config) {
  .validate_symbol(symbol)
  .validate_scalar_character(interval, "interval")
  .futures_validate_limit(limit, max = 1500)
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .validate_json_list_flag(json_list)

  payload <- .request_public(
    config,
    path,
    query = list(symbol = symbol, interval = interval, startTime = startTime, endTime = endTime, limit = limit)
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  .futures_as_klines_dt(payload)
}

test_that("futures market endpoints validate and shape responses", {
  cfg <- config_futures()

  expect_error(
    futures_get_order_book("BTCUSDT", limit = 7, config = cfg),
    "limit"
  )

  expect_error(
    futures_get_aggregate_trades(
      symbol = "BTCUSDT",
      fromId = 1,
      startTime = 0,
      config = cfg
    ),
    "fromId"
  )

  local_mocked_bindings(
    .request_public = function(config, path, query = NULL) {
      switch(
        path,
        "/fapi/v1/trades" = list(list(id = "1", price = "100", qty = "2", quoteQty = "200", time = 1000)),
        "/fapi/v1/aggTrades" = list(list(a = "1", p = "100", q = "2", f = "1", l = "1", T = 2000)),
        "/fapi/v1/fundingRate" = list(list(symbol = "BTCUSDT", fundingRate = "0.01", fundingTime = 3000, markPrice = "100")),
        "/fapi/v1/markPriceKlines" = list(list(0, "1", "2", "0.5", "1.5", "10", 1000, "15", 2, "4", "6", "0")),
        list(path = path, query = query)
      )
    },
    .package = "binxr"
  )

  trades_out <- futures_get_recent_trades("BTCUSDT")
  agg_out <- futures_get_aggregate_trades("BTCUSDT")
  funding_out <- futures_get_funding_rate_history(symbol = "BTCUSDT")
  kline_out <- futures_get_mark_price_klines("BTCUSDT", "1m")

  expect_s3_class(trades_out, "data.table")
  expect_equal(trades_out$id, 1)
  expect_s3_class(trades_out$time, "POSIXct")

  expect_s3_class(agg_out, "data.table")
  expect_equal(agg_out$a, 1)
  expect_s3_class(agg_out$T, "POSIXct")

  expect_s3_class(funding_out, "data.table")
  expect_equal(funding_out$fundingRate, 0.01)
  expect_s3_class(funding_out$fundingTime, "POSIXct")

  expect_s3_class(kline_out, "data.table")
  expect_equal(kline_out$open, 1)
})

#' Set Binance Futures margin type
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param margin_type One of `"CROSSED"` or `"ISOLATED"`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_set_margin_type <- function(
    symbol,
    margin_type = c("CROSSED", "ISOLATED"),
    config = config_futures()) {
  .validate_symbol(symbol)
  margin_type <- match.arg(margin_type)
  .request_signed(
    config,
    "/fapi/v1/marginType",
    params = list(symbol = symbol, marginType = margin_type),
    method = "POST"
  )
}

#' @rdname futures_set_margin_type
#' @param marginType Legacy alias for `margin_type`.
#' @export
set_fapi_account_margin_type <- function(
    symbol,
    marginType = c("CROSSED", "ISOLATED"),
    config = binxr_config_futures()) {
  .binxr_deprecate_alias("set_fapi_account_margin_type", "futures_set_margin_type")
  futures_set_margin_type(symbol = symbol, margin_type = match.arg(marginType), config = config)
}

#' Set Binance Futures leverage
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param leverage Desired leverage.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_set_leverage <- function(symbol, leverage, config = config_futures()) {
  .validate_symbol(symbol)
  .validate_positive_integerish(leverage, "leverage")
  .request_signed(
    config,
    "/fapi/v1/leverage",
    params = list(symbol = symbol, leverage = leverage),
    method = "POST"
  )
}

#' @rdname futures_set_leverage
#' @export
set_fapi_account_leverage <- function(symbol, leverage, config = binxr_config_futures()) {
  .binxr_deprecate_alias("set_fapi_account_leverage", "futures_set_leverage")
  futures_set_leverage(symbol = symbol, leverage = leverage, config = config)
}

#' Set Binance Futures position mode
#'
#' @param dual_side_position `TRUE` for hedge mode, `FALSE` for one-way mode.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_set_position_mode <- function(dual_side_position = TRUE, config = config_futures()) {
  .validate_scalar_logical(dual_side_position, "dual_side_position")
  .request_signed(
    config,
    "/fapi/v1/positionSide",
    params = list(dualSidePosition = if (dual_side_position) "true" else "false"),
    method = "POST"
  )
}

#' @rdname futures_set_position_mode
#' @param dualSidePosition Legacy alias for `dual_side_position`.
#' @export
set_fapi_account_position_side <- function(dualSidePosition = TRUE, config = binxr_config_futures()) {
  .binxr_deprecate_alias("set_fapi_account_position_side", "futures_set_position_mode")
  futures_set_position_mode(dual_side_position = dualSidePosition, config = config)
}

#' Place a Binance Futures order
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param side One of `"BUY"` or `"SELL"`.
#' @param type Order type.
#' @param quantity Order quantity.
#' @param price Optional limit price.
#' @param time_in_force Optional time-in-force value. Required for `LIMIT`.
#' @param position_side One of `"BOTH"`, `"LONG"`, or `"SHORT"`.
#' @param reduce_only Whether the order is reduce-only.
#' @param stop_price Optional trigger price for conditional orders.
#' @param working_type Optional trigger source. Must only be supplied with
#'   `stop_price`.
#' @param new_client_order_id Optional client order identifier.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_place_order <- function(
    symbol,
    side = c("BUY", "SELL"),
    type = c(
      "LIMIT", "MARKET", "STOP", "STOP_MARKET",
      "TAKE_PROFIT", "TAKE_PROFIT_MARKET", "TRAILING_STOP_MARKET"
    ),
    quantity,
    price = NULL,
    time_in_force = NULL,
    position_side = c("BOTH", "LONG", "SHORT"),
    reduce_only = FALSE,
    stop_price = NULL,
    working_type = NULL,
    new_client_order_id = NULL,
    config = config_futures()) {
  .validate_symbol(symbol)
  side <- match.arg(side)
  type <- match.arg(type)
  position_side <- match.arg(position_side)
  .validate_positive_number(quantity, "quantity")
  .validate_optional_scalar_numeric(price, "price")
  .validate_optional_scalar_numeric(stop_price, "stop_price")
  .validate_optional_scalar_character(time_in_force, "time_in_force")
  .validate_optional_scalar_character(working_type, "working_type")
  .validate_optional_scalar_character(new_client_order_id, "new_client_order_id")
  .validate_scalar_logical(reduce_only, "reduce_only")

  if (identical(type, "LIMIT")) {
    if (is.null(price)) {
      stop("`price` is required when `type = \"LIMIT\"`.", call. = FALSE)
    }
    if (is.null(time_in_force)) {
      stop("`time_in_force` is required when `type = \"LIMIT\"`.", call. = FALSE)
    }
    .validate_one_of(time_in_force, c("GTC", "IOC", "FOK", "GTX"), "time_in_force")
  } else if (!is.null(time_in_force)) {
    stop("`time_in_force` must only be supplied when `type = \"LIMIT\"`.", call. = FALSE)
  }

  if (is.null(stop_price) && !is.null(working_type)) {
    stop("`working_type` must only be supplied when `stop_price` is provided.", call. = FALSE)
  }
  if (!is.null(working_type)) {
    .validate_one_of(working_type, c("CONTRACT_PRICE", "MARK_PRICE"), "working_type")
  }

  .request_signed(
    config,
    "/fapi/v1/order",
    params = list(
      symbol = symbol,
      side = side,
      type = type,
      quantity = quantity,
      price = price,
      timeInForce = time_in_force,
      positionSide = position_side,
      reduceOnly = if (isTRUE(reduce_only)) "true" else "false",
      stopPrice = stop_price,
      workingType = working_type,
      newClientOrderId = new_client_order_id
    ),
    method = "POST"
  )
}

#' @rdname futures_place_order
#' @param timeInForce Legacy alias for `time_in_force`.
#' @param positionSide Legacy alias for `position_side`.
#' @param reduceOnly Legacy alias for `reduce_only`.
#' @param stopPrice Legacy alias for `stop_price`.
#' @param workingType Legacy alias for `working_type`.
#' @param newClientOrderId Legacy alias for `new_client_order_id`.
#' @export
place_fapi_trade_order <- function(
    symbol,
    side = c("BUY", "SELL"),
    type = c(
      "LIMIT", "MARKET", "STOP", "STOP_MARKET",
      "TAKE_PROFIT", "TAKE_PROFIT_MARKET", "TRAILING_STOP_MARKET"
    ),
    quantity,
    price = NULL,
    timeInForce = c("GTC", "IOC", "FOK", "GTX"),
    positionSide = c("BOTH", "LONG", "SHORT"),
    reduceOnly = FALSE,
    stopPrice = NULL,
    workingType = c("CONTRACT_PRICE", "MARK_PRICE"),
    newClientOrderId = NULL,
    config = binxr_config_futures()) {
  .binxr_deprecate_alias("place_fapi_trade_order", "futures_place_order")
  type <- match.arg(type)
  futures_place_order(
    symbol = symbol,
    side = match.arg(side),
    type = type,
    quantity = quantity,
    price = price,
    time_in_force = if (identical(type, "LIMIT")) match.arg(timeInForce) else NULL,
    position_side = match.arg(positionSide),
    reduce_only = reduceOnly,
    stop_price = stopPrice,
    working_type = if (!is.null(stopPrice)) match.arg(workingType) else NULL,
    new_client_order_id = newClientOrderId,
    config = config
  )
}

#' Cancel a Binance Futures order
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param order_id Optional exchange order ID.
#' @param orig_client_order_id Optional client order ID.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_cancel_order <- function(
    symbol,
    order_id = NULL,
    orig_client_order_id = NULL,
    config = config_futures()) {
  .validate_symbol(symbol)
  .validate_required_any(order_id, orig_client_order_id, arg_names = c("order_id", "orig_client_order_id"))
  .validate_optional_scalar_numeric(order_id, "order_id")
  .validate_optional_scalar_character(orig_client_order_id, "orig_client_order_id")
  .request_signed(
    config,
    "/fapi/v1/order",
    params = list(symbol = symbol, orderId = order_id, origClientOrderId = orig_client_order_id),
    method = "DELETE"
  )
}

#' @rdname futures_cancel_order
#' @param orderId Legacy alias for `order_id`.
#' @param origClientOrderId Legacy alias for `orig_client_order_id`.
#' @export
cancel_fapi_trade_order <- function(
    symbol,
    orderId = NULL,
    origClientOrderId = NULL,
    config = binxr_config_futures()) {
  .binxr_deprecate_alias("cancel_fapi_trade_order", "futures_cancel_order")
  futures_cancel_order(
    symbol = symbol,
    order_id = orderId,
    orig_client_order_id = origClientOrderId,
    config = config
  )
}

#' Cancel all open Binance Futures orders for a symbol
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_cancel_all_orders <- function(symbol, config = config_futures()) {
  .validate_symbol(symbol)
  .request_signed(
    config,
    "/fapi/v1/allOpenOrders",
    params = list(symbol = symbol),
    method = "DELETE"
  )
}

#' @rdname futures_cancel_all_orders
#' @export
cancel_fapi_trade_orders_all <- function(symbol, config = binxr_config_futures()) {
  .binxr_deprecate_alias("cancel_fapi_trade_orders_all", "futures_cancel_all_orders")
  futures_cancel_all_orders(symbol = symbol, config = config)
}

#' Get Binance Futures open orders
#'
#' @param symbol Optional trading pair symbol. If `NULL`, returns all open
#'   orders.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_open_orders <- function(symbol = NULL, json_list = FALSE, config = config_futures()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  payload <- .request_signed(
    config,
    "/fapi/v1/openOrders",
    params = list(symbol = symbol),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }
  orders_dt <- .maybe_as_dt(payload)
  .normalize_order_time_cols(orders_dt)
}

#' @rdname futures_get_open_orders
#' @export
get_fapi_trade_open_orders <- function(symbol = NULL, config = binxr_config_futures()) {
  .binxr_deprecate_alias("get_fapi_trade_open_orders", "futures_get_open_orders")
  futures_get_open_orders(symbol = symbol, config = config)
}

#' Get Binance Futures order history
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param limit Maximum number of orders to return.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_orders <- function(symbol, limit = 500, json_list = FALSE, config = config_futures()) {
  .validate_symbol(symbol)
  .validate_positive_integerish(limit, "limit")
  payload <- .request_signed(
    config,
    "/fapi/v1/allOrders",
    params = list(symbol = symbol, limit = limit),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }
  orders_dt <- .maybe_as_dt(payload)
  .normalize_order_time_cols(orders_dt)
}

#' @rdname futures_get_orders
#' @export
get_fapi_trade_orders <- function(symbol, limit = 500, config = binxr_config_futures()) {
  .binxr_deprecate_alias("get_fapi_trade_orders", "futures_get_orders")
  futures_get_orders(symbol = symbol, limit = limit, config = config)
}

#' Get a Binance Futures order
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param order_id Optional exchange order ID.
#' @param orig_client_order_id Optional client order ID.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_get_order <- function(
    symbol,
    order_id = NULL,
    orig_client_order_id = NULL,
    config = config_futures()) {
  .validate_symbol(symbol)
  .validate_required_any(order_id, orig_client_order_id, arg_names = c("order_id", "orig_client_order_id"))
  .validate_optional_scalar_numeric(order_id, "order_id")
  .validate_optional_scalar_character(orig_client_order_id, "orig_client_order_id")
  .request_signed(
    config,
    "/fapi/v1/order",
    params = list(symbol = symbol, orderId = order_id, origClientOrderId = orig_client_order_id),
    method = "GET"
  )
}

#' @rdname futures_get_order
#' @param orderId Legacy alias for `order_id`.
#' @param origClientOrderId Legacy alias for `orig_client_order_id`.
#' @export
get_fapi_trade_order <- function(
    symbol,
    orderId = NULL,
    origClientOrderId = NULL,
    config = binxr_config_futures()) {
  .binxr_deprecate_alias("get_fapi_trade_order", "futures_get_order")
  futures_get_order(
    symbol = symbol,
    order_id = orderId,
    orig_client_order_id = origClientOrderId,
    config = config
  )
}

#' Test a Binance Futures order
#'
#' @inheritParams futures_place_order
#'
#' @return A parsed list.
#' @export
futures_test_order <- function(
    symbol,
    side = c("BUY", "SELL"),
    type = c(
      "LIMIT", "MARKET", "STOP", "STOP_MARKET",
      "TAKE_PROFIT", "TAKE_PROFIT_MARKET", "TRAILING_STOP_MARKET"
    ),
    quantity,
    price = NULL,
    time_in_force = NULL,
    position_side = c("BOTH", "LONG", "SHORT"),
    reduce_only = FALSE,
    stop_price = NULL,
    working_type = NULL,
    new_client_order_id = NULL,
    config = config_futures()) {
  .validate_symbol(symbol)
  side <- match.arg(side)
  type <- match.arg(type)
  position_side <- match.arg(position_side)
  .validate_positive_number(quantity, "quantity")
  .validate_optional_scalar_numeric(price, "price")
  .validate_optional_scalar_numeric(stop_price, "stop_price")
  .validate_optional_scalar_character(time_in_force, "time_in_force")
  .validate_optional_scalar_character(working_type, "working_type")
  .validate_optional_scalar_character(new_client_order_id, "new_client_order_id")
  .validate_scalar_logical(reduce_only, "reduce_only")

  if (identical(type, "LIMIT")) {
    if (is.null(price)) {
      stop("`price` is required when `type = \"LIMIT\"`.", call. = FALSE)
    }
    if (is.null(time_in_force)) {
      stop("`time_in_force` is required when `type = \"LIMIT\"`.", call. = FALSE)
    }
    .validate_one_of(time_in_force, c("GTC", "IOC", "FOK", "GTX"), "time_in_force")
  } else if (!is.null(time_in_force)) {
    stop("`time_in_force` must only be supplied when `type = \"LIMIT\"`.", call. = FALSE)
  }

  if (is.null(stop_price) && !is.null(working_type)) {
    stop("`working_type` must only be supplied when `stop_price` is provided.", call. = FALSE)
  }
  if (!is.null(working_type)) {
    .validate_one_of(working_type, c("CONTRACT_PRICE", "MARK_PRICE"), "working_type")
  }

  .request_signed(
    config,
    "/fapi/v1/order/test",
    params = list(
      symbol = symbol,
      side = side,
      type = type,
      quantity = quantity,
      price = price,
      timeInForce = time_in_force,
      positionSide = position_side,
      reduceOnly = if (isTRUE(reduce_only)) "true" else "false",
      stopPrice = stop_price,
      workingType = working_type,
      newClientOrderId = new_client_order_id
    ),
    method = "POST"
  )
}

#' Set Binance Futures multi-assets mode
#'
#' @param multi_assets_margin `TRUE` to enable multi-assets mode, `FALSE` to disable it.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_set_multi_assets_mode <- function(multi_assets_margin = TRUE, config = config_futures()) {
  .validate_scalar_logical(multi_assets_margin, "multi_assets_margin")
  .request_signed(
    config,
    "/fapi/v1/multiAssetsMargin",
    params = list(multiAssetsMargin = if (isTRUE(multi_assets_margin)) "true" else "false"),
    method = "POST"
  )
}

#' Set Binance Futures countdown cancel-all
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param countdown_time Countdown time in milliseconds.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_countdown_cancel_all <- function(symbol, countdown_time, config = config_futures()) {
  .validate_symbol(symbol)
  .validate_positive_integerish(countdown_time, "countdown_time")
  .request_signed(
    config,
    "/fapi/v1/countdownCancelAll",
    params = list(symbol = symbol, countdownTime = countdown_time),
    method = "POST"
  )
}

#' Get Binance Futures account trades
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param order_id Optional order ID.
#' @param startTime Optional start time in milliseconds since Unix epoch.
#' @param endTime Optional end time in milliseconds since Unix epoch.
#' @param fromId Optional trade ID to start from.
#' @param limit Maximum number of rows to return. Must not exceed `1000`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_account_trades <- function(
    symbol,
    order_id = NULL,
    startTime = NULL,
    endTime = NULL,
    fromId = NULL,
    limit = 500,
    json_list = FALSE,
    config = config_futures()) {
  .validate_symbol(symbol)
  .validate_optional_scalar_numeric(order_id, "order_id")
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .validate_optional_scalar_numeric(fromId, "fromId")
  .futures_validate_limit(limit)
  .validate_json_list_flag(json_list)

  if (!is.null(fromId) && (!is.null(startTime) || !is.null(endTime))) {
    stop("`fromId` must not be supplied with `startTime` or `endTime`.", call. = FALSE)
  }
  .futures_validate_time_window(startTime, endTime, 7L * 24L * 60L * 60L * 1000L, "7 days")

  payload <- .request_signed(
    config,
    "/fapi/v1/userTrades",
    params = list(
      symbol = symbol,
      orderId = order_id,
      startTime = startTime,
      endTime = endTime,
      fromId = fromId,
      limit = limit
    ),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  trades_dt <- .maybe_as_dt(payload)
  trades_dt <- .coerce_numeric_cols(
    trades_dt,
    c("commission", "id", "orderId", "price", "qty", "quoteQty", "realizedPnl")
  )
  .normalize_time_cols(trades_dt, "time")
}

#' Get Binance Futures force orders
#'
#' @param symbol Optional trading pair symbol.
#' @param auto_close_type Optional auto-close type.
#' @param startTime Optional start time in milliseconds since Unix epoch.
#' @param endTime Optional end time in milliseconds since Unix epoch.
#' @param limit Maximum number of rows to return. Must not exceed `100`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_force_orders <- function(
    symbol = NULL,
    auto_close_type = NULL,
    startTime = NULL,
    endTime = NULL,
    limit = 50,
    json_list = FALSE,
    config = config_futures()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  .validate_optional_scalar_character(auto_close_type, "auto_close_type")
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .futures_validate_limit(limit, max = 100)
  .validate_json_list_flag(json_list)

  payload <- .request_signed(
    config,
    "/fapi/v1/forceOrders",
    params = list(symbol = symbol, autoCloseType = auto_close_type, startTime = startTime, endTime = endTime, limit = limit),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  orders_dt <- .maybe_as_dt(payload)
  orders_dt <- .coerce_numeric_cols(
    orders_dt,
    c("price", "origQty", "executedQty", "averagePrice", "cumQuote", "orderId")
  )
  .normalize_time_cols(orders_dt, "time")
}

test_that("futures_place_order rejects silently ignored parameters", {
  cfg <- config_futures(api_key = "k", secret_key = "s")

  expect_error(
    futures_place_order(
      symbol = "BTCUSDT",
      side = "BUY",
      type = "MARKET",
      quantity = 1,
      time_in_force = "GTC",
      config = cfg
    ),
    "time_in_force"
  )

  expect_error(
    futures_place_order(
      symbol = "BTCUSDT",
      side = "BUY",
      type = "LIMIT",
      quantity = 1,
      price = 100,
      time_in_force = "GTC",
      working_type = "MARK_PRICE",
      config = cfg
    ),
    "working_type"
  )
})

test_that("deprecated market aliases warn and forward to canonical names", {
  local_mocked_bindings(
    .request_public = function(config, path, query = NULL) {
      switch(
        path,
        "/api/v3/ticker/price" = list(symbol = query$symbol, price = "100"),
        "/fapi/v1/time" = list(serverTime = 1000),
        list(path = path, query = query)
      )
    },
    .package = "binxr"
  )

  spot_out <- NULL
  futures_out <- NULL
  time_out <- NULL
  expect_warning(
    spot_out <- get_spot_mark_price("BTCUSDT"),
    "deprecated"
  )
  expect_warning(
    futures_out <- get_fapi_mark_price("ETHUSDT"),
    "deprecated"
  )
  expect_warning(
    time_out <- get_fapi_system_time(),
    "deprecated"
  )

  expect_identical(spot_out$symbol, "BTCUSDT")
  expect_identical(futures_out$query$symbol, "ETHUSDT")
  expect_s3_class(time_out, "POSIXct")
})

test_that("deprecated futures trade aliases warn and forward to canonical names", {
  cfg <- config_futures(api_key = "k", secret_key = "s")

  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "PUT", "DELETE")) {
      list(path = path, params = params, method = method)
    },
    .package = "binxr"
  )

  order_out <- NULL
  place_out <- NULL
  expect_warning(
    order_out <- get_fapi_trade_order("BTCUSDT", orderId = 1, config = cfg),
    "deprecated"
  )
  expect_warning(
    place_out <- place_fapi_trade_order(
      symbol = "BTCUSDT",
      side = "BUY",
      type = "LIMIT",
      quantity = 1,
      price = 100,
      timeInForce = "GTC",
      config = cfg
    ),
    "deprecated"
  )

  expect_identical(order_out$path, "/fapi/v1/order")
  expect_identical(order_out$params$orderId, 1)
  expect_identical(place_out$method, "POST")
  expect_identical(place_out$params$timeInForce, "GTC")
})

test_that("futures trade expansion validates and forwards params", {
  cfg <- config_futures(api_key = "k", secret_key = "s")

  expect_error(
    futures_get_account_trades(
      symbol = "BTCUSDT",
      fromId = 1,
      startTime = 0,
      config = cfg
    ),
    "fromId"
  )

  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "PUT", "DELETE")) {
      switch(
        path,
        "/fapi/v1/userTrades" = list(list(id = "1", orderId = "2", price = "100", qty = "1", quoteQty = "100", realizedPnl = "0.5", commission = "-0.1", time = 1000)),
        "/fapi/v1/forceOrders" = list(list(orderId = "3", price = "90", origQty = "1", executedQty = "1", averagePrice = "90", cumQuote = "90", time = 2000)),
        list(path = path, params = params, method = method)
      )
    },
    .package = "binxr"
  )

  test_out <- futures_test_order(
    symbol = "BTCUSDT",
    side = "BUY",
    type = "LIMIT",
    quantity = 1,
    price = 100,
    time_in_force = "GTC",
    config = cfg
  )
  multi_assets_out <- futures_set_multi_assets_mode(TRUE, config = cfg)
  countdown_out <- futures_countdown_cancel_all("BTCUSDT", 5000, config = cfg)
  trades_out <- futures_get_account_trades("BTCUSDT", config = cfg)
  force_out <- futures_get_force_orders(symbol = "BTCUSDT", config = cfg)

  expect_identical(test_out$path, "/fapi/v1/order/test")
  expect_identical(multi_assets_out$params$multiAssetsMargin, "true")
  expect_identical(countdown_out$params$countdownTime, 5000)

  expect_s3_class(trades_out, "data.table")
  expect_equal(trades_out$id, 1)
  expect_s3_class(trades_out$time, "POSIXct")

  expect_s3_class(force_out, "data.table")
  expect_equal(force_out$orderId, 3)
  expect_s3_class(force_out$time, "POSIXct")
})

library(testthat)
library(binxr)

if (inherits(try(loadNamespace("waldo"), silent = TRUE), "try-error")) {
  assignInNamespace(
    "waldo_compare",
    function(x, y, ..., x_arg = "x", y_arg = "y") {
      if (identical(x, y)) {
        character()
      } else {
        paste0(x_arg, " is not identical to ", y_arg)
      }
    },
    ns = "testthat"
  )
}

test_check("binxr")

#' (Optional) Coerce account to data.frames
#' @param account get_account() result
#' @return list(assets=..., positions=...)
#' @examples
#' account <- list(
#'   assets = list(list(asset = "USDT", walletBalance = "10")),
#'   positions = list(list(symbol = "ETHUSDT", positionAmt = "0"))
#' )
#' coerce_account_frames(account)
#' @export
coerce_account_frames <- function(account) {
  to_df <- function(x) {
    if (!length(x)) {
      return(data.frame())
    }
    jsonlite::fromJSON(jsonlite::toJSON(x), simplifyDataFrame = TRUE)
  }
  list(assets = to_df(account$assets), positions = to_df(account$positions))
}

#' Get Binance Futures account info
#'
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_get_account <- function(config = config_futures()) {
  .request_signed(config, "/fapi/v2/account", params = list(), method = "GET")
}

#' @rdname futures_get_account
#' @export
get_fapi_account <- function(config = binxr_config_futures()) {
  .binxr_deprecate_alias("get_fapi_account", "futures_get_account")
  futures_get_account(config = config)
}

#' Get Binance Futures position risk
#'
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_position_risk <- function(json_list = FALSE, config = config_futures()) {
  payload <- .request_signed(config, "/fapi/v2/positionRisk", params = list(), method = "GET")
  if (isTRUE(json_list)) {
    return(payload)
  }

  risk_dt <- .maybe_as_dt(payload)
  risk_dt <- .coerce_numeric_cols(
    risk_dt,
    c("positionAmt", "entryPrice", "breakEvenPrice", "markPrice", "unRealizedProfit",
      "liquidationPrice", "isolatedMargin", "notional", "isolatedWallet", "adl")
  )
  .normalize_time_cols(risk_dt, "updateTime")
}

#' @rdname futures_get_position_risk
#' @export
get_fapi_account_position_risk <- function(config = binxr_config_futures()) {
  .binxr_deprecate_alias("get_fapi_account_position_risk", "futures_get_position_risk")
  futures_get_position_risk(config = config)
}

#' Get standardized Binance Futures positions
#'
#' @param acc Optional account payload from [futures_get_account()]. If `NULL`,
#'   account data is fetched.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_positions <- function(acc = NULL, json_list = FALSE, config = config_futures()) {
  if (is.null(acc)) {
    acc <- futures_get_account(config = config)
  }
  positions <- acc$positions %||% list()
  if (isTRUE(json_list)) {
    return(positions)
  }

  pos_dt <- .maybe_as_dt(positions)
  pos_dt <- .coerce_numeric_cols(
    pos_dt,
    c(
      "positionAmt", "initialMargin", "maintMargin", "entryPrice", "breakEvenPrice",
      "unrealizedProfit", "positionInitialMargin", "openOrderInitialMargin",
      "isolatedMargin", "notional", "isolatedWallet", "bidNotional", "askNotional"
    )
  )
  if ("leverage" %in% names(pos_dt)) {
    data.table::set(pos_dt, j = "leverage", value = as.integer(pos_dt$leverage))
  }
  .normalize_time_cols(pos_dt, "updateTime")
}

#' @rdname futures_get_positions
#' @export
get_fapi_account_positions <- function(acc = NULL, config = binxr_config_futures()) {
  .binxr_deprecate_alias("get_fapi_account_positions", "futures_get_positions")
  futures_get_positions(acc = acc, config = config)
}

#' Get Binance Futures account summary
#'
#' @param acc Optional account payload from [futures_get_account()]. If `NULL`,
#'   account data is fetched.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A one-row `data.table` by default, or a parsed list when
#'   `json_list = TRUE`.
#' @export
futures_get_account_summary <- function(acc = NULL, json_list = FALSE, config = config_futures()) {
  if (is.null(acc)) {
    acc <- futures_get_account(config = config)
  }

  summary_list <- list(
    totalWalletBalance = as.numeric(acc$totalWalletBalance),
    totalUnrealizedProfit = as.numeric(acc$totalUnrealizedProfit),
    totalInitialMargin = as.numeric(acc$totalInitialMargin),
    totalMaintMargin = as.numeric(acc$totalMaintMargin),
    availableBalance = as.numeric(acc$availableBalance),
    maxWithdrawAmount = as.numeric(acc$maxWithdrawAmount)
  )

  if (isTRUE(json_list)) {
    return(summary_list)
  }

  data.table::as.data.table(summary_list)
}

#' @rdname futures_get_account_summary
#' @export
get_fapi_account_summary <- function(acc = NULL, config = binxr_config_futures()) {
  .binxr_deprecate_alias("get_fapi_account_summary", "futures_get_account_summary")
  futures_get_account_summary(acc = acc, config = config)
}

#' Get Binance Futures balances
#'
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_balance <- function(json_list = FALSE, config = config_futures()) {
  payload <- .request_signed(config, "/fapi/v3/balance", params = list(), method = "GET")
  if (isTRUE(json_list)) {
    return(payload)
  }

  balance_dt <- .maybe_as_dt(payload)
  balance_dt <- .coerce_numeric_cols(
    balance_dt,
    c("balance", "crossWalletBalance", "crossUnPnl", "availableBalance", "maxWithdrawAmount")
  )
  .normalize_time_cols(balance_dt, "updateTime")
}

#' Get Binance Futures position mode
#'
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_get_position_mode <- function(config = config_futures()) {
  .request_signed(config, "/fapi/v1/positionSide/dual", params = list(), method = "GET")
}

#' Get Binance Futures multi-assets mode
#'
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_get_multi_assets_mode <- function(config = config_futures()) {
  .request_signed(config, "/fapi/v1/multiAssetsMargin", params = list(), method = "GET")
}

#' Get Binance Futures commission rate
#'
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A parsed list.
#' @export
futures_get_commission_rate <- function(symbol, config = config_futures()) {
  .validate_symbol(symbol)
  .request_signed(config, "/fapi/v1/commissionRate", params = list(symbol = symbol), method = "GET")
}

#' Get Binance Futures order rate limits
#'
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A futures configuration created by [config_futures()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
futures_get_order_rate_limit <- function(json_list = FALSE, config = config_futures()) {
  payload <- .request_signed(config, "/fapi/v1/rateLimit/order", params = list(), method = "GET")
  if (isTRUE(json_list)) {
    return(payload)
  }

  .coerce_numeric_cols(.maybe_as_dt(payload), c("intervalNum", "limit"))
}

test_that("futures account expansion endpoints call signed paths and shape outputs", {
  cfg <- config_futures(api_key = "k", secret_key = "s")

  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "PUT", "DELETE")) {
      switch(
        path,
        "/fapi/v3/balance" = list(list(asset = "USDT", balance = "10", crossWalletBalance = "9", crossUnPnl = "1", availableBalance = "8", maxWithdrawAmount = "7", updateTime = 1000)),
        "/fapi/v1/rateLimit/order" = list(list(intervalNum = 1, limit = 1200)),
        list(path = path, params = params, method = method)
      )
    },
    .package = "binxr"
  )

  balance_out <- futures_get_balance(config = cfg)
  position_mode_out <- futures_get_position_mode(config = cfg)
  multi_assets_out <- futures_get_multi_assets_mode(config = cfg)
  commission_out <- futures_get_commission_rate("BTCUSDT", config = cfg)
  rate_limit_out <- futures_get_order_rate_limit(config = cfg)

  expect_s3_class(balance_out, "data.table")
  expect_equal(balance_out$balance, 10)
  expect_s3_class(balance_out$updateTime, "POSIXct")

  expect_identical(position_mode_out$path, "/fapi/v1/positionSide/dual")
  expect_identical(multi_assets_out$path, "/fapi/v1/multiAssetsMargin")
  expect_identical(commission_out$params$symbol, "BTCUSDT")

  expect_s3_class(rate_limit_out, "data.table")
  expect_equal(rate_limit_out$limit, 1200)
})

test_that("deprecated futures account aliases warn and forward", {
  acc <- list(
    totalWalletBalance = "10",
    totalUnrealizedProfit = "2",
    totalInitialMargin = "3",
    totalMaintMargin = "1",
    availableBalance = "7",
    maxWithdrawAmount = "6"
  )

  out <- NULL
  expect_warning(
    out <- get_fapi_account_summary(acc = acc),
    "deprecated"
  )

  expect_s3_class(out, "data.table")
  expect_equal(out$availableBalance, 7)
})

test_that("futures_get_positions keeps zero and short positions by default", {
  acc <- list(
    positions = list(
      list(symbol = "BTCUSDT", positionAmt = "0", initialMargin = "0", leverage = "20", updateTime = 0),
      list(symbol = "ETHUSDT", positionAmt = "-1.5", initialMargin = "12.3", leverage = "10", updateTime = 1000)
    )
  )

  out <- futures_get_positions(acc = acc)

  expect_s3_class(out, "data.table")
  expect_equal(nrow(out), 2L)
  expect_equal(out$positionAmt, c(0, -1.5))
})

test_that("futures_get_positions returns raw list when json_list is TRUE", {
  positions <- list(list(symbol = "BTCUSDT", positionAmt = "1"))
  acc <- list(positions = positions)

  out <- futures_get_positions(acc = acc, json_list = TRUE)

  expect_identical(out, positions)
})

test_that("futures_get_account_summary supports data.table and list return shapes", {
  acc <- list(
    totalWalletBalance = "10",
    totalUnrealizedProfit = "2",
    totalInitialMargin = "3",
    totalMaintMargin = "1",
    availableBalance = "7",
    maxWithdrawAmount = "6"
  )

  dt_out <- futures_get_account_summary(acc = acc)
  list_out <- futures_get_account_summary(acc = acc, json_list = TRUE)

  expect_s3_class(dt_out, "data.table")
  expect_equal(dt_out$totalWalletBalance, 10)
  expect_type(list_out, "list")
  expect_equal(list_out$availableBalance, 7)
})

#' Place a Binance Options order
#'
#' @param symbol Option symbol, for example `"BTC-200730-9000-C"`.
#' @param side One of `"BUY"` or `"SELL"`.
#' @param type Order type. Only `"LIMIT"` is currently supported.
#' @param quantity Order quantity.
#' @param price Order price.
#' @param time_in_force Time in force. Default is `"GTC"`.
#' @param reduce_only Whether the order is reduce-only.
#' @param post_only Whether the order is post-only.
#' @param new_order_resp_type Response type. One of `"ACK"` or `"RESULT"`.
#' @param client_order_id Optional client order ID.
#' @param is_mmp Whether the order is an MMP order.
#' @param self_trade_prevention_mode Optional STP mode.
#' @param config An options configuration created by [config_options()].
#'
#' @return A parsed list.
#' @export
options_place_order <- function(
    symbol,
    side = c("BUY", "SELL"),
    type = c("LIMIT"),
    quantity,
    price,
    time_in_force = c("GTC", "IOC", "FOK"),
    reduce_only = FALSE,
    post_only = FALSE,
    new_order_resp_type = c("ACK", "RESULT"),
    client_order_id = NULL,
    is_mmp = FALSE,
    self_trade_prevention_mode = NULL,
    config = config_options()) {
  .validate_symbol(symbol)
  side <- match.arg(side)
  type <- match.arg(type)
  time_in_force <- match.arg(time_in_force)
  new_order_resp_type <- match.arg(new_order_resp_type)
  .validate_positive_number(quantity, "quantity")
  .validate_positive_number(price, "price")
  .validate_scalar_logical(reduce_only, "reduce_only")
  .validate_scalar_logical(post_only, "post_only")
  .validate_optional_scalar_character(client_order_id, "client_order_id")
  .validate_scalar_logical(is_mmp, "is_mmp")
  .validate_optional_scalar_character(self_trade_prevention_mode, "self_trade_prevention_mode")

  if (!is.null(self_trade_prevention_mode)) {
    .validate_one_of(
      self_trade_prevention_mode,
      c("EXPIRE_TAKER", "EXPIRE_MAKER", "EXPIRE_BOTH"),
      "self_trade_prevention_mode"
    )
  }

  .request_signed(
    config,
    "/eapi/v1/order",
    params = list(
      symbol = symbol,
      side = side,
      type = type,
      quantity = quantity,
      price = price,
      timeInForce = time_in_force,
      reduceOnly = if (isTRUE(reduce_only)) "true" else "false",
      postOnly = if (isTRUE(post_only)) "true" else "false",
      newOrderRespType = new_order_resp_type,
      clientOrderId = client_order_id,
      isMmp = if (isTRUE(is_mmp)) "true" else "false",
      selfTradePreventionMode = self_trade_prevention_mode
    ),
    method = "POST"
  )
}

#' Get a Binance Options order
#'
#' @param symbol Option symbol, for example `"BTC-200730-9000-C"`.
#' @param order_id Optional exchange order ID.
#' @param client_order_id Optional client order ID.
#' @param config An options configuration created by [config_options()].
#'
#' @return A parsed list.
#' @export
options_get_order <- function(symbol, order_id = NULL, client_order_id = NULL, config = config_options()) {
  .validate_symbol(symbol)
  .validate_required_any(order_id, client_order_id, arg_names = c("order_id", "client_order_id"))
  .validate_optional_scalar_numeric(order_id, "order_id")
  .validate_optional_scalar_character(client_order_id, "client_order_id")

  .request_signed(
    config,
    "/eapi/v1/order",
    params = list(symbol = symbol, orderId = order_id, clientOrderId = client_order_id),
    method = "GET"
  )
}

#' Cancel a Binance Options order
#'
#' @param symbol Option symbol, for example `"BTC-200730-9000-C"`.
#' @param order_id Optional exchange order ID.
#' @param client_order_id Optional client order ID.
#' @param config An options configuration created by [config_options()].
#'
#' @return A parsed list.
#' @export
options_cancel_order <- function(symbol, order_id = NULL, client_order_id = NULL, config = config_options()) {
  .validate_symbol(symbol)
  .validate_required_any(order_id, client_order_id, arg_names = c("order_id", "client_order_id"))
  .validate_optional_scalar_numeric(order_id, "order_id")
  .validate_optional_scalar_character(client_order_id, "client_order_id")

  .request_signed(
    config,
    "/eapi/v1/order",
    params = list(symbol = symbol, orderId = order_id, clientOrderId = client_order_id),
    method = "DELETE"
  )
}

#' Cancel all Binance Options orders for a symbol
#'
#' @param symbol Option symbol, for example `"BTC-200730-9000-C"`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A parsed list.
#' @export
options_cancel_all_orders <- function(symbol, config = config_options()) {
  .validate_symbol(symbol)
  .request_signed(config, "/eapi/v1/allOpenOrders", params = list(symbol = symbol), method = "DELETE")
}

#' Cancel all Binance Options orders by underlying
#'
#' @param underlying Underlying symbol, for example `"BTCUSDT"`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A parsed list.
#' @export
options_cancel_all_orders_by_underlying <- function(underlying, config = config_options()) {
  .validate_symbol(underlying, "underlying")
  .request_signed(
    config,
    "/eapi/v1/allOpenOrdersByUnderlying",
    params = list(underlying = underlying),
    method = "DELETE"
  )
}

#' Get Binance Options open orders
#'
#' @param symbol Optional option symbol.
#' @param order_id Optional order ID to start from.
#' @param startTime Optional start time in milliseconds since Unix epoch.
#' @param endTime Optional end time in milliseconds since Unix epoch.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
options_get_open_orders <- function(
    symbol = NULL,
    order_id = NULL,
    startTime = NULL,
    endTime = NULL,
    json_list = FALSE,
    config = config_options()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  .validate_optional_scalar_numeric(order_id, "order_id")
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .validate_json_list_flag(json_list)

  payload <- .request_signed(
    config,
    "/eapi/v1/openOrders",
    params = list(symbol = symbol, orderId = order_id, startTime = startTime, endTime = endTime),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  .options_as_orders_dt(payload)
}

#' Get Binance Options order history
#'
#' @param symbol Option symbol, for example `"BTC-200730-9000-C"`.
#' @param order_id Optional order ID to start from.
#' @param startTime Optional start time in milliseconds since Unix epoch.
#' @param endTime Optional end time in milliseconds since Unix epoch.
#' @param limit Maximum number of rows to return. Must not exceed `1000`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
options_get_order_history <- function(
    symbol,
    order_id = NULL,
    startTime = NULL,
    endTime = NULL,
    limit = 100,
    json_list = FALSE,
    config = config_options()) {
  .validate_symbol(symbol)
  .validate_optional_scalar_numeric(order_id, "order_id")
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .options_validate_limit(limit, max = 1000)
  .validate_json_list_flag(json_list)

  payload <- .request_signed(
    config,
    "/eapi/v1/historyOrders",
    params = list(symbol = symbol, orderId = order_id, startTime = startTime, endTime = endTime, limit = limit),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  .options_as_orders_dt(payload)
}

#' Get Binance Options account trades
#'
#' @param symbol Optional option symbol.
#' @param fromId Optional trade ID to start from.
#' @param startTime Optional start time in milliseconds since Unix epoch.
#' @param endTime Optional end time in milliseconds since Unix epoch.
#' @param limit Maximum number of rows to return. Must not exceed `1000`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
options_get_account_trades <- function(
    symbol = NULL,
    fromId = NULL,
    startTime = NULL,
    endTime = NULL,
    limit = 100,
    json_list = FALSE,
    config = config_options()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  .validate_optional_scalar_numeric(fromId, "fromId")
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .options_validate_limit(limit, max = 1000)
  .validate_json_list_flag(json_list)

  payload <- .request_signed(
    config,
    "/eapi/v1/userTrades",
    params = list(symbol = symbol, fromId = fromId, startTime = startTime, endTime = endTime, limit = limit),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  trades_dt <- .maybe_as_dt(payload)
  trades_dt <- .coerce_numeric_cols(
    trades_dt,
    c("id", "tradeId", "orderId", "price", "quantity", "fee", "realizedProfit")
  )
  .normalize_time_cols(trades_dt, "time")
}

#' Get Binance Options positions
#'
#' @param symbol Optional option symbol.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
options_get_positions <- function(symbol = NULL, json_list = FALSE, config = config_options()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  .validate_json_list_flag(json_list)

  payload <- .request_signed(config, "/eapi/v1/position", params = list(symbol = symbol), method = "GET")
  if (isTRUE(json_list)) {
    return(payload)
  }

  pos_dt <- .maybe_as_dt(payload)
  pos_dt <- .coerce_numeric_cols(
    pos_dt,
    c("entryPrice", "quantity", "markValue", "unrealizedPNL", "markPrice", "strikePrice", "bidQuantity", "askQuantity")
  )
  .normalize_time_cols(pos_dt, c("expiryDate", "time"))
}

#' Get Binance Options user commission
#'
#' @param config An options configuration created by [config_options()].
#'
#' @return A parsed list.
#' @export
options_get_commission <- function(config = config_options()) {
  .request_signed(config, "/eapi/v1/commission", params = list(), method = "GET")
}

#' Get Binance Options exercise records
#'
#' @param symbol Optional option symbol.
#' @param startTime Optional start time in milliseconds since Unix epoch.
#' @param endTime Optional end time in milliseconds since Unix epoch.
#' @param limit Maximum number of rows to return. Must not exceed `1000`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
options_get_exercise_records <- function(
    symbol = NULL,
    startTime = NULL,
    endTime = NULL,
    limit = 1000,
    json_list = FALSE,
    config = config_options()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .options_validate_limit(limit, max = 1000)
  .validate_json_list_flag(json_list)

  payload <- .request_signed(
    config,
    "/eapi/v1/exerciseRecord",
    params = list(symbol = symbol, startTime = startTime, endTime = endTime, limit = limit),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  exercise_dt <- .maybe_as_dt(payload)
  exercise_dt <- .coerce_numeric_cols(exercise_dt, c("id", "exercisePrice", "quantity", "amount", "fee"))
  .normalize_time_cols(exercise_dt, "createDate")
}

#' @noRd
.options_as_orders_dt <- function(payload) {
  orders_dt <- .maybe_as_dt(payload)
  orders_dt <- .coerce_numeric_cols(
    orders_dt,
    c("orderId", "price", "quantity", "executedQty", "avgPrice")
  )
  .normalize_time_cols(orders_dt, c("createTime", "updateTime"))
}

#' Create a Binance futures configuration
#'
#' @param api_key character API key. Leave `NULL` for unsigned/public requests.
#' @param secret_key character Secret key. Leave `NULL` for unsigned/public requests.
#' @param base_url character Base URL for the Binance Futures API.
#' @param recv_window integer recvWindow in milliseconds used for signed requests.
#' @param timeout numeric request timeout in seconds.
#' @param verbose logical whether to enable verbose request debugging in future helpers.
#'
#' @return A `binxr_config` list.
#' @examples
#' cfg <- config_futures(api_key = NULL, secret_key = NULL)
#' cfg$product
#' @export
config_futures <- function(
    api_key = Sys.getenv("BINX_API_KEY", unset = ""),
    secret_key = Sys.getenv("BINX_SECRET_KEY", unset = ""),
    base_url = "https://fapi.binance.com",
    recv_window = 10000L,
    timeout = 30,
    verbose = FALSE) {
  .new_config(
    product = "futures",
    api_key = api_key,
    secret_key = secret_key,
    base_url = base_url,
    recv_window = recv_window,
    timeout = timeout,
    verbose = verbose
  )
}

#' Create a Binance spot configuration
#'
#' @param api_key character API key. Leave `NULL` for unsigned/public requests.
#' @param secret_key character Secret key. Leave `NULL` for unsigned/public requests.
#' @param base_url character Base URL for the Binance Spot API.
#' @param recv_window integer recvWindow in milliseconds used for signed requests.
#' @param timeout numeric request timeout in seconds.
#' @param verbose logical whether to enable verbose request debugging in future helpers.
#'
#' @return A `binxr_config` list.
#' @examples
#' cfg <- config_spot(api_key = NULL, secret_key = NULL)
#' cfg$product
#' @export
config_spot <- function(
    api_key = Sys.getenv("BINX_API_KEY", unset = ""),
    secret_key = Sys.getenv("BINX_SECRET_KEY", unset = ""),
    base_url = "https://api.binance.com",
    recv_window = 10000L,
    timeout = 30,
    verbose = FALSE) {
  .new_config(
    product = "spot",
    api_key = api_key,
    secret_key = secret_key,
    base_url = base_url,
    recv_window = recv_window,
    timeout = timeout,
    verbose = verbose
  )
}

#' Create a Binance options configuration
#'
#' @param api_key character API key. Leave `NULL` for unsigned/public requests.
#' @param secret_key character Secret key. Leave `NULL` for unsigned/public requests.
#' @param base_url character Base URL for the Binance Options API.
#' @param recv_window integer recvWindow in milliseconds used for signed requests.
#' @param timeout numeric request timeout in seconds.
#' @param verbose logical whether to enable verbose request debugging in future helpers.
#'
#' @return A `binxr_config` list.
#' @examples
#' cfg <- config_options(api_key = NULL, secret_key = NULL)
#' cfg$product
#' @export
config_options <- function(
    api_key = Sys.getenv("BINX_API_KEY", unset = ""),
    secret_key = Sys.getenv("BINX_SECRET_KEY", unset = ""),
    base_url = "https://eapi.binance.com",
    recv_window = 10000L,
    timeout = 30,
    verbose = FALSE) {
  .new_config(
    product = "options",
    api_key = api_key,
    secret_key = secret_key,
    base_url = base_url,
    recv_window = recv_window,
    timeout = timeout,
    verbose = verbose
  )
}

#' Backward-compatible futures config constructor
#'
#' @inheritParams config_futures
#' @param base Legacy alias for `base_url`.
#' @param recvWindow Legacy alias for `recv_window`.
#' @return A `binxr_config` list.
#' @export
binxr_config_futures <- function(
    api_key = Sys.getenv("BINX_API_KEY", unset = ""),
    secret_key = Sys.getenv("BINX_SECRET_KEY", unset = ""),
    base = "https://fapi.binance.com",
    recvWindow = 10000L) {
  config_futures(
    api_key = api_key,
    secret_key = secret_key,
    base_url = base,
    recv_window = recvWindow
  )
}

#' Backward-compatible spot config constructor
#'
#' @inheritParams config_spot
#' @param base Legacy alias for `base_url`.
#' @param recvWindow Legacy alias for `recv_window`.
#' @return A `binxr_config` list.
#' @export
binxr_config_spot <- function(
    api_key = Sys.getenv("BINX_API_KEY", unset = ""),
    secret_key = Sys.getenv("BINX_SECRET_KEY", unset = ""),
    base = "https://api.binance.com",
    recvWindow = 10000L) {
  config_spot(
    api_key = api_key,
    secret_key = secret_key,
    base_url = base,
    recv_window = recvWindow
  )
}

#' Backward-compatible options config constructor
#'
#' @inheritParams config_options
#' @param base Legacy alias for `base_url`.
#' @param recvWindow Legacy alias for `recv_window`.
#' @return A `binxr_config` list.
#' @export
binxr_config_options <- function(
    api_key = Sys.getenv("BINX_API_KEY", unset = ""),
    secret_key = Sys.getenv("BINX_SECRET_KEY", unset = ""),
    base = "https://eapi.binance.com",
    recvWindow = 10000L) {
  config_options(
    api_key = api_key,
    secret_key = secret_key,
    base_url = base,
    recv_window = recvWindow
  )
}

#' @noRd
.new_config <- function(
    product,
    api_key,
    secret_key,
    base_url,
    recv_window,
    timeout,
    verbose) {
  api_key <- .normalize_secret(api_key)
  secret_key <- .normalize_secret(secret_key)

  .validate_one_of(product, c("spot", "futures", "options"), "product")
  .validate_scalar_character(base_url, "base_url")
  .validate_positive_number(timeout, "timeout")
  .validate_positive_integerish(recv_window, "recv_window")
  .validate_scalar_logical(verbose, "verbose")

  structure(
    list(
      product = product,
      base_url = base_url,
      base = base_url,
      api_key = api_key,
      secret_key = secret_key,
      recv_window = as.integer(recv_window),
      recvWindow = as.integer(recv_window),
      timeout = as.numeric(timeout),
      verbose = verbose
    ),
    class = "binxr_config"
  )
}

#' @noRd
.normalize_secret <- function(x) {
  if (is.null(x) || identical(x, "")) {
    return(NULL)
  }
  .validate_scalar_character(x, deparse(substitute(x)))
  x
}

test_that("public configs can be created without credentials", {
  cfg <- config_futures(api_key = "", secret_key = "")

  expect_s3_class(cfg, "binxr_config")
  expect_identical(cfg$product, "futures")
  expect_null(cfg$api_key)
  expect_null(cfg$secret_key)
  expect_equal(cfg$base_url, "https://fapi.binance.com")
})

test_that("signed config validation requires both credentials", {
  cfg <- config_spot(api_key = "", secret_key = "")

  expect_error(
    binxr:::.validate_signed_config(cfg),
    "Signed endpoints require both `api_key` and `secret_key`"
  )
})

test_that("legacy config constructors stay compatible", {
  cfg <- binxr_config_futures(api_key = "k", secret_key = "s")

  expect_identical(cfg$product, "futures")
  expect_identical(cfg$api_key, "k")
  expect_identical(cfg$secret_key, "s")
})

test_that("options configs can be created and legacy wrapper stays compatible", {
  cfg <- config_options(api_key = "", secret_key = "")
  legacy_cfg <- binxr_config_options(api_key = "k", secret_key = "s")

  expect_identical(cfg$product, "options")
  expect_equal(cfg$base_url, "https://eapi.binance.com")
  expect_null(cfg$api_key)
  expect_null(cfg$secret_key)

  expect_identical(legacy_cfg$product, "options")
  expect_identical(legacy_cfg$api_key, "k")
  expect_identical(legacy_cfg$secret_key, "s")
})

#' Test Binance Spot API connectivity
#'
#' @param config A spot configuration created by [config_spot()].
#'
#' @return An empty parsed list on success.
#' @export
spot_ping <- function(config = config_spot()) {
  .request_public(config, "/api/v3/ping")
}

#' Get Binance Spot server time
#'
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A `POSIXct` timestamp.
#' @export
spot_get_server_time <- function(config = config_spot()) {
  .ms_to_datetime(.binance_server_time(config))
}

#' Get Binance Spot exchange information
#'
#' @param symbol Optional trading pair symbol.
#' @param symbols Optional character vector of trading pair symbols.
#' @param permissions Optional character vector of permission strings.
#' @param show_permission_sets Logical flag forwarded as `showPermissionSets`.
#' @param symbol_status Optional trading status filter.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_get_exchange_info <- function(
    symbol = NULL,
    symbols = NULL,
    permissions = NULL,
    show_permission_sets = TRUE,
    symbol_status = NULL,
    config = config_spot()) {
  .validate_optional_scalar_character(symbol, "symbol")
  .validate_optional_character_vector(symbols, "symbols")
  .validate_optional_character_vector(permissions, "permissions")
  .validate_scalar_logical(show_permission_sets, "show_permission_sets")
  .validate_optional_scalar_character(symbol_status, "symbol_status")

  if (!is.null(symbol_status) && (!is.null(symbol) || !is.null(symbols))) {
    stop("`symbol_status` cannot be used with `symbol` or `symbols`.", call. = FALSE)
  }

  payload <- .request_public(
    config,
    "/api/v3/exchangeInfo",
    query = list(
      symbol = symbol,
      symbols = .json_array_param(symbols),
      permissions = .json_array_param(permissions),
      showPermissionSets = tolower(as.character(show_permission_sets)),
      symbolStatus = symbol_status
    )
  )

  if (!is.null(payload$symbols) && is.list(payload$symbols) && all(vapply(payload$symbols, is.list, logical(1)))) {
    names(payload$symbols) <- vapply(payload$symbols, `[[`, "", "symbol")
  }
  payload
}

#' Get Binance Spot execution rules
#'
#' @param symbol Optional trading pair symbol.
#' @param symbols Optional character vector of trading pair symbols.
#' @param symbol_status Optional trading status filter.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_get_execution_rules <- function(
    symbol = NULL,
    symbols = NULL,
    symbol_status = NULL,
    config = config_spot()) {
  .validate_optional_scalar_character(symbol, "symbol")
  .validate_optional_character_vector(symbols, "symbols")
  .validate_optional_scalar_character(symbol_status, "symbol_status")
  .validate_at_most_one(symbol, symbols, symbol_status, arg_names = c("symbol", "symbols", "symbol_status"))

  .request_public(
    config,
    "/api/v3/executionRules",
    query = list(
      symbol = symbol,
      symbols = .json_array_param(symbols),
      symbolStatus = symbol_status
    )
  )
}

#' @noRd
.json_array_param <- function(x) {
  if (is.null(x)) {
    return(NULL)
  }
  jsonlite::toJSON(unname(x), auto_unbox = FALSE)
}

test_that("spot exchange info validates symbol_status exclusivity", {
  expect_error(
    spot_get_exchange_info(symbol = "BTCUSDT", symbol_status = "TRADING"),
    "cannot be used with `symbol` or `symbols`"
  )
})

test_that("spot execution rules allows only one selector", {
  expect_error(
    spot_get_execution_rules(symbol = "BTCUSDT", symbol_status = "TRADING"),
    "Only one of symbol, symbols, symbol_status may be supplied"
  )
})

test_that("spot exchange info shapes array parameters as JSON strings", {
  local_mocked_bindings(
    .request_public = function(config, path, query = NULL) {
      query
    },
    .package = "binxr"
  )

  out <- spot_get_exchange_info(
    symbols = c("BTCUSDT", "ETHUSDT"),
    permissions = c("SPOT", "MARGIN"),
    show_permission_sets = FALSE
  )

  expect_identical(as.character(out$symbols), "[\"BTCUSDT\",\"ETHUSDT\"]")
  expect_identical(as.character(out$permissions), "[\"SPOT\",\"MARGIN\"]")
  expect_identical(out$showPermissionSets, "false")
})

test_that("spot ping uses public request helper", {
  local_mocked_bindings(
    .request_public = function(config, path, query = NULL) {
      list(path = path)
    },
    .package = "binxr"
  )

  out <- spot_ping()

  expect_identical(out$path, "/api/v3/ping")
})

#' @noRd
.validate_scalar_character <- function(x, arg) {
  if (!is.character(x) || length(x) != 1L || is.na(x) || !nzchar(x)) {
    stop(sprintf("`%s` must be a non-empty character scalar.", arg), call. = FALSE)
  }
  invisible(x)
}

#' @noRd
.validate_optional_scalar_character <- function(x, arg) {
  if (is.null(x)) {
    return(invisible(x))
  }
  .validate_scalar_character(x, arg)
}

#' @noRd
.validate_scalar_logical <- function(x, arg) {
  if (!is.logical(x) || length(x) != 1L || is.na(x)) {
    stop(sprintf("`%s` must be TRUE or FALSE.", arg), call. = FALSE)
  }
  invisible(x)
}

#' @noRd
.validate_optional_scalar_numeric <- function(x, arg) {
  if (is.null(x)) {
    return(invisible(x))
  }
  if (!is.numeric(x) || length(x) != 1L || is.na(x)) {
    stop(sprintf("`%s` must be a numeric scalar when provided.", arg), call. = FALSE)
  }
  invisible(x)
}

#' @noRd
.validate_positive_number <- function(x, arg) {
  if (!is.numeric(x) || length(x) != 1L || is.na(x) || x <= 0) {
    stop(sprintf("`%s` must be a positive number.", arg), call. = FALSE)
  }
  invisible(x)
}

#' @noRd
.validate_positive_integerish <- function(x, arg) {
  if (!is.numeric(x) || length(x) != 1L || is.na(x) || x <= 0 || x != as.integer(x)) {
    stop(sprintf("`%s` must be a positive whole number.", arg), call. = FALSE)
  }
  invisible(x)
}

#' @noRd
.validate_required_any <- function(..., arg_names) {
  values <- list(...)
  present <- !vapply(values, is.null, logical(1))
  if (!any(present)) {
    stop(
      sprintf("Provide at least one of: %s.", paste(arg_names, collapse = ", ")),
      call. = FALSE
    )
  }
  invisible(TRUE)
}

#' @noRd
.validate_one_of <- function(x, choices, arg) {
  if (!is.character(x) || length(x) != 1L || is.na(x) || !(x %in% choices)) {
    stop(
      sprintf("`%s` must be one of: %s.", arg, paste(choices, collapse = ", ")),
      call. = FALSE
    )
  }
  invisible(x)
}

#' @noRd
.validate_symbol <- function(symbol, arg = "symbol") {
  .validate_scalar_character(symbol, arg)
}

#' @noRd
.validate_optional_character_vector <- function(x, arg) {
  if (is.null(x)) {
    return(invisible(x))
  }
  if (!is.character(x) || !length(x) || anyNA(x) || any(!nzchar(x))) {
    stop(sprintf("`%s` must be a non-empty character vector when provided.", arg), call. = FALSE)
  }
  invisible(x)
}

#' @noRd
.validate_at_most_one <- function(..., arg_names) {
  values <- list(...)
  present <- !vapply(values, is.null, logical(1))
  if (sum(present) > 1L) {
    stop(
      sprintf("Only one of %s may be supplied at a time.", paste(arg_names, collapse = ", ")),
      call. = FALSE
    )
  }
  invisible(TRUE)
}

#' @noRd
.validate_config <- function(config, arg = "config") {
  if (!inherits(config, "binxr_config")) {
    stop(sprintf("`%s` must be created by a binxr config constructor.", arg), call. = FALSE)
  }
  invisible(config)
}

#' @noRd
.validate_public_config <- function(config) {
  .validate_config(config)
  invisible(config)
}

#' @noRd
.validate_signed_config <- function(config) {
  .validate_config(config)
  if (is.null(config$api_key) || is.null(config$secret_key)) {
    stop(
      "Signed endpoints require both `api_key` and `secret_key` in `config`.",
      call. = FALSE
    )
  }
  invisible(config)
}

#' @noRd
.validate_json_list_flag <- function(json_list) {
  .validate_scalar_logical(json_list, "json_list")
}

#' @import data.table
#' @importFrom rlang %||%
NULL

#' Convert milliseconds to POSIXct
#'
#' @param ms numeric Milliseconds since Unix epoch.
#' @return POSIXct Converted timestamp.
#' @noRd
.ms_to_datetime <- function(ms) {
  as.POSIXct(ms / 1000, origin = "1970-01-01", tz = Sys.timezone())
}

#' Round price and quantity to exchange increments
#'
#' @param exchangeInfo Exchange info from [futures_get_exchange_info()].
#' @param symbol Trading pair symbol, for example `"ETHUSDT"`.
#' @param price Optional price to round using the symbol tick size.
#' @param quantity Optional quantity to round using the symbol step size.
#' @param round_dir Direction: `"down"` or `"up"`.
#'
#' @return A list with elements `price` and `quantity`.
#' @examples
#' exchange_info <- list(
#'   symbols = list(
#'     ETHUSDT = list(
#'       filters = list(
#'         list(filterType = "PRICE_FILTER", tickSize = "0.01"),
#'         list(filterType = "LOT_SIZE", stepSize = "0.001")
#'       )
#'     )
#'   )
#' )
#' round_price_qty(exchange_info, "ETHUSDT", price = 1800.123, quantity = 0.12345)
#' @export
round_price_qty <- function(exchangeInfo, symbol, price = NULL, quantity = NULL, round_dir = c("down", "up")) {
  round_dir <- match.arg(round_dir)
  s <- exchangeInfo$symbols[[symbol]]
  if (is.null(s)) stop("symbol not found in exchangeInfo")
  tick <- 1
  step <- 1
  for (f in s$filters) {
    if (f$filterType == "PRICE_FILTER") tick <- as.numeric(f$tickSize)
    if (f$filterType == "LOT_SIZE") step <- as.numeric(f$stepSize)
  }
  round_inc <- function(x, inc) {
    if (is.null(x)) return(NULL)
    if (round_dir == "down") floor((x + 1e-12) / inc) * inc
    else ceiling((x - 1e-12) / inc) * inc
  }
  list(price = round_inc(price, tick), quantity = round_inc(quantity, step))
}

#' @rdname round_price_qty
#' @export
util_round_price_qty <- function(exchangeInfo, symbol, price = NULL, quantity = NULL, round_dir = c("down", "up")) {
  round_price_qty(
    exchangeInfo = exchangeInfo,
    symbol = symbol,
    price = price,
    quantity = quantity,
    round_dir = round_dir
  )
}

#' Get Binance Options margin account information
#'
#' @param config An options configuration created by [config_options()].
#'
#' @return A parsed list.
#' @export
options_get_margin_account <- function(config = config_options()) {
  .request_signed(config, "/eapi/v1/marginAccount", params = list(), method = "GET")
}

#' Get Binance Options funding flow
#'
#' @param currency Asset type. Currently Binance supports `"USDT"`.
#' @param recordId Optional record ID to start from.
#' @param startTime Optional start time in milliseconds since Unix epoch.
#' @param endTime Optional end time in milliseconds since Unix epoch.
#' @param limit Maximum number of rows to return. Must not exceed `1000`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
options_get_funding_flow <- function(
    currency = "USDT",
    recordId = NULL,
    startTime = NULL,
    endTime = NULL,
    limit = 100,
    json_list = FALSE,
    config = config_options()) {
  .validate_scalar_character(currency, "currency")
  .validate_optional_scalar_numeric(recordId, "recordId")
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .options_validate_limit(limit, max = 1000)
  .validate_json_list_flag(json_list)

  payload <- .request_signed(
    config,
    "/eapi/v1/bill",
    params = list(currency = currency, recordId = recordId, startTime = startTime, endTime = endTime, limit = limit),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  bill_dt <- .maybe_as_dt(payload)
  bill_dt <- .coerce_numeric_cols(bill_dt, c("id", "amount"))
  .normalize_time_cols(bill_dt, "createDate")
}

#' @noRd
.binance_server_time <- function(config) {
  .validate_public_config(config)

  path <- switch(
    config$product,
    spot = "/api/v3/time",
    futures = "/fapi/v1/time",
    options = "/eapi/v1/time",
    stop(sprintf("Unsupported product `%s`.", config$product), call. = FALSE)
  )

  .request_public(config, path)$serverTime
}

#' @noRd
.request_public <- function(config, path, query = NULL) {
  .validate_public_config(config)
  .validate_scalar_character(path, "path")

  req <- httr2::request(paste0(config$base_url, path)) |>
    httr2::req_timeout(config$timeout)

  if (!is.null(query)) {
    req <- httr2::req_url_query(req, !!!.compact_query(query))
  }

  .perform_request(req, method = "GET", endpoint = path, product = config$product)
}

#' @noRd
.request_signed <- function(config, path, params = NULL, method = c("GET", "POST", "PUT", "DELETE")) {
  .validate_signed_config(config)
  .validate_scalar_character(path, "path")
  method <- match.arg(method)

  params <- .compact_query(params %||% list())
  if (is.null(params$recvWindow)) {
    params$recvWindow <- config$recv_window
  }
  if (is.null(params$timestamp)) {
    params$timestamp <- .binance_server_time(config)
  }

  signature <- .build_signature(config$secret_key, params)

  req <- httr2::request(paste0(config$base_url, path)) |>
    httr2::req_timeout(config$timeout) |>
    httr2::req_headers("X-MBX-APIKEY" = config$api_key) |>
    httr2::req_method(method) |>
    httr2::req_url_query(!!!params, signature = signature)

  .perform_request(req, method = method, endpoint = path, product = config$product)
}

#' @noRd
.perform_request <- function(req, method, endpoint, product) {
  resp <- tryCatch(httr2::req_perform(req), error = function(e) e)
  .parse_response(resp, endpoint = endpoint, method = method, product = product)
}

#' @noRd
.build_signature <- function(secret, params) {
  payload <- .encode_query(params)
  digest::hmac(key = secret, object = payload, algo = "sha256")
}

#' @noRd
.compact_query <- function(params) {
  if (is.null(params)) {
    return(list())
  }
  if (!is.list(params)) {
    stop("`params`/`query` must be a named list.", call. = FALSE)
  }
  keep <- !vapply(params, is.null, logical(1))
  params[keep]
}

#' @noRd
.encode_query <- function(params) {
  params <- .compact_query(params)
  if (!length(params)) {
    return("")
  }

  names(params) <- names(params) %||% rep.int("", length(params))
  pieces <- Map(
    function(key, value) {
      if (length(value) != 1L) {
        stop(sprintf("Query parameter `%s` must have length 1.", key), call. = FALSE)
      }
      paste0(
        utils::URLencode(key, reserved = TRUE),
        "=",
        utils::URLencode(as.character(value), reserved = TRUE)
      )
    },
    names(params),
    params
  )
  paste(unlist(pieces, use.names = FALSE), collapse = "&")
}

#' @noRd
.parse_response <- function(resp, endpoint = NA_character_, method = NA_character_, product = NA_character_) {
  if (inherits(resp, "httr2_response")) {
    status <- httr2::resp_status(resp)
    body <- tryCatch(httr2::resp_body_string(resp), error = function(e) "")

    if (status >= 200 && status < 300) {
      if (!nzchar(body)) {
        return(list())
      }
      return(httr2::resp_body_json(resp, simplifyVector = FALSE))
    }

    payload <- tryCatch(jsonlite::fromJSON(body, simplifyVector = FALSE), error = function(e) NULL)
    code <- payload$code %||% NA
    msg <- payload$msg %||% body %||% "Unknown Binance API error"

    stop(
      structure(
        list(
          message = sprintf(
            "%s %s failed [%s] code=%s: %s",
            method %||% "HTTP",
            endpoint %||% "<unknown>",
            status,
            code,
            msg
          ),
          status = status,
          code = code,
          api_message = msg,
          endpoint = endpoint,
          method = method,
          product = product
        ),
        class = c("binxr_api_error", "error", "condition")
      )
    )
  }

  if (inherits(resp, "error")) {
    stop(
      structure(
        list(
          message = conditionMessage(resp),
          endpoint = endpoint,
          method = method,
          product = product
        ),
        class = c("binxr_transport_error", "error", "condition")
      )
    )
  }

  stop("Unknown response type.", call. = FALSE)
}

#' @noRd
.binxr_now <- function(config) {
  .binance_server_time(config)
}

#' @noRd
.signed_req <- function(config, endpoint, params, method = c("GET", "POST", "PUT", "DELETE")) {
  .request_signed(config, endpoint, params = params, method = method)
}

#' @noRd
.public_get <- function(config, path, query = NULL) {
  .request_public(config, path, query = query)
}

#' @noRd
.qs_build <- function(kv) {
  .encode_query(kv)
}

#' @noRd
.hmac256 <- function(secret, msg) {
  digest::hmac(key = secret, object = msg, algo = "sha256")
}

test_that("query encoding preserves reserved characters safely", {
  encoded <- binxr:::.encode_query(list(symbol = "BTCUSDT", newClientOrderId = "id with spaces/and?chars"))

  expect_match(encoded, "^symbol=BTCUSDT&newClientOrderId=")
  expect_true(grepl("id%20with%20spaces%2Fand%3Fchars", encoded, fixed = TRUE))
})

test_that("parse_response surfaces structured API errors", {
  resp <- structure(
    list(),
    class = "httr2_response"
  )

  local_mocked_bindings(
    resp_status = function(x) 400L,
    resp_body_string = function(x) '{"code":-1102,"msg":"Mandatory parameter missing"}',
    .package = "httr2"
  )

  expect_error(
    binxr:::.parse_response(resp, endpoint = "/fapi/v1/order", method = "POST", product = "futures"),
    class = "binxr_api_error"
  )
})

#' Test Binance Options connectivity
#'
#' @param config An options configuration created by [config_options()].
#'
#' @return A parsed list.
#' @export
options_ping <- function(config = config_options()) {
  .request_public(config, "/eapi/v1/ping")
}

#' Get Binance Options server time
#'
#' @param config An options configuration created by [config_options()].
#'
#' @return A `POSIXct` timestamp.
#' @export
options_get_server_time <- function(config = config_options()) {
  .ms_to_datetime(.binance_server_time(config))
}

#' Get Binance Options exchange info
#'
#' @param config An options configuration created by [config_options()].
#'
#' @return A parsed list. Option symbol and asset lists are named when present.
#' @export
options_get_exchange_info <- function(config = config_options()) {
  payload <- .request_public(config, "/eapi/v1/exchangeInfo")
  if (!is.null(payload$optionSymbols)) {
    names(payload$optionSymbols) <- vapply(payload$optionSymbols, `[[`, "", "symbol")
  }
  if (!is.null(payload$optionAssets)) {
    names(payload$optionAssets) <- vapply(payload$optionAssets, `[[`, "", "name")
  }
  payload
}

#' Get Binance Options order book
#'
#' @param symbol Option symbol, for example `"BTC-200730-9000-C"`.
#' @param limit Optional depth limit. One of `10`, `20`, `50`, `100`, `500`, or
#'   `1000`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A parsed list.
#' @export
options_get_order_book <- function(symbol, limit = NULL, config = config_options()) {
  .validate_symbol(symbol)
  if (!is.null(limit)) {
    .validate_positive_integerish(limit, "limit")
    if (!(limit %in% c(10L, 20L, 50L, 100L, 500L, 1000L))) {
      stop("`limit` must be one of: 10, 20, 50, 100, 500, 1000.", call. = FALSE)
    }
  }

  .request_public(config, "/eapi/v1/depth", query = list(symbol = symbol, limit = limit))
}

#' Get Binance Options recent trades
#'
#' @param symbol Option symbol, for example `"BTC-200730-9000-C"`.
#' @param limit Maximum number of trades to return. Must not exceed `500`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
options_get_recent_trades <- function(symbol, limit = 100, json_list = FALSE, config = config_options()) {
  .validate_symbol(symbol)
  .options_validate_limit(limit, max = 500)
  .validate_json_list_flag(json_list)

  payload <- .request_public(config, "/eapi/v1/trades", query = list(symbol = symbol, limit = limit))
  if (isTRUE(json_list)) {
    return(payload)
  }

  .options_as_trades_dt(payload)
}

#' Get Binance Options recent block trades
#'
#' @param symbol Optional option symbol.
#' @param limit Maximum number of trades to return. Must not exceed `500`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
options_get_recent_block_trades <- function(symbol = NULL, limit = 100, json_list = FALSE, config = config_options()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  .options_validate_limit(limit, max = 500)
  .validate_json_list_flag(json_list)

  payload <- .request_public(config, "/eapi/v1/blockTrades", query = list(symbol = symbol, limit = limit))
  if (isTRUE(json_list)) {
    return(payload)
  }

  .options_as_trades_dt(payload)
}

#' Get Binance Options klines
#'
#' @param symbol Option symbol, for example `"BTC-200730-9000-C"`.
#' @param interval Kline interval string.
#' @param startTime Optional start time in milliseconds since Unix epoch.
#' @param endTime Optional end time in milliseconds since Unix epoch.
#' @param limit Maximum number of rows to return. Must not exceed `1500`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
options_get_klines <- function(
    symbol,
    interval,
    startTime = NULL,
    endTime = NULL,
    limit = 500,
    json_list = FALSE,
    config = config_options()) {
  .validate_symbol(symbol)
  .validate_scalar_character(interval, "interval")
  .options_validate_limit(limit, max = 1500)
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .validate_json_list_flag(json_list)

  payload <- .request_public(
    config,
    "/eapi/v1/klines",
    query = list(symbol = symbol, interval = interval, startTime = startTime, endTime = endTime, limit = limit)
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  .futures_as_klines_dt(payload)
}

#' Get Binance Options mark prices
#'
#' @param symbol Optional option symbol.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
options_get_mark_price <- function(symbol = NULL, json_list = FALSE, config = config_options()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  .validate_json_list_flag(json_list)

  payload <- .request_public(config, "/eapi/v1/mark", query = list(symbol = symbol))
  if (isTRUE(json_list)) {
    return(payload)
  }

  mark_dt <- .maybe_as_dt(payload)
  .coerce_numeric_cols(
    mark_dt,
    c("markPrice", "bidIV", "askIV", "markIV", "delta", "theta", "gamma", "vega",
      "highPriceLimit", "lowPriceLimit", "riskFreeInterest")
  )
}

#' Get Binance Options underlying index price
#'
#' @param underlying Underlying spot pair, for example `"BTCUSDT"`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A parsed list.
#' @export
options_get_index_price <- function(underlying, config = config_options()) {
  .validate_symbol(underlying, "underlying")
  .request_public(config, "/eapi/v1/index", query = list(underlying = underlying))
}

#' Get Binance Options 24hr ticker statistics
#'
#' @param symbol Optional option symbol.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
options_get_24hr_ticker <- function(symbol = NULL, json_list = FALSE, config = config_options()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  .validate_json_list_flag(json_list)

  payload <- .request_public(config, "/eapi/v1/ticker", query = list(symbol = symbol))
  if (isTRUE(json_list)) {
    return(payload)
  }

  ticker_dt <- .maybe_as_dt(payload)
  ticker_dt <- .coerce_numeric_cols(
    ticker_dt,
    c("priceChange", "priceChangePercent", "lastPrice", "lastQty", "open", "high", "low",
      "volume", "amount", "bidPrice", "askPrice", "firstTradeId", "tradeCount", "strikePrice", "exercisePrice")
  )
  .normalize_time_cols(ticker_dt, c("openTime", "closeTime"))
}

#' Get Binance Options open interest
#'
#' @param underlying_asset Underlying asset, for example `"BTCUSDT"`.
#' @param expiration Expiration date, for example `"221225"`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
options_get_open_interest <- function(
    underlying_asset,
    expiration,
    json_list = FALSE,
    config = config_options()) {
  .validate_symbol(underlying_asset, "underlying_asset")
  .validate_scalar_character(expiration, "expiration")
  .validate_json_list_flag(json_list)

  payload <- .request_public(
    config,
    "/eapi/v1/openInterest",
    query = list(underlyingAsset = underlying_asset, expiration = expiration)
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  oi_dt <- .maybe_as_dt(payload)
  oi_dt <- .coerce_numeric_cols(oi_dt, c("sumOpenInterest", "sumOpenInterestUsd", "timestamp"))
  .normalize_time_cols(oi_dt, "timestamp")
}

#' Get Binance Options historical exercise records
#'
#' @param underlying Optional underlying symbol, for example `"BTCUSDT"`.
#' @param startTime Optional start time in milliseconds since Unix epoch.
#' @param endTime Optional end time in milliseconds since Unix epoch.
#' @param limit Maximum number of rows to return. Must not exceed `100`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config An options configuration created by [config_options()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
options_get_exercise_history <- function(
    underlying = NULL,
    startTime = NULL,
    endTime = NULL,
    limit = 100,
    json_list = FALSE,
    config = config_options()) {
  if (!is.null(underlying)) {
    .validate_symbol(underlying, "underlying")
  }
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .options_validate_limit(limit, max = 100)
  .validate_json_list_flag(json_list)

  payload <- .request_public(
    config,
    "/eapi/v1/exerciseHistory",
    query = list(underlying = underlying, startTime = startTime, endTime = endTime, limit = limit)
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  history_dt <- .maybe_as_dt(payload)
  history_dt <- .coerce_numeric_cols(history_dt, c("strikePrice", "realStrikePrice"))
  .normalize_time_cols(history_dt, "expiryDate")
}

#' @noRd
.options_validate_limit <- function(limit, max, arg = "limit") {
  .validate_positive_integerish(limit, arg)
  if (limit > max) {
    stop(sprintf("`%s` must be less than or equal to %d.", arg, max), call. = FALSE)
  }
  invisible(limit)
}

#' @noRd
.options_as_trades_dt <- function(payload) {
  trades_dt <- .maybe_as_dt(payload)
  trades_dt <- .coerce_numeric_cols(trades_dt, c("id", "tradeId", "price", "qty", "quoteQty", "side"))
  .normalize_time_cols(trades_dt, "time")
}

test_that("options market/account/trade endpoints validate and shape responses", {
  cfg_public <- config_options()
  cfg_signed <- config_options(api_key = "k", secret_key = "s")

  expect_error(
    options_get_order("BTC-200730-9000-C", config = cfg_signed),
    "Provide at least one of: order_id, client_order_id"
  )

  local_mocked_bindings(
    .request_public = function(config, path, query = NULL) {
      switch(
        path,
        "/eapi/v1/trades" = list(list(id = "1", tradeId = "2", price = "100", qty = "1", quoteQty = "100", side = "-1", time = 1000)),
        "/eapi/v1/blockTrades" = list(list(id = "3", tradeId = "4", price = "90", qty = "2", quoteQty = "180", side = "1", time = 2000)),
        "/eapi/v1/klines" = list(list(0, "1", "2", "0.5", "1.5", "10", 1000, "15", 2, "4", "6", "0")),
        "/eapi/v1/mark" = list(list(symbol = "BTC-200730-9000-C", markPrice = "100", delta = "0.1")),
        "/eapi/v1/ticker" = list(list(symbol = "BTC-200730-9000-C", priceChange = "1", openTime = 0, closeTime = 1000)),
        "/eapi/v1/openInterest" = list(list(symbol = "ETH-221119-1175-P", sumOpenInterest = "4.01", sumOpenInterestUsd = "4880.2985", timestamp = "1668754020000")),
        "/eapi/v1/exerciseHistory" = list(list(symbol = "BTC-220121-60000-P", strikePrice = "60000", realStrikePrice = "38844.6965", expiryDate = 1000)),
        list(path = path, query = query)
      )
    },
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "PUT", "DELETE")) {
      switch(
        path,
        "/eapi/v1/order" = list(path = path, params = params, method = method),
        "/eapi/v1/openOrders" = list(list(orderId = "1", price = "100", quantity = "1", executedQty = "0", avgPrice = "0", createTime = 1000, updateTime = 1000)),
        "/eapi/v1/historyOrders" = list(list(orderId = "2", price = "110", quantity = "1", executedQty = "1", avgPrice = "110", createTime = 1000, updateTime = 2000)),
        "/eapi/v1/userTrades" = list(list(id = "5", tradeId = "6", orderId = "2", price = "100", quantity = "1", fee = "0", realizedProfit = "1", time = 3000)),
        "/eapi/v1/position" = list(list(entryPrice = "100", quantity = "-0.1", markValue = "105", unrealizedPNL = "5", markPrice = "1050", strikePrice = "9000", expiryDate = 1000, time = 2000)),
        "/eapi/v1/bill" = list(list(id = "7", amount = "-0.5", createDate = 4000)),
        list(path = path, params = params, method = method)
      )
    },
    .package = "binxr"
  )

  expect_identical(options_ping(cfg_public)$path, "/eapi/v1/ping")

  recent_trades <- options_get_recent_trades("BTC-200730-9000-C", config = cfg_public)
  block_trades <- options_get_recent_block_trades(config = cfg_public)
  klines <- options_get_klines("BTC-200730-9000-C", "1m", config = cfg_public)
  mark <- options_get_mark_price(config = cfg_public)
  ticker <- options_get_24hr_ticker(config = cfg_public)
  open_interest <- options_get_open_interest("BTCUSDT", "221225", config = cfg_public)
  exercise_history <- options_get_exercise_history(config = cfg_public)

  place <- options_place_order("BTC-200730-9000-C", quantity = 1, price = 100, config = cfg_signed)
  open_orders <- options_get_open_orders(config = cfg_signed)
  order_history <- options_get_order_history("BTC-200730-9000-C", config = cfg_signed)
  trades <- options_get_account_trades(config = cfg_signed)
  positions <- options_get_positions(config = cfg_signed)
  funding <- options_get_funding_flow(config = cfg_signed)

  expect_s3_class(recent_trades, "data.table")
  expect_s3_class(block_trades, "data.table")
  expect_s3_class(klines, "data.table")
  expect_s3_class(mark, "data.table")
  expect_s3_class(ticker, "data.table")
  expect_s3_class(open_interest, "data.table")
  expect_s3_class(exercise_history, "data.table")

  expect_identical(place$path, "/eapi/v1/order")
  expect_s3_class(open_orders, "data.table")
  expect_s3_class(order_history, "data.table")
  expect_s3_class(trades, "data.table")
  expect_s3_class(positions, "data.table")
  expect_s3_class(funding, "data.table")
})

#' @noRd
.maybe_as_dt <- function(x, json_list = FALSE) {
  .validate_json_list_flag(json_list)
  if (isTRUE(json_list)) {
    return(x)
  }
  .as_dt_preserve_lists(x)
}

#' @noRd
.as_dt_preserve_lists <- function(x) {
  if (inherits(x, "data.table")) {
    return(data.table::copy(x))
  }

  if (is.data.frame(x)) {
    return(data.table::as.data.table(x))
  }

  if (is.list(x) && length(x) && all(vapply(x, is.list, logical(1)))) {
    return(data.table::rbindlist(x, fill = TRUE))
  }

  if (is.list(x)) {
    return(data.table::as.data.table(list(value = list(x))))
  }

  data.table::data.table(value = x)
}

#' @noRd
.normalize_time_cols <- function(dt, cols, tz = Sys.timezone()) {
  if (!inherits(dt, "data.table")) {
    dt <- data.table::as.data.table(dt)
  }

  for (col in intersect(cols, names(dt))) {
    data.table::set(
      dt,
      j = col,
      value = as.POSIXct(as.numeric(dt[[col]]) / 1000, origin = "1970-01-01", tz = tz)
    )
  }
  dt
}

#' @noRd
.coerce_numeric_cols <- function(dt, cols) {
  if (!inherits(dt, "data.table")) {
    dt <- data.table::as.data.table(dt)
  }

  for (col in intersect(cols, names(dt))) {
    data.table::set(dt, j = col, value = as.numeric(dt[[col]]))
  }
  dt
}

#' @noRd
.normalize_order_time_cols <- function(dt, tz = Sys.timezone()) {
  .normalize_time_cols(dt, c("time", "updateTime", "workingTime", "transactTime"), tz = tz)
}

test_that("maybe_as_dt returns data.table by default", {
  x <- list(
    list(symbol = "BTCUSDT", filters = list(list(filterType = "PRICE_FILTER"))),
    list(symbol = "ETHUSDT", filters = list(list(filterType = "LOT_SIZE")))
  )

  dt <- binxr:::.maybe_as_dt(x)

  expect_s3_class(dt, "data.table")
  expect_true(is.list(dt$filters))
  expect_equal(nrow(dt), 2L)
})

test_that("json_list flag preserves original parsed list", {
  x <- list(list(symbol = "BTCUSDT"))

  out <- binxr:::.maybe_as_dt(x, json_list = TRUE)

  expect_identical(out, x)
})

#' Get Binance Spot account information
#'
#' @param omit_zero_balances Whether to omit zero balances from the response.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_get_account <- function(omit_zero_balances = FALSE, config = config_spot()) {
  .validate_scalar_logical(omit_zero_balances, "omit_zero_balances")

  .request_signed(
    config,
    "/api/v3/account",
    params = list(omitZeroBalances = if (isTRUE(omit_zero_balances)) "true" else "false"),
    method = "GET"
  )
}

#' Get a Binance Spot order
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param order_id Optional exchange order ID.
#' @param orig_client_order_id Optional client order ID.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_get_order <- function(
    symbol,
    order_id = NULL,
    orig_client_order_id = NULL,
    config = config_spot()) {
  .validate_symbol(symbol)
  .validate_required_any(order_id, orig_client_order_id, arg_names = c("order_id", "orig_client_order_id"))
  .validate_optional_scalar_numeric(order_id, "order_id")
  .validate_optional_scalar_character(orig_client_order_id, "orig_client_order_id")

  .request_signed(
    config,
    "/api/v3/order",
    params = list(
      symbol = symbol,
      orderId = order_id,
      origClientOrderId = orig_client_order_id
    ),
    method = "GET"
  )
}

#' Get Binance Spot open orders
#'
#' @param symbol Optional trading pair symbol. If `NULL`, returns all open
#'   orders.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
spot_get_open_orders <- function(symbol = NULL, json_list = FALSE, config = config_spot()) {
  if (!is.null(symbol)) {
    .validate_symbol(symbol)
  }
  .validate_json_list_flag(json_list)

  payload <- .request_signed(
    config,
    "/api/v3/openOrders",
    params = list(symbol = symbol),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  .spot_as_orders_dt(payload)
}

#' Get Binance Spot order history
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param order_id Optional exchange order ID to start from.
#' @param start_time Optional start time in milliseconds.
#' @param end_time Optional end time in milliseconds.
#' @param limit Maximum number of orders to return. Must not exceed `1000`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
spot_get_orders <- function(
    symbol,
    order_id = NULL,
    start_time = NULL,
    end_time = NULL,
    limit = 500,
    json_list = FALSE,
    config = config_spot()) {
  .validate_symbol(symbol)
  .validate_optional_scalar_numeric(order_id, "order_id")
  .validate_optional_scalar_numeric(start_time, "start_time")
  .validate_optional_scalar_numeric(end_time, "end_time")
  .spot_validate_limit(limit)
  .validate_json_list_flag(json_list)
  .spot_validate_time_range_24h(start_time, end_time)

  payload <- .request_signed(
    config,
    "/api/v3/allOrders",
    params = list(
      symbol = symbol,
      orderId = order_id,
      startTime = start_time,
      endTime = end_time,
      limit = limit
    ),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  .spot_as_orders_dt(payload)
}

#' Get a Binance Spot order list
#'
#' @param order_list_id Optional exchange order-list ID.
#' @param orig_client_order_id Optional order-list client ID.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_get_order_list <- function(
    order_list_id = NULL,
    orig_client_order_id = NULL,
    config = config_spot()) {
  .validate_required_any(order_list_id, orig_client_order_id, arg_names = c("order_list_id", "orig_client_order_id"))
  .validate_optional_scalar_numeric(order_list_id, "order_list_id")
  .validate_optional_scalar_character(orig_client_order_id, "orig_client_order_id")

  .request_signed(
    config,
    "/api/v3/orderList",
    params = list(
      orderListId = order_list_id,
      origClientOrderId = orig_client_order_id
    ),
    method = "GET"
  )
}

#' Get all Binance Spot order lists
#'
#' @param from_id Optional order-list ID to start from.
#' @param start_time Optional start time in milliseconds.
#' @param end_time Optional end time in milliseconds.
#' @param limit Maximum number of order lists to return. Must not exceed
#'   `1000`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
spot_get_all_order_lists <- function(
    from_id = NULL,
    start_time = NULL,
    end_time = NULL,
    limit = 500,
    json_list = FALSE,
    config = config_spot()) {
  .validate_optional_scalar_numeric(from_id, "from_id")
  .validate_optional_scalar_numeric(start_time, "start_time")
  .validate_optional_scalar_numeric(end_time, "end_time")
  .spot_validate_limit(limit)
  .validate_json_list_flag(json_list)

  if (!is.null(from_id) && (!is.null(start_time) || !is.null(end_time))) {
    stop("`from_id` must not be supplied with `start_time` or `end_time`.", call. = FALSE)
  }
  .spot_validate_time_range_24h(start_time, end_time)

  payload <- .request_signed(
    config,
    "/api/v3/allOrderList",
    params = list(
      fromId = from_id,
      startTime = start_time,
      endTime = end_time,
      limit = limit
    ),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  .spot_as_order_lists_dt(payload)
}

#' Get Binance Spot open order lists
#'
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
spot_get_open_order_lists <- function(json_list = FALSE, config = config_spot()) {
  .validate_json_list_flag(json_list)

  payload <- .request_signed(
    config,
    "/api/v3/openOrderList",
    params = list(),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  .spot_as_order_lists_dt(payload)
}

#' Get Binance Spot account trades
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param order_id Optional exchange order ID.
#' @param start_time Optional start time in milliseconds.
#' @param end_time Optional end time in milliseconds.
#' @param from_id Optional trade ID to start from.
#' @param limit Maximum number of trades to return. Must not exceed `1000`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
spot_get_account_trades <- function(
    symbol,
    order_id = NULL,
    start_time = NULL,
    end_time = NULL,
    from_id = NULL,
    limit = 500,
    json_list = FALSE,
    config = config_spot()) {
  .validate_symbol(symbol)
  .validate_optional_scalar_numeric(order_id, "order_id")
  .validate_optional_scalar_numeric(start_time, "start_time")
  .validate_optional_scalar_numeric(end_time, "end_time")
  .validate_optional_scalar_numeric(from_id, "from_id")
  .spot_validate_limit(limit)
  .validate_json_list_flag(json_list)

  if (!is.null(order_id) && (!is.null(start_time) || !is.null(end_time))) {
    stop("`order_id` must not be supplied with `start_time` or `end_time`.", call. = FALSE)
  }
  if (!is.null(from_id) && (!is.null(start_time) || !is.null(end_time))) {
    stop("`from_id` must not be supplied with `start_time` or `end_time`.", call. = FALSE)
  }
  .spot_validate_time_range_24h(start_time, end_time)

  payload <- .request_signed(
    config,
    "/api/v3/myTrades",
    params = list(
      symbol = symbol,
      orderId = order_id,
      startTime = start_time,
      endTime = end_time,
      fromId = from_id,
      limit = limit
    ),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  .spot_as_trades_dt(payload)
}

#' Get Binance Spot unfilled order counts
#'
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
spot_get_unfilled_order_count <- function(json_list = FALSE, config = config_spot()) {
  .validate_json_list_flag(json_list)

  payload <- .request_signed(
    config,
    "/api/v3/rateLimit/order",
    params = list(),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  counts_dt <- .maybe_as_dt(payload)
  .coerce_numeric_cols(counts_dt, c("intervalNum", "limit", "count"))
}

#' Get Binance Spot prevented matches
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param prevented_match_id Optional prevented-match ID.
#' @param order_id Optional exchange order ID.
#' @param from_prevented_match_id Optional prevented-match ID to start from.
#' @param limit Maximum number of records to return. Must not exceed `1000`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
spot_get_prevented_matches <- function(
    symbol,
    prevented_match_id = NULL,
    order_id = NULL,
    from_prevented_match_id = NULL,
    limit = 500,
    json_list = FALSE,
    config = config_spot()) {
  .validate_symbol(symbol)
  .validate_optional_scalar_numeric(prevented_match_id, "prevented_match_id")
  .validate_optional_scalar_numeric(order_id, "order_id")
  .validate_optional_scalar_numeric(from_prevented_match_id, "from_prevented_match_id")
  .spot_validate_limit(limit)
  .validate_json_list_flag(json_list)

  .validate_required_any(prevented_match_id, order_id, arg_names = c("prevented_match_id", "order_id"))
  .validate_at_most_one(prevented_match_id, order_id, arg_names = c("prevented_match_id", "order_id"))
  if (!is.null(from_prevented_match_id) && is.null(order_id)) {
    stop("`from_prevented_match_id` requires `order_id`.", call. = FALSE)
  }
  if (!is.null(from_prevented_match_id) && !is.null(prevented_match_id)) {
    stop("`from_prevented_match_id` must not be supplied with `prevented_match_id`.", call. = FALSE)
  }

  payload <- .request_signed(
    config,
    "/api/v3/myPreventedMatches",
    params = list(
      symbol = symbol,
      preventedMatchId = prevented_match_id,
      orderId = order_id,
      fromPreventedMatchId = from_prevented_match_id,
      limit = limit
    ),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  matches_dt <- .maybe_as_dt(payload)
  matches_dt <- .coerce_numeric_cols(
    matches_dt,
    c("preventedMatchId", "takerOrderId", "makerOrderId", "tradeGroupId", "price", "makerPreventedQuantity")
  )
  .normalize_time_cols(matches_dt, "transactTime")
}

#' Get Binance Spot allocations
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param start_time Optional start time in milliseconds.
#' @param end_time Optional end time in milliseconds.
#' @param from_allocation_id Optional allocation ID to start from.
#' @param limit Maximum number of records to return. Must not exceed `1000`.
#' @param order_id Optional exchange order ID.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
spot_get_allocations <- function(
    symbol,
    start_time = NULL,
    end_time = NULL,
    from_allocation_id = NULL,
    limit = 500,
    order_id = NULL,
    json_list = FALSE,
    config = config_spot()) {
  .validate_symbol(symbol)
  .validate_optional_scalar_numeric(start_time, "start_time")
  .validate_optional_scalar_numeric(end_time, "end_time")
  .validate_optional_scalar_numeric(from_allocation_id, "from_allocation_id")
  .spot_validate_limit(limit)
  .validate_optional_scalar_numeric(order_id, "order_id")
  .validate_json_list_flag(json_list)

  if (!is.null(from_allocation_id) && (!is.null(start_time) || !is.null(end_time))) {
    stop("`from_allocation_id` must not be supplied with `start_time` or `end_time`.", call. = FALSE)
  }
  if (!is.null(order_id) && (!is.null(start_time) || !is.null(end_time))) {
    stop("`order_id` must not be supplied with `start_time` or `end_time`.", call. = FALSE)
  }
  .spot_validate_time_range_24h(start_time, end_time)

  payload <- .request_signed(
    config,
    "/api/v3/myAllocations",
    params = list(
      symbol = symbol,
      startTime = start_time,
      endTime = end_time,
      fromAllocationId = from_allocation_id,
      limit = limit,
      orderId = order_id
    ),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  allocations_dt <- .maybe_as_dt(payload)
  allocations_dt <- .coerce_numeric_cols(
    allocations_dt,
    c("allocationId", "orderId", "orderListId", "price", "qty", "quoteQty", "commission")
  )
  .normalize_time_cols(allocations_dt, "time")
}

#' Get Binance Spot commission rates
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_get_commission_rates <- function(symbol, config = config_spot()) {
  .validate_symbol(symbol)

  .request_signed(
    config,
    "/api/v3/account/commission",
    params = list(symbol = symbol),
    method = "GET"
  )
}

#' Get Binance Spot order amendments
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param order_id Exchange order ID.
#' @param from_execution_id Optional execution ID to start from.
#' @param limit Maximum number of records to return. Must not exceed `1000`.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
spot_get_order_amendments <- function(
    symbol,
    order_id,
    from_execution_id = NULL,
    limit = 500,
    json_list = FALSE,
    config = config_spot()) {
  .validate_symbol(symbol)
  .validate_optional_scalar_numeric(from_execution_id, "from_execution_id")
  .spot_validate_limit(limit)
  .validate_json_list_flag(json_list)
  .validate_optional_scalar_numeric(order_id, "order_id")
  if (is.null(order_id)) {
    stop("`order_id` must be supplied.", call. = FALSE)
  }

  payload <- .request_signed(
    config,
    "/api/v3/order/amendments",
    params = list(
      symbol = symbol,
      orderId = order_id,
      fromExecutionId = from_execution_id,
      limit = limit
    ),
    method = "GET"
  )
  if (isTRUE(json_list)) {
    return(payload)
  }

  amendments_dt <- .maybe_as_dt(payload)
  amendments_dt <- .coerce_numeric_cols(
    amendments_dt,
    c("orderId", "executionId", "origQty", "newQty")
  )
  .normalize_time_cols(amendments_dt, "time")
}

#' Get Binance Spot relevant account filters
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_get_relevant_filters <- function(symbol, config = config_spot()) {
  .validate_symbol(symbol)

  .request_signed(
    config,
    "/api/v3/myFilters",
    params = list(symbol = symbol),
    method = "GET"
  )
}

#' @noRd
.spot_validate_limit <- function(limit, max = 1000L, arg = "limit") {
  .validate_positive_integerish(limit, arg)
  if (limit > max) {
    stop(sprintf("`%s` must be less than or equal to %d.", arg, max), call. = FALSE)
  }
  invisible(limit)
}

#' @noRd
.spot_validate_time_range_24h <- function(start_time, end_time) {
  if (is.null(start_time) || is.null(end_time)) {
    return(invisible(TRUE))
  }
  if (end_time < start_time) {
    stop("`end_time` must be greater than or equal to `start_time`.", call. = FALSE)
  }
  if ((end_time - start_time) > 24 * 60 * 60 * 1000) {
    stop("The time between `start_time` and `end_time` must not exceed 24 hours.", call. = FALSE)
  }
  invisible(TRUE)
}

#' @noRd
.spot_as_orders_dt <- function(payload) {
  orders_dt <- .maybe_as_dt(payload)
  orders_dt <- .coerce_numeric_cols(
    orders_dt,
    c(
      "orderId", "orderListId", "price", "origQty", "executedQty",
      "cummulativeQuoteQty", "stopPrice", "icebergQty", "origQuoteOrderQty"
    )
  )
  .normalize_order_time_cols(orders_dt)
}

#' @noRd
.spot_as_order_lists_dt <- function(payload) {
  lists_dt <- .maybe_as_dt(payload)
  lists_dt <- .coerce_numeric_cols(lists_dt, c("orderListId"))
  .normalize_time_cols(lists_dt, "transactionTime")
}

#' @noRd
.spot_as_trades_dt <- function(payload) {
  trades_dt <- .maybe_as_dt(payload)
  trades_dt <- .coerce_numeric_cols(
    trades_dt,
    c("id", "orderId", "orderListId", "price", "qty", "quoteQty", "commission")
  )
  .normalize_time_cols(trades_dt, "time")
}

test_that("spot account endpoints validate selectors and supported parameter combinations", {
  cfg <- config_spot(api_key = "k", secret_key = "s")

  expect_error(
    spot_get_order(symbol = "BTCUSDT", config = cfg),
    "Provide at least one of: order_id, orig_client_order_id"
  )

  expect_error(
    spot_get_all_order_lists(
      from_id = 1,
      start_time = 0,
      config = cfg
    ),
    "from_id"
  )

  expect_error(
    spot_get_account_trades(
      symbol = "BTCUSDT",
      order_id = 10,
      start_time = 0,
      config = cfg
    ),
    "order_id"
  )

  expect_error(
    spot_get_prevented_matches(
      symbol = "BTCUSDT",
      from_prevented_match_id = 2,
      config = cfg
    ),
    "Provide at least one of: prevented_match_id, order_id"
  )

  expect_error(
    spot_get_allocations(
      symbol = "BTCUSDT",
      start_time = 0,
      end_time = 24 * 60 * 60 * 1000 + 1,
      config = cfg
    ),
    "24 hours"
  )
})

test_that("spot account endpoints call signed paths and shape tabular responses", {
  cfg <- config_spot(api_key = "k", secret_key = "s")

  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "PUT", "DELETE")) {
      if (identical(path, "/api/v3/account")) {
        return(list(path = path, params = params, method = method))
      }
      if (identical(path, "/api/v3/order")) {
        return(list(path = path, params = params, method = method))
      }
      if (identical(path, "/api/v3/openOrders")) {
        return(list(list(orderId = "1", time = 1000, updateTime = 2000, price = "100.5")))
      }
      if (identical(path, "/api/v3/myTrades")) {
        return(list(list(id = "2", orderId = "9", price = "101", qty = "0.5", time = 3000)))
      }
      if (identical(path, "/api/v3/rateLimit/order")) {
        return(list(list(intervalNum = 10, limit = 50, count = 1)))
      }
      if (identical(path, "/api/v3/order/amendments")) {
        return(list(list(orderId = "9", executionId = "22", origQty = "5", newQty = "4", time = 4000)))
      }
      list(path = path, params = params, method = method)
    },
    .package = "binxr"
  )

  account_out <- spot_get_account(omit_zero_balances = TRUE, config = cfg)
  order_out <- spot_get_order(symbol = "BTCUSDT", order_id = 10, config = cfg)
  open_orders_out <- spot_get_open_orders(symbol = "BTCUSDT", config = cfg)
  trades_out <- spot_get_account_trades(symbol = "BTCUSDT", config = cfg)
  counts_out <- spot_get_unfilled_order_count(config = cfg)
  amendments_out <- spot_get_order_amendments(symbol = "BTCUSDT", order_id = 9, config = cfg)

  expect_identical(account_out$path, "/api/v3/account")
  expect_identical(account_out$params$omitZeroBalances, "true")
  expect_identical(order_out$path, "/api/v3/order")
  expect_identical(order_out$params$orderId, 10)

  expect_s3_class(open_orders_out, "data.table")
  expect_equal(open_orders_out$orderId, 1)
  expect_s3_class(open_orders_out$time, "POSIXct")

  expect_s3_class(trades_out, "data.table")
  expect_equal(trades_out$id, 2)
  expect_s3_class(trades_out$time, "POSIXct")

  expect_s3_class(counts_out, "data.table")
  expect_equal(counts_out$count, 1)

  expect_s3_class(amendments_out, "data.table")
  expect_equal(amendments_out$executionId, 22)
  expect_s3_class(amendments_out$time, "POSIXct")
})

test_that("spot account json_list flags return raw payloads", {
  cfg <- config_spot(api_key = "k", secret_key = "s")
  raw_payload <- list(list(orderId = "1"))

  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "PUT", "DELETE")) {
      raw_payload
    },
    .package = "binxr"
  )

  expect_identical(spot_get_open_orders(json_list = TRUE, config = cfg), raw_payload)
  expect_identical(spot_get_open_order_lists(json_list = TRUE, config = cfg), raw_payload)
  expect_identical(spot_get_unfilled_order_count(json_list = TRUE, config = cfg), raw_payload)
})

#' Get Binance Spot order book
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param limit Optional order book depth limit.
#' @param symbol_status Optional trading status filter.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_get_order_book <- function(symbol, limit = NULL, symbol_status = NULL, config = config_spot()) {
  .validate_symbol(symbol)
  .validate_optional_scalar_numeric(limit, "limit")
  .validate_optional_scalar_character(symbol_status, "symbol_status")

  .request_public(
    config,
    "/api/v3/depth",
    query = list(
      symbol = symbol,
      limit = limit,
      symbolStatus = symbol_status
    )
  )
}

#' Get Binance Spot recent trades
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param limit Maximum number of rows to return.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
spot_get_recent_trades <- function(symbol, limit = 500, json_list = FALSE, config = config_spot()) {
  .validate_symbol(symbol)
  .validate_positive_integerish(limit, "limit")

  payload <- .request_public(
    config,
    "/api/v3/trades",
    query = list(
      symbol = symbol,
      limit = limit
    )
  )

  .spot_market_maybe_tabular(payload, json_list = json_list)
}

#' Get Binance Spot historical trades
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param limit Maximum number of rows to return.
#' @param fromId Optional trade identifier to fetch from.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
spot_get_historical_trades <- function(
    symbol,
    limit = 500,
    fromId = NULL,
    json_list = FALSE,
    config = config_spot()) {
  .validate_symbol(symbol)
  .validate_positive_integerish(limit, "limit")
  .validate_optional_scalar_numeric(fromId, "fromId")

  payload <- .request_public(
    config,
    "/api/v3/historicalTrades",
    query = list(
      symbol = symbol,
      limit = limit,
      fromId = fromId
    )
  )

  .spot_market_maybe_tabular(payload, json_list = json_list)
}

#' Get Binance Spot aggregate trades
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param fromId Optional aggregate trade identifier to fetch from, inclusive.
#' @param startTime Optional start time in milliseconds since Unix epoch, inclusive.
#' @param endTime Optional end time in milliseconds since Unix epoch, inclusive.
#' @param limit Maximum number of rows to return.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
spot_get_aggregate_trades <- function(
    symbol,
    fromId = NULL,
    startTime = NULL,
    endTime = NULL,
    limit = 500,
    json_list = FALSE,
    config = config_spot()) {
  .validate_symbol(symbol)
  .validate_optional_scalar_numeric(fromId, "fromId")
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .validate_positive_integerish(limit, "limit")

  payload <- .request_public(
    config,
    "/api/v3/aggTrades",
    query = list(
      symbol = symbol,
      fromId = fromId,
      startTime = startTime,
      endTime = endTime,
      limit = limit
    )
  )

  .spot_market_maybe_tabular(payload, json_list = json_list)
}

#' Get Binance Spot klines
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param interval Kline interval string.
#' @param startTime Optional start time in milliseconds since Unix epoch.
#' @param endTime Optional end time in milliseconds since Unix epoch.
#' @param timeZone Optional timezone string accepted by Binance.
#' @param limit Maximum number of rows to return.
#' @param json_list If `TRUE`, return the parsed list instead of a `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
spot_get_klines <- function(
    symbol,
    interval,
    startTime = NULL,
    endTime = NULL,
    timeZone = NULL,
    limit = 500,
    json_list = FALSE,
    config = config_spot()) {
  .validate_symbol(symbol)
  .validate_scalar_character(interval, "interval")
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .validate_optional_scalar_character(timeZone, "timeZone")
  .validate_positive_integerish(limit, "limit")

  payload <- .request_public(
    config,
    "/api/v3/klines",
    query = list(
      symbol = symbol,
      interval = interval,
      startTime = startTime,
      endTime = endTime,
      timeZone = timeZone,
      limit = limit
    )
  )

  if (isTRUE(json_list)) {
    return(payload)
  }

  .spot_klines_to_dt(payload)
}

#' Get Binance Spot UI klines
#'
#' @inheritParams spot_get_klines
#'
#' @return A `data.table` by default, or a parsed list when `json_list = TRUE`.
#' @export
spot_get_ui_klines <- function(
    symbol,
    interval,
    startTime = NULL,
    endTime = NULL,
    timeZone = NULL,
    limit = 500,
    json_list = FALSE,
    config = config_spot()) {
  .validate_symbol(symbol)
  .validate_scalar_character(interval, "interval")
  .validate_optional_scalar_numeric(startTime, "startTime")
  .validate_optional_scalar_numeric(endTime, "endTime")
  .validate_optional_scalar_character(timeZone, "timeZone")
  .validate_positive_integerish(limit, "limit")

  payload <- .request_public(
    config,
    "/api/v3/uiKlines",
    query = list(
      symbol = symbol,
      interval = interval,
      startTime = startTime,
      endTime = endTime,
      timeZone = timeZone,
      limit = limit
    )
  )

  if (isTRUE(json_list)) {
    return(payload)
  }

  .spot_klines_to_dt(payload)
}

#' Get Binance Spot average price
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_get_average_price <- function(symbol, config = config_spot()) {
  .validate_symbol(symbol)
  .request_public(config, "/api/v3/avgPrice", query = list(symbol = symbol))
}

#' Get Binance Spot 24-hour ticker statistics
#'
#' @param symbol Optional trading pair symbol.
#' @param symbols Optional character vector of trading pair symbols.
#' @param type One of `"FULL"` or `"MINI"`.
#' @param symbol_status Optional trading status filter.
#' @param json_list If `TRUE`, return the parsed list instead of converting
#'   array responses to `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list for single-symbol requests, or a `data.table` for
#'   multi-symbol/all-symbol requests unless `json_list = TRUE`.
#' @export
spot_get_24hr_ticker <- function(
    symbol = NULL,
    symbols = NULL,
    type = c("FULL", "MINI"),
    symbol_status = NULL,
    json_list = FALSE,
    config = config_spot()) {
  type <- match.arg(type)
  query <- .spot_symbol_selector_query(
    symbol = symbol,
    symbols = symbols,
    symbol_status = symbol_status
  )
  payload <- .request_public(
    config,
    "/api/v3/ticker/24hr",
    query = c(query, list(type = type))
  )
  .spot_market_maybe_tabular(payload, json_list = json_list)
}

#' Get Binance Spot trading day ticker statistics
#'
#' @param symbol Optional trading pair symbol.
#' @param symbols Optional character vector of trading pair symbols.
#' @param timeZone Optional timezone string accepted by Binance.
#' @param type One of `"FULL"` or `"MINI"`.
#' @param symbol_status Optional trading status filter.
#' @param json_list If `TRUE`, return the parsed list instead of converting
#'   array responses to `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list for single-symbol requests, or a `data.table` for
#'   multi-symbol requests unless `json_list = TRUE`.
#' @export
spot_get_trading_day_ticker <- function(
    symbol = NULL,
    symbols = NULL,
    timeZone = NULL,
    type = c("FULL", "MINI"),
    symbol_status = NULL,
    json_list = FALSE,
    config = config_spot()) {
  type <- match.arg(type)
  .validate_optional_scalar_character(timeZone, "timeZone")
  query <- .spot_symbol_selector_query(
    symbol = symbol,
    symbols = symbols,
    symbol_status = symbol_status,
    require_any = TRUE
  )
  payload <- .request_public(
    config,
    "/api/v3/ticker/tradingDay",
    query = c(
      query,
      list(
        timeZone = timeZone,
        type = type
      )
    )
  )
  .spot_market_maybe_tabular(payload, json_list = json_list)
}

#' Get Binance Spot rolling window ticker statistics
#'
#' @param symbol Optional trading pair symbol.
#' @param symbols Optional character vector of trading pair symbols.
#' @param windowSize Optional rolling window size such as `"1m"`, `"4h"`, or `"7d"`.
#' @param type One of `"FULL"` or `"MINI"`.
#' @param symbol_status Optional trading status filter.
#' @param json_list If `TRUE`, return the parsed list instead of converting
#'   array responses to `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list for single-symbol requests, or a `data.table` for
#'   multi-symbol requests unless `json_list = TRUE`.
#' @export
spot_get_rolling_window_ticker <- function(
    symbol = NULL,
    symbols = NULL,
    windowSize = NULL,
    type = c("FULL", "MINI"),
    symbol_status = NULL,
    json_list = FALSE,
    config = config_spot()) {
  type <- match.arg(type)
  .validate_optional_window_size(windowSize, "windowSize")
  query <- .spot_symbol_selector_query(
    symbol = symbol,
    symbols = symbols,
    symbol_status = symbol_status,
    require_any = TRUE
  )
  payload <- .request_public(
    config,
    "/api/v3/ticker",
    query = c(
      query,
      list(
        windowSize = windowSize,
        type = type
      )
    )
  )
  .spot_market_maybe_tabular(payload, json_list = json_list)
}

#' Get Binance Spot ticker price
#'
#' @param symbol Optional trading pair symbol.
#' @param symbols Optional character vector of trading pair symbols.
#' @param symbol_status Optional trading status filter.
#' @param json_list If `TRUE`, return the parsed list instead of converting
#'   array responses to `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list for single-symbol requests, or a `data.table` for
#'   multi-symbol/all-symbol requests unless `json_list = TRUE`.
#' @export
spot_get_ticker_price <- function(
    symbol = NULL,
    symbols = NULL,
    symbol_status = NULL,
    json_list = FALSE,
    config = config_spot()) {
  query <- .spot_symbol_selector_query(
    symbol = symbol,
    symbols = symbols,
    symbol_status = symbol_status
  )
  payload <- .request_public(config, "/api/v3/ticker/price", query = query)
  .spot_market_maybe_tabular(payload, json_list = json_list)
}

#' @rdname spot_get_ticker_price
#' @export
get_spot_mark_price <- function(symbol = NULL, config = binxr_config_spot()) {
  .binxr_deprecate_alias("get_spot_mark_price", "spot_get_ticker_price")
  spot_get_ticker_price(symbol = symbol, config = config)
}

#' Get Binance Spot book ticker
#'
#' @param symbol Optional trading pair symbol.
#' @param symbols Optional character vector of trading pair symbols.
#' @param symbol_status Optional trading status filter.
#' @param json_list If `TRUE`, return the parsed list instead of converting
#'   array responses to `data.table`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list for single-symbol requests, or a `data.table` for
#'   multi-symbol/all-symbol requests unless `json_list = TRUE`.
#' @export
spot_get_book_ticker <- function(
    symbol = NULL,
    symbols = NULL,
    symbol_status = NULL,
    json_list = FALSE,
    config = config_spot()) {
  query <- .spot_symbol_selector_query(
    symbol = symbol,
    symbols = symbols,
    symbol_status = symbol_status
  )
  payload <- .request_public(config, "/api/v3/ticker/bookTicker", query = query)
  .spot_market_maybe_tabular(payload, json_list = json_list)
}

#' Get Binance Spot reference price
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_get_reference_price <- function(symbol, config = config_spot()) {
  .validate_symbol(symbol)
  .request_public(config, "/api/v3/referencePrice", query = list(symbol = symbol))
}

#' Get Binance Spot reference price calculation metadata
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param symbol_status Optional trading status filter.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_get_reference_price_calculation <- function(
    symbol,
    symbol_status = NULL,
    config = config_spot()) {
  .validate_symbol(symbol)
  .validate_optional_scalar_character(symbol_status, "symbol_status")

  .request_public(
    config,
    "/api/v3/referencePrice/calculation",
    query = list(
      symbol = symbol,
      symbolStatus = symbol_status
    )
  )
}

#' @noRd
.spot_symbol_selector_query <- function(
    symbol = NULL,
    symbols = NULL,
    symbol_status = NULL,
    require_any = FALSE) {
  .validate_optional_scalar_character(symbol, "symbol")
  .validate_optional_character_vector(symbols, "symbols")
  .validate_optional_scalar_character(symbol_status, "symbol_status")
  .validate_at_most_one(symbol, symbols, arg_names = c("symbol", "symbols"))
  if (isTRUE(require_any)) {
    .validate_required_any(symbol, symbols, arg_names = c("symbol", "symbols"))
  }

  list(
    symbol = symbol,
    symbols = .json_array_param(symbols),
    symbolStatus = symbol_status
  )
}

#' @noRd
.validate_optional_window_size <- function(x, arg) {
  if (is.null(x)) {
    return(invisible(x))
  }
  .validate_scalar_character(x, arg)
  valid_pattern <- "^(?:([1-9]|[1-5][0-9])m|([1-9]|1[0-9]|2[0-3])h|([1-7])d)$"
  if (!grepl(valid_pattern, x)) {
    stop(
      sprintf("`%s` must be one of 1m-59m, 1h-23h, or 1d-7d.", arg),
      call. = FALSE
    )
  }
  invisible(x)
}

#' @noRd
.spot_market_maybe_tabular <- function(payload, json_list = FALSE) {
  if (isTRUE(json_list)) {
    return(payload)
  }
  if (is.list(payload) && length(payload) && all(vapply(payload, is.list, logical(1)))) {
    return(.maybe_as_dt(payload))
  }
  payload
}

#' @noRd
.spot_klines_to_dt <- function(payload) {
  kline_dt <- data.table::rbindlist(payload)
  names(kline_dt) <- c(
    "datetime", "open", "high", "low", "close", "volume",
    "datetime_close", "quote_asset_volume", "num_trades",
    "taker_buy_base_vol", "taker_buy_quote_vol", "ignore"
  )
  kline_dt <- .normalize_time_cols(kline_dt, c("datetime", "datetime_close"))
  .coerce_numeric_cols(
    kline_dt,
    c("open", "high", "low", "close", "volume", "quote_asset_volume",
      "num_trades", "taker_buy_base_vol", "taker_buy_quote_vol", "ignore")
  )
}

test_that("spot ticker price rejects symbol and symbols together", {
  expect_error(
    spot_get_ticker_price(symbol = "BTCUSDT", symbols = c("ETHUSDT")),
    "Only one of symbol, symbols may be supplied"
  )
})

test_that("spot ticker price converts multi-symbol responses to data.table", {
  local_mocked_bindings(
    .request_public = function(config, path, query = NULL) {
      list(
        list(symbol = "BTCUSDT", price = "100"),
        list(symbol = "ETHUSDT", price = "200")
      )
    },
    .package = "binxr"
  )

  out <- spot_get_ticker_price(symbols = c("BTCUSDT", "ETHUSDT"))

  expect_s3_class(out, "data.table")
  expect_equal(nrow(out), 2L)
})

test_that("spot klines default to data.table and support json_list", {
  payload <- list(
    list(1, "1", "2", "0.5", "1.5", "10", 2, "20", 3, "4", "5", "0")
  )

  local_mocked_bindings(
    .request_public = function(config, path, query = NULL) payload,
    .package = "binxr"
  )

  dt_out <- spot_get_klines("BTCUSDT", "1m")
  list_out <- spot_get_klines("BTCUSDT", "1m", json_list = TRUE)

  expect_s3_class(dt_out, "data.table")
  expect_true("datetime" %in% names(dt_out))
  expect_identical(list_out, payload)
})

test_that("spot order book passes symbolStatus through explicitly", {
  local_mocked_bindings(
    .request_public = function(config, path, query = NULL) query,
    .package = "binxr"
  )

  out <- spot_get_order_book("BTCUSDT", limit = 100, symbol_status = "TRADING")

  expect_identical(out$symbol, "BTCUSDT")
  expect_identical(out$limit, 100)
  expect_identical(out$symbolStatus, "TRADING")
})

test_that("spot 24hr ticker includes type and supports tabular array responses", {
  local_mocked_bindings(
    .request_public = function(config, path, query = NULL) {
      if (is.null(query$symbol)) {
        return(list(
          list(symbol = "BTCUSDT", lastPrice = "100"),
          list(symbol = "ETHUSDT", lastPrice = "200")
        ))
      }
      list(symbol = query$symbol, lastPrice = "100")
    },
    .package = "binxr"
  )

  dt_out <- spot_get_24hr_ticker(symbols = c("BTCUSDT", "ETHUSDT"), type = "MINI")
  list_out <- spot_get_24hr_ticker(symbol = "BTCUSDT", type = "FULL")

  expect_s3_class(dt_out, "data.table")
  expect_equal(nrow(dt_out), 2L)
  expect_type(list_out, "list")
  expect_identical(list_out$symbol, "BTCUSDT")
})

test_that("spot recent and historical trades default to data.table and support json_list", {
  payload <- list(
    list(
      id = 1,
      price = "100",
      qty = "2",
      quoteQty = "200",
      time = 1700000000000,
      isBuyerMaker = TRUE,
      isBestMatch = TRUE
    )
  )

  local_mocked_bindings(
    .request_public = function(config, path, query = NULL) payload,
    .package = "binxr"
  )

  recent_dt <- spot_get_recent_trades("BTCUSDT")
  historical_list <- spot_get_historical_trades("BTCUSDT", fromId = 10, json_list = TRUE)

  expect_s3_class(recent_dt, "data.table")
  expect_identical(historical_list, payload)
})

test_that("spot aggregate trades pass selector parameters through and default to data.table", {
  local_mocked_bindings(
    .request_public = function(config, path, query = NULL) {
      if (!is.null(query$fromId)) {
        return(query)
      }
      list(list(a = 1, p = "1", q = "2", f = 1, l = 2, T = 3, m = TRUE, M = TRUE))
    },
    .package = "binxr"
  )

  query_out <- spot_get_aggregate_trades(
    "BTCUSDT",
    fromId = 5,
    startTime = 10,
    endTime = 20,
    limit = 50,
    json_list = TRUE
  )
  dt_out <- spot_get_aggregate_trades("BTCUSDT")

  expect_identical(query_out$symbol, "BTCUSDT")
  expect_identical(query_out$fromId, 5)
  expect_identical(query_out$startTime, 10)
  expect_identical(query_out$endTime, 20)
  expect_identical(query_out$limit, 50)
  expect_s3_class(dt_out, "data.table")
})

test_that("spot trading day ticker requires a selector and supports tabular responses", {
  expect_error(
    spot_get_trading_day_ticker(),
    "Provide at least one of: symbol, symbols"
  )

  local_mocked_bindings(
    .request_public = function(config, path, query = NULL) {
      if (!is.null(query$symbols)) {
        return(list(
          list(symbol = "BTCUSDT", lastPrice = "100"),
          list(symbol = "ETHUSDT", lastPrice = "200")
        ))
      }
      query
    },
    .package = "binxr"
  )

  dt_out <- spot_get_trading_day_ticker(
    symbols = c("BTCUSDT", "ETHUSDT"),
    timeZone = "8",
    type = "MINI",
    symbol_status = "TRADING"
  )
  query_out <- spot_get_trading_day_ticker(symbol = "BTCUSDT", timeZone = "0", json_list = TRUE)

  expect_s3_class(dt_out, "data.table")
  expect_identical(query_out$symbol, "BTCUSDT")
  expect_identical(query_out$timeZone, "0")
  expect_identical(query_out$type, "FULL")
})

test_that("spot rolling window ticker validates windowSize and requires a selector", {
  expect_error(
    spot_get_rolling_window_ticker(),
    "Provide at least one of: symbol, symbols"
  )
  expect_error(
    spot_get_rolling_window_ticker(symbol = "BTCUSDT", windowSize = "1d2h"),
    "must be one of 1m-59m, 1h-23h, or 1d-7d"
  )

  local_mocked_bindings(
    .request_public = function(config, path, query = NULL) query,
    .package = "binxr"
  )

  out <- spot_get_rolling_window_ticker(
    symbol = "BTCUSDT",
    windowSize = "4h",
    type = "MINI",
    symbol_status = "BREAK",
    json_list = TRUE
  )

  expect_identical(out$symbol, "BTCUSDT")
  expect_identical(out$windowSize, "4h")
  expect_identical(out$type, "MINI")
  expect_identical(out$symbolStatus, "BREAK")
})

test_that("spot reference price endpoints pass required parameters through", {
  local_mocked_bindings(
    .request_public = function(config, path, query = NULL) list(path = path, query = query),
    .package = "binxr"
  )

  ref_out <- spot_get_reference_price("BTCUSDT")
  calc_out <- spot_get_reference_price_calculation("BTCUSDT", symbol_status = "HALT")

  expect_identical(ref_out$path, "/api/v3/referencePrice")
  expect_identical(ref_out$query$symbol, "BTCUSDT")
  expect_identical(calc_out$path, "/api/v3/referencePrice/calculation")
  expect_identical(calc_out$query$symbolStatus, "HALT")
})

#' Place a Binance Spot order
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param side One of `"BUY"` or `"SELL"`.
#' @param type Order type.
#' @param time_in_force Optional time-in-force value.
#' @param quantity Optional order quantity.
#' @param quote_order_qty Optional quote-order quantity for `MARKET` orders.
#' @param price Optional limit price.
#' @param new_client_order_id Optional client order identifier.
#' @param strategy_id Optional numeric strategy identifier.
#' @param strategy_type Optional numeric strategy type. Must be at least `1000000`.
#' @param stop_price Optional trigger price.
#' @param trailing_delta Optional trailing delta value.
#' @param iceberg_qty Optional iceberg quantity.
#' @param new_order_resp_type Optional response type. One of `"ACK"`, `"RESULT"`, or `"FULL"`.
#' @param self_trade_prevention_mode Optional self-trade prevention mode.
#' @param peg_price_type Optional pegged price type. One of `"PRIMARY_PEG"` or `"MARKET_PEG"`.
#' @param peg_offset_value Optional peg offset value.
#' @param peg_offset_type Optional peg offset type. Only `"PRICE_LEVEL"` is supported.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_place_order <- function(
    symbol,
    side = c("BUY", "SELL"),
    type = c(
      "LIMIT", "MARKET", "STOP_LOSS", "STOP_LOSS_LIMIT",
      "TAKE_PROFIT", "TAKE_PROFIT_LIMIT", "LIMIT_MAKER"
    ),
    time_in_force = NULL,
    quantity = NULL,
    quote_order_qty = NULL,
    price = NULL,
    new_client_order_id = NULL,
    strategy_id = NULL,
    strategy_type = NULL,
    stop_price = NULL,
    trailing_delta = NULL,
    iceberg_qty = NULL,
    new_order_resp_type = NULL,
    self_trade_prevention_mode = NULL,
    peg_price_type = NULL,
    peg_offset_value = NULL,
    peg_offset_type = NULL,
    config = config_spot()) {
  side <- match.arg(side)
  type <- match.arg(type)
  .validate_symbol(symbol)
  .validate_optional_scalar_character(time_in_force, "time_in_force")
  .validate_optional_scalar_numeric(quantity, "quantity")
  .validate_optional_scalar_numeric(quote_order_qty, "quote_order_qty")
  .validate_optional_scalar_numeric(price, "price")
  .validate_optional_scalar_character(new_client_order_id, "new_client_order_id")
  .validate_optional_scalar_numeric(strategy_id, "strategy_id")
  .validate_optional_scalar_numeric(strategy_type, "strategy_type")
  .validate_optional_scalar_numeric(stop_price, "stop_price")
  .validate_optional_scalar_numeric(trailing_delta, "trailing_delta")
  .validate_optional_scalar_numeric(iceberg_qty, "iceberg_qty")
  .validate_optional_scalar_character(new_order_resp_type, "new_order_resp_type")
  .validate_optional_scalar_character(self_trade_prevention_mode, "self_trade_prevention_mode")
  .validate_optional_scalar_character(peg_price_type, "peg_price_type")
  .validate_optional_scalar_numeric(peg_offset_value, "peg_offset_value")
  .validate_optional_scalar_character(peg_offset_type, "peg_offset_type")

  .spot_validate_order_request(
    type = type,
    time_in_force = time_in_force,
    quantity = quantity,
    quote_order_qty = quote_order_qty,
    price = price,
    strategy_type = strategy_type,
    stop_price = stop_price,
    trailing_delta = trailing_delta,
    iceberg_qty = iceberg_qty,
    new_order_resp_type = new_order_resp_type,
    peg_price_type = peg_price_type,
    peg_offset_value = peg_offset_value,
    peg_offset_type = peg_offset_type
  )

  .request_signed(
    config,
    "/api/v3/order",
    params = list(
      symbol = symbol,
      side = side,
      type = type,
      timeInForce = time_in_force,
      quantity = quantity,
      quoteOrderQty = quote_order_qty,
      price = price,
      newClientOrderId = new_client_order_id,
      strategyId = strategy_id,
      strategyType = strategy_type,
      stopPrice = stop_price,
      trailingDelta = trailing_delta,
      icebergQty = iceberg_qty,
      newOrderRespType = new_order_resp_type,
      selfTradePreventionMode = self_trade_prevention_mode,
      pegPriceType = peg_price_type,
      pegOffsetValue = peg_offset_value,
      pegOffsetType = peg_offset_type
    ),
    method = "POST"
  )
}

#' Test a Binance Spot order
#'
#' @inheritParams spot_place_order
#' @param compute_commission_rates Whether to return commission-rate details.
#'
#' @return A parsed list.
#' @export
spot_test_order <- function(
    symbol,
    side = c("BUY", "SELL"),
    type = c(
      "LIMIT", "MARKET", "STOP_LOSS", "STOP_LOSS_LIMIT",
      "TAKE_PROFIT", "TAKE_PROFIT_LIMIT", "LIMIT_MAKER"
    ),
    time_in_force = NULL,
    quantity = NULL,
    quote_order_qty = NULL,
    price = NULL,
    new_client_order_id = NULL,
    strategy_id = NULL,
    strategy_type = NULL,
    stop_price = NULL,
    trailing_delta = NULL,
    iceberg_qty = NULL,
    new_order_resp_type = NULL,
    self_trade_prevention_mode = NULL,
    peg_price_type = NULL,
    peg_offset_value = NULL,
    peg_offset_type = NULL,
    compute_commission_rates = FALSE,
    config = config_spot()) {
  .validate_scalar_logical(compute_commission_rates, "compute_commission_rates")

  side <- match.arg(side)
  type <- match.arg(type)
  .validate_symbol(symbol)
  .validate_optional_scalar_character(time_in_force, "time_in_force")
  .validate_optional_scalar_numeric(quantity, "quantity")
  .validate_optional_scalar_numeric(quote_order_qty, "quote_order_qty")
  .validate_optional_scalar_numeric(price, "price")
  .validate_optional_scalar_character(new_client_order_id, "new_client_order_id")
  .validate_optional_scalar_numeric(strategy_id, "strategy_id")
  .validate_optional_scalar_numeric(strategy_type, "strategy_type")
  .validate_optional_scalar_numeric(stop_price, "stop_price")
  .validate_optional_scalar_numeric(trailing_delta, "trailing_delta")
  .validate_optional_scalar_numeric(iceberg_qty, "iceberg_qty")
  .validate_optional_scalar_character(new_order_resp_type, "new_order_resp_type")
  .validate_optional_scalar_character(self_trade_prevention_mode, "self_trade_prevention_mode")
  .validate_optional_scalar_character(peg_price_type, "peg_price_type")
  .validate_optional_scalar_numeric(peg_offset_value, "peg_offset_value")
  .validate_optional_scalar_character(peg_offset_type, "peg_offset_type")

  .spot_validate_order_request(
    type = type,
    time_in_force = time_in_force,
    quantity = quantity,
    quote_order_qty = quote_order_qty,
    price = price,
    strategy_type = strategy_type,
    stop_price = stop_price,
    trailing_delta = trailing_delta,
    iceberg_qty = iceberg_qty,
    new_order_resp_type = new_order_resp_type,
    peg_price_type = peg_price_type,
    peg_offset_value = peg_offset_value,
    peg_offset_type = peg_offset_type
  )

  .request_signed(
    config,
    "/api/v3/order/test",
    params = list(
      symbol = symbol,
      side = side,
      type = type,
      timeInForce = time_in_force,
      quantity = quantity,
      quoteOrderQty = quote_order_qty,
      price = price,
      newClientOrderId = new_client_order_id,
      strategyId = strategy_id,
      strategyType = strategy_type,
      stopPrice = stop_price,
      trailingDelta = trailing_delta,
      icebergQty = iceberg_qty,
      newOrderRespType = new_order_resp_type,
      selfTradePreventionMode = self_trade_prevention_mode,
      pegPriceType = peg_price_type,
      pegOffsetValue = peg_offset_value,
      pegOffsetType = peg_offset_type,
      computeCommissionRates = if (isTRUE(compute_commission_rates)) "true" else "false"
    ),
    method = "POST"
  )
}

#' Cancel a Binance Spot order
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param order_id Optional exchange order ID.
#' @param orig_client_order_id Optional client order ID.
#' @param new_client_order_id Optional client order ID for the cancel request.
#' @param cancel_restrictions Optional cancel restriction. One of `"ONLY_NEW"` or `"ONLY_PARTIALLY_FILLED"`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_cancel_order <- function(
    symbol,
    order_id = NULL,
    orig_client_order_id = NULL,
    new_client_order_id = NULL,
    cancel_restrictions = NULL,
    config = config_spot()) {
  .validate_symbol(symbol)
  .validate_required_any(order_id, orig_client_order_id, arg_names = c("order_id", "orig_client_order_id"))
  .validate_optional_scalar_numeric(order_id, "order_id")
  .validate_optional_scalar_character(orig_client_order_id, "orig_client_order_id")
  .validate_optional_scalar_character(new_client_order_id, "new_client_order_id")
  .validate_optional_scalar_character(cancel_restrictions, "cancel_restrictions")
  if (!is.null(cancel_restrictions)) {
    .validate_one_of(cancel_restrictions, c("ONLY_NEW", "ONLY_PARTIALLY_FILLED"), "cancel_restrictions")
  }

  .request_signed(
    config,
    "/api/v3/order",
    params = list(
      symbol = symbol,
      orderId = order_id,
      origClientOrderId = orig_client_order_id,
      newClientOrderId = new_client_order_id,
      cancelRestrictions = cancel_restrictions
    ),
    method = "DELETE"
  )
}

#' Cancel all open Binance Spot orders for a symbol
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list by default.
#' @export
spot_cancel_all_orders <- function(symbol, config = config_spot()) {
  .validate_symbol(symbol)
  .request_signed(
    config,
    "/api/v3/openOrders",
    params = list(symbol = symbol),
    method = "DELETE"
  )
}

#' Cancel and replace a Binance Spot order
#'
#' @inheritParams spot_place_order
#' @param cancel_replace_mode One of `"STOP_ON_FAILURE"` or `"ALLOW_FAILURE"`.
#' @param cancel_new_client_order_id Optional client order ID for the cancel request.
#' @param cancel_orig_client_order_id Optional original client order ID to cancel.
#' @param cancel_order_id Optional exchange order ID to cancel.
#' @param cancel_restrictions Optional cancel restriction. One of `"ONLY_NEW"` or `"ONLY_PARTIALLY_FILLED"`.
#' @param order_rate_limit_exceeded_mode Optional order-rate-limit behavior. One of `"DO_NOTHING"` or `"CANCEL_ONLY"`.
#'
#' @return A parsed list.
#' @export
spot_cancel_replace_order <- function(
    symbol,
    side = c("BUY", "SELL"),
    type = c(
      "LIMIT", "MARKET", "STOP_LOSS", "STOP_LOSS_LIMIT",
      "TAKE_PROFIT", "TAKE_PROFIT_LIMIT", "LIMIT_MAKER"
    ),
    cancel_replace_mode = c("STOP_ON_FAILURE", "ALLOW_FAILURE"),
    time_in_force = NULL,
    quantity = NULL,
    quote_order_qty = NULL,
    price = NULL,
    cancel_new_client_order_id = NULL,
    cancel_orig_client_order_id = NULL,
    cancel_order_id = NULL,
    new_client_order_id = NULL,
    strategy_id = NULL,
    strategy_type = NULL,
    stop_price = NULL,
    trailing_delta = NULL,
    iceberg_qty = NULL,
    new_order_resp_type = NULL,
    self_trade_prevention_mode = NULL,
    cancel_restrictions = NULL,
    order_rate_limit_exceeded_mode = NULL,
    peg_price_type = NULL,
    peg_offset_value = NULL,
    peg_offset_type = NULL,
    config = config_spot()) {
  side <- match.arg(side)
  type <- match.arg(type)
  cancel_replace_mode <- match.arg(cancel_replace_mode)
  .validate_symbol(symbol)
  .validate_optional_scalar_character(cancel_new_client_order_id, "cancel_new_client_order_id")
  .validate_optional_scalar_character(cancel_orig_client_order_id, "cancel_orig_client_order_id")
  .validate_optional_scalar_numeric(cancel_order_id, "cancel_order_id")
  .validate_required_any(cancel_order_id, cancel_orig_client_order_id, arg_names = c("cancel_order_id", "cancel_orig_client_order_id"))
  .validate_optional_scalar_character(cancel_restrictions, "cancel_restrictions")
  .validate_optional_scalar_character(order_rate_limit_exceeded_mode, "order_rate_limit_exceeded_mode")

  if (!is.null(cancel_restrictions)) {
    .validate_one_of(cancel_restrictions, c("ONLY_NEW", "ONLY_PARTIALLY_FILLED"), "cancel_restrictions")
  }
  if (!is.null(order_rate_limit_exceeded_mode)) {
    .validate_one_of(order_rate_limit_exceeded_mode, c("DO_NOTHING", "CANCEL_ONLY"), "order_rate_limit_exceeded_mode")
  }

  .spot_validate_order_request(
    type = type,
    time_in_force = time_in_force,
    quantity = quantity,
    quote_order_qty = quote_order_qty,
    price = price,
    strategy_type = strategy_type,
    stop_price = stop_price,
    trailing_delta = trailing_delta,
    iceberg_qty = iceberg_qty,
    new_order_resp_type = new_order_resp_type,
    peg_price_type = peg_price_type,
    peg_offset_value = peg_offset_value,
    peg_offset_type = peg_offset_type
  )

  .request_signed(
    config,
    "/api/v3/order/cancelReplace",
    params = list(
      symbol = symbol,
      side = side,
      type = type,
      cancelReplaceMode = cancel_replace_mode,
      timeInForce = time_in_force,
      quantity = quantity,
      quoteOrderQty = quote_order_qty,
      price = price,
      cancelNewClientOrderId = cancel_new_client_order_id,
      cancelOrigClientOrderId = cancel_orig_client_order_id,
      cancelOrderId = cancel_order_id,
      newClientOrderId = new_client_order_id,
      strategyId = strategy_id,
      strategyType = strategy_type,
      stopPrice = stop_price,
      trailingDelta = trailing_delta,
      icebergQty = iceberg_qty,
      newOrderRespType = new_order_resp_type,
      selfTradePreventionMode = self_trade_prevention_mode,
      cancelRestrictions = cancel_restrictions,
      orderRateLimitExceededMode = order_rate_limit_exceeded_mode,
      pegPriceType = peg_price_type,
      pegOffsetValue = peg_offset_value,
      pegOffsetType = peg_offset_type
    ),
    method = "POST"
  )
}

#' Amend a Binance Spot order while keeping priority
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param order_id Optional exchange order ID.
#' @param orig_client_order_id Optional client order ID.
#' @param new_client_order_id Optional new client order ID after amendment.
#' @param new_qty New order quantity. Must be positive.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_amend_order_keep_priority <- function(
    symbol,
    order_id = NULL,
    orig_client_order_id = NULL,
    new_client_order_id = NULL,
    new_qty,
    config = config_spot()) {
  .validate_symbol(symbol)
  .validate_required_any(order_id, orig_client_order_id, arg_names = c("order_id", "orig_client_order_id"))
  .validate_optional_scalar_numeric(order_id, "order_id")
  .validate_optional_scalar_character(orig_client_order_id, "orig_client_order_id")
  .validate_optional_scalar_character(new_client_order_id, "new_client_order_id")
  .validate_positive_number(new_qty, "new_qty")

  .request_signed(
    config,
    "/api/v3/order/amend/keepPriority",
    params = list(
      symbol = symbol,
      orderId = order_id,
      origClientOrderId = orig_client_order_id,
      newClientOrderId = new_client_order_id,
      newQty = new_qty
    ),
    method = "PUT"
  )
}

#' Place a deprecated Binance Spot OCO order
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param side One of `"BUY"` or `"SELL"`.
#' @param quantity Order quantity shared by both legs.
#' @param price Limit price for the maker leg.
#' @param stop_price Stop trigger price.
#' @param list_client_order_id Optional order-list client ID.
#' @param limit_client_order_id Optional limit-leg client ID.
#' @param limit_strategy_id Optional limit-leg strategy ID.
#' @param limit_strategy_type Optional limit-leg strategy type.
#' @param limit_iceberg_qty Optional limit-leg iceberg quantity.
#' @param trailing_delta Optional shared trailing delta.
#' @param stop_client_order_id Optional stop-leg client ID.
#' @param stop_strategy_id Optional stop-leg strategy ID.
#' @param stop_strategy_type Optional stop-leg strategy type.
#' @param stop_limit_price Optional stop-limit price.
#' @param stop_iceberg_qty Optional stop-leg iceberg quantity.
#' @param stop_limit_time_in_force Optional stop-limit time-in-force.
#' @param new_order_resp_type Optional response type. One of `"ACK"`, `"RESULT"`, or `"FULL"`.
#' @param self_trade_prevention_mode Optional self-trade prevention mode.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_place_oco_order <- function(
    symbol,
    side = c("BUY", "SELL"),
    quantity,
    price,
    stop_price,
    list_client_order_id = NULL,
    limit_client_order_id = NULL,
    limit_strategy_id = NULL,
    limit_strategy_type = NULL,
    limit_iceberg_qty = NULL,
    trailing_delta = NULL,
    stop_client_order_id = NULL,
    stop_strategy_id = NULL,
    stop_strategy_type = NULL,
    stop_limit_price = NULL,
    stop_iceberg_qty = NULL,
    stop_limit_time_in_force = NULL,
    new_order_resp_type = NULL,
    self_trade_prevention_mode = NULL,
    config = config_spot()) {
  side <- match.arg(side)
  .validate_symbol(symbol)
  .validate_positive_number(quantity, "quantity")
  .validate_positive_number(price, "price")
  .validate_positive_number(stop_price, "stop_price")
  .validate_optional_scalar_character(list_client_order_id, "list_client_order_id")
  .validate_optional_scalar_character(limit_client_order_id, "limit_client_order_id")
  .validate_optional_scalar_character(stop_client_order_id, "stop_client_order_id")
  .validate_optional_scalar_numeric(limit_strategy_id, "limit_strategy_id")
  .validate_optional_scalar_numeric(limit_strategy_type, "limit_strategy_type")
  .validate_optional_scalar_numeric(limit_iceberg_qty, "limit_iceberg_qty")
  .validate_optional_scalar_numeric(trailing_delta, "trailing_delta")
  .validate_optional_scalar_numeric(stop_strategy_id, "stop_strategy_id")
  .validate_optional_scalar_numeric(stop_strategy_type, "stop_strategy_type")
  .validate_optional_scalar_numeric(stop_limit_price, "stop_limit_price")
  .validate_optional_scalar_numeric(stop_iceberg_qty, "stop_iceberg_qty")
  .validate_optional_scalar_character(stop_limit_time_in_force, "stop_limit_time_in_force")
  .validate_optional_scalar_character(new_order_resp_type, "new_order_resp_type")
  .validate_optional_scalar_character(self_trade_prevention_mode, "self_trade_prevention_mode")

  .spot_validate_strategy_type(limit_strategy_type, "limit_strategy_type")
  .spot_validate_strategy_type(stop_strategy_type, "stop_strategy_type")
  .spot_validate_positive_optional(limit_iceberg_qty, "limit_iceberg_qty")
  .spot_validate_positive_optional(trailing_delta, "trailing_delta")
  .spot_validate_positive_optional(stop_limit_price, "stop_limit_price")
  .spot_validate_positive_optional(stop_iceberg_qty, "stop_iceberg_qty")
  .spot_validate_optional_resp_type(new_order_resp_type)

  if (!is.null(stop_limit_price) && is.null(stop_limit_time_in_force)) {
    stop("`stop_limit_time_in_force` is required when `stop_limit_price` is supplied.", call. = FALSE)
  }
  if (!is.null(stop_limit_time_in_force)) {
    .validate_one_of(stop_limit_time_in_force, c("GTC", "FOK", "IOC"), "stop_limit_time_in_force")
  }

  .request_signed(
    config,
    "/api/v3/order/oco",
    params = list(
      symbol = symbol,
      listClientOrderId = list_client_order_id,
      side = side,
      quantity = quantity,
      limitClientOrderId = limit_client_order_id,
      price = price,
      limitStrategyId = limit_strategy_id,
      limitStrategyType = limit_strategy_type,
      limitIcebergQty = limit_iceberg_qty,
      trailingDelta = trailing_delta,
      stopClientOrderId = stop_client_order_id,
      stopPrice = stop_price,
      stopStrategyId = stop_strategy_id,
      stopStrategyType = stop_strategy_type,
      stopLimitPrice = stop_limit_price,
      stopIcebergQty = stop_iceberg_qty,
      stopLimitTimeInForce = stop_limit_time_in_force,
      newOrderRespType = new_order_resp_type,
      selfTradePreventionMode = self_trade_prevention_mode
    ),
    method = "POST"
  )
}

#' Place a Binance Spot OCO order list
#'
#' @inheritParams spot_place_oco_order
#' @param above_type Type for the above leg.
#' @param above_client_order_id Optional above-leg client ID.
#' @param above_iceberg_qty Optional above-leg iceberg quantity.
#' @param above_price Optional above-leg limit price.
#' @param above_stop_price Optional above-leg stop price.
#' @param above_trailing_delta Optional above-leg trailing delta.
#' @param above_time_in_force Optional above-leg time-in-force.
#' @param above_strategy_id Optional above-leg strategy ID.
#' @param above_strategy_type Optional above-leg strategy type.
#' @param above_peg_price_type Optional above-leg peg price type.
#' @param above_peg_offset_type Optional above-leg peg offset type.
#' @param above_peg_offset_value Optional above-leg peg offset value.
#' @param below_type Type for the below leg.
#' @param below_client_order_id Optional below-leg client ID.
#' @param below_iceberg_qty Optional below-leg iceberg quantity.
#' @param below_price Optional below-leg limit price.
#' @param below_stop_price Optional below-leg stop price.
#' @param below_trailing_delta Optional below-leg trailing delta.
#' @param below_time_in_force Optional below-leg time-in-force.
#' @param below_strategy_id Optional below-leg strategy ID.
#' @param below_strategy_type Optional below-leg strategy type.
#' @param below_peg_price_type Optional below-leg peg price type.
#' @param below_peg_offset_type Optional below-leg peg offset type.
#' @param below_peg_offset_value Optional below-leg peg offset value.
#'
#' @return A parsed list.
#' @export
spot_place_order_list_oco <- function(
    symbol,
    side = c("BUY", "SELL"),
    quantity,
    above_type = c("STOP_LOSS_LIMIT", "STOP_LOSS", "LIMIT_MAKER", "TAKE_PROFIT", "TAKE_PROFIT_LIMIT"),
    above_client_order_id = NULL,
    above_iceberg_qty = NULL,
    above_price = NULL,
    above_stop_price = NULL,
    above_trailing_delta = NULL,
    above_time_in_force = NULL,
    above_strategy_id = NULL,
    above_strategy_type = NULL,
    above_peg_price_type = NULL,
    above_peg_offset_type = NULL,
    above_peg_offset_value = NULL,
    below_type = c("STOP_LOSS", "STOP_LOSS_LIMIT", "TAKE_PROFIT", "TAKE_PROFIT_LIMIT"),
    below_client_order_id = NULL,
    below_iceberg_qty = NULL,
    below_price = NULL,
    below_stop_price = NULL,
    below_trailing_delta = NULL,
    below_time_in_force = NULL,
    below_strategy_id = NULL,
    below_strategy_type = NULL,
    below_peg_price_type = NULL,
    below_peg_offset_type = NULL,
    below_peg_offset_value = NULL,
    list_client_order_id = NULL,
    new_order_resp_type = NULL,
    self_trade_prevention_mode = NULL,
    config = config_spot()) {
  side <- match.arg(side)
  above_type <- match.arg(above_type)
  below_type <- match.arg(below_type)
  .validate_symbol(symbol)
  .validate_positive_number(quantity, "quantity")
  .validate_optional_scalar_character(list_client_order_id, "list_client_order_id")
  .validate_optional_scalar_character(above_client_order_id, "above_client_order_id")
  .validate_optional_scalar_character(below_client_order_id, "below_client_order_id")
  .spot_validate_order_list_leg(
    type = above_type,
    price = above_price,
    stop_price = above_stop_price,
    trailing_delta = above_trailing_delta,
    time_in_force = above_time_in_force,
    iceberg_qty = above_iceberg_qty,
    strategy_type = above_strategy_type,
    peg_price_type = above_peg_price_type,
    peg_offset_type = above_peg_offset_type,
    peg_offset_value = above_peg_offset_value,
    prefix = "above"
  )
  .spot_validate_order_list_leg(
    type = below_type,
    price = below_price,
    stop_price = below_stop_price,
    trailing_delta = below_trailing_delta,
    time_in_force = below_time_in_force,
    iceberg_qty = below_iceberg_qty,
    strategy_type = below_strategy_type,
    peg_price_type = below_peg_price_type,
    peg_offset_type = below_peg_offset_type,
    peg_offset_value = below_peg_offset_value,
    prefix = "below"
  )
  .spot_validate_optional_resp_type(new_order_resp_type)

  .request_signed(
    config,
    "/api/v3/orderList/oco",
    params = list(
      symbol = symbol,
      listClientOrderId = list_client_order_id,
      side = side,
      quantity = quantity,
      aboveType = above_type,
      aboveClientOrderId = above_client_order_id,
      aboveIcebergQty = above_iceberg_qty,
      abovePrice = above_price,
      aboveStopPrice = above_stop_price,
      aboveTrailingDelta = above_trailing_delta,
      aboveTimeInForce = above_time_in_force,
      aboveStrategyId = above_strategy_id,
      aboveStrategyType = above_strategy_type,
      abovePegPriceType = above_peg_price_type,
      abovePegOffsetType = above_peg_offset_type,
      abovePegOffsetValue = above_peg_offset_value,
      belowType = below_type,
      belowClientOrderId = below_client_order_id,
      belowIcebergQty = below_iceberg_qty,
      belowPrice = below_price,
      belowStopPrice = below_stop_price,
      belowTrailingDelta = below_trailing_delta,
      belowTimeInForce = below_time_in_force,
      belowStrategyId = below_strategy_id,
      belowStrategyType = below_strategy_type,
      belowPegPriceType = below_peg_price_type,
      belowPegOffsetType = below_peg_offset_type,
      belowPegOffsetValue = below_peg_offset_value,
      newOrderRespType = new_order_resp_type,
      selfTradePreventionMode = self_trade_prevention_mode
    ),
    method = "POST"
  )
}

#' Place a Binance Spot OTO order list
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param list_client_order_id Optional order-list client ID.
#' @param new_order_resp_type Optional response type. One of `"ACK"`, `"RESULT"`, or `"FULL"`.
#' @param self_trade_prevention_mode Optional self-trade prevention mode.
#' @param working_type Working leg type. One of `"LIMIT"` or `"LIMIT_MAKER"`.
#' @param working_side Working leg side. One of `"BUY"` or `"SELL"`.
#' @param working_client_order_id Optional working-leg client ID.
#' @param working_price Working leg price.
#' @param working_quantity Working leg quantity.
#' @param working_iceberg_qty Optional working-leg iceberg quantity.
#' @param working_time_in_force Optional working-leg time-in-force.
#' @param working_strategy_id Optional working-leg strategy ID.
#' @param working_strategy_type Optional working-leg strategy type.
#' @param working_peg_price_type Optional working-leg peg price type.
#' @param working_peg_offset_type Optional working-leg peg offset type.
#' @param working_peg_offset_value Optional working-leg peg offset value.
#' @param pending_type Pending leg order type.
#' @param pending_side Pending leg side. One of `"BUY"` or `"SELL"`.
#' @param pending_client_order_id Optional pending-leg client ID.
#' @param pending_price Optional pending-leg price.
#' @param pending_stop_price Optional pending-leg stop price.
#' @param pending_trailing_delta Optional pending-leg trailing delta.
#' @param pending_quantity Pending leg quantity.
#' @param pending_iceberg_qty Optional pending-leg iceberg quantity.
#' @param pending_time_in_force Optional pending-leg time-in-force.
#' @param pending_strategy_id Optional pending-leg strategy ID.
#' @param pending_strategy_type Optional pending-leg strategy type.
#' @param pending_peg_price_type Optional pending-leg peg price type.
#' @param pending_peg_offset_type Optional pending-leg peg offset type.
#' @param pending_peg_offset_value Optional pending-leg peg offset value.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_place_order_list_oto <- function(
    symbol,
    list_client_order_id = NULL,
    new_order_resp_type = NULL,
    self_trade_prevention_mode = NULL,
    working_type = c("LIMIT", "LIMIT_MAKER"),
    working_side = c("BUY", "SELL"),
    working_client_order_id = NULL,
    working_price,
    working_quantity,
    working_iceberg_qty = NULL,
    working_time_in_force = NULL,
    working_strategy_id = NULL,
    working_strategy_type = NULL,
    working_peg_price_type = NULL,
    working_peg_offset_type = NULL,
    working_peg_offset_value = NULL,
    pending_type = c("LIMIT", "MARKET", "STOP_LOSS", "STOP_LOSS_LIMIT", "TAKE_PROFIT", "TAKE_PROFIT_LIMIT", "LIMIT_MAKER"),
    pending_side = c("BUY", "SELL"),
    pending_client_order_id = NULL,
    pending_price = NULL,
    pending_stop_price = NULL,
    pending_trailing_delta = NULL,
    pending_quantity,
    pending_iceberg_qty = NULL,
    pending_time_in_force = NULL,
    pending_strategy_id = NULL,
    pending_strategy_type = NULL,
    pending_peg_price_type = NULL,
    pending_peg_offset_type = NULL,
    pending_peg_offset_value = NULL,
    config = config_spot()) {
  working_type <- match.arg(working_type)
  working_side <- match.arg(working_side)
  pending_type <- match.arg(pending_type)
  pending_side <- match.arg(pending_side)
  .validate_symbol(symbol)
  .validate_positive_number(working_price, "working_price")
  .validate_positive_number(working_quantity, "working_quantity")
  .validate_positive_number(pending_quantity, "pending_quantity")
  if (identical(pending_type, "MARKET") && !is.null(pending_price)) {
    stop("`pending_price` must not be supplied when `pending_type = \"MARKET\"`.", call. = FALSE)
  }
  .spot_validate_working_order_list_leg(
    type = working_type,
    time_in_force = working_time_in_force,
    iceberg_qty = working_iceberg_qty,
    strategy_type = working_strategy_type,
    peg_price_type = working_peg_price_type,
    peg_offset_type = working_peg_offset_type,
    peg_offset_value = working_peg_offset_value,
    prefix = "working"
  )
  .spot_validate_order_list_leg(
    type = pending_type,
    price = pending_price,
    stop_price = pending_stop_price,
    trailing_delta = pending_trailing_delta,
    time_in_force = pending_time_in_force,
    iceberg_qty = pending_iceberg_qty,
    strategy_type = pending_strategy_type,
    peg_price_type = pending_peg_price_type,
    peg_offset_type = pending_peg_offset_type,
    peg_offset_value = pending_peg_offset_value,
    prefix = "pending",
    allow_market = TRUE
  )
  .spot_validate_optional_resp_type(new_order_resp_type)

  .request_signed(
    config,
    "/api/v3/orderList/oto",
    params = list(
      symbol = symbol,
      listClientOrderId = list_client_order_id,
      newOrderRespType = new_order_resp_type,
      selfTradePreventionMode = self_trade_prevention_mode,
      workingType = working_type,
      workingSide = working_side,
      workingClientOrderId = working_client_order_id,
      workingPrice = working_price,
      workingQuantity = working_quantity,
      workingIcebergQty = working_iceberg_qty,
      workingTimeInForce = working_time_in_force,
      workingStrategyId = working_strategy_id,
      workingStrategyType = working_strategy_type,
      workingPegPriceType = working_peg_price_type,
      workingPegOffsetType = working_peg_offset_type,
      workingPegOffsetValue = working_peg_offset_value,
      pendingType = pending_type,
      pendingSide = pending_side,
      pendingClientOrderId = pending_client_order_id,
      pendingPrice = pending_price,
      pendingStopPrice = pending_stop_price,
      pendingTrailingDelta = pending_trailing_delta,
      pendingQuantity = pending_quantity,
      pendingIcebergQty = pending_iceberg_qty,
      pendingTimeInForce = pending_time_in_force,
      pendingStrategyId = pending_strategy_id,
      pendingStrategyType = pending_strategy_type,
      pendingPegPriceType = pending_peg_price_type,
      pendingPegOffsetType = pending_peg_offset_type,
      pendingPegOffsetValue = pending_peg_offset_value
    ),
    method = "POST"
  )
}

#' Place a Binance Spot OTOCO order list
#'
#' @inheritParams spot_place_order_list_oto
#' @param pending_side Shared side for the pending OCO legs.
#' @param pending_above_type Type for the pending-above leg.
#' @param pending_above_client_order_id Optional pending-above client ID.
#' @param pending_above_price Optional pending-above price.
#' @param pending_above_stop_price Optional pending-above stop price.
#' @param pending_above_trailing_delta Optional pending-above trailing delta.
#' @param pending_above_iceberg_qty Optional pending-above iceberg quantity.
#' @param pending_above_time_in_force Optional pending-above time-in-force.
#' @param pending_above_strategy_id Optional pending-above strategy ID.
#' @param pending_above_strategy_type Optional pending-above strategy type.
#' @param pending_above_peg_price_type Optional pending-above peg price type.
#' @param pending_above_peg_offset_type Optional pending-above peg offset type.
#' @param pending_above_peg_offset_value Optional pending-above peg offset value.
#' @param pending_below_type Type for the pending-below leg.
#' @param pending_below_client_order_id Optional pending-below client ID.
#' @param pending_below_price Optional pending-below price.
#' @param pending_below_stop_price Optional pending-below stop price.
#' @param pending_below_trailing_delta Optional pending-below trailing delta.
#' @param pending_below_iceberg_qty Optional pending-below iceberg quantity.
#' @param pending_below_time_in_force Optional pending-below time-in-force.
#' @param pending_below_strategy_id Optional pending-below strategy ID.
#' @param pending_below_strategy_type Optional pending-below strategy type.
#' @param pending_below_peg_price_type Optional pending-below peg price type.
#' @param pending_below_peg_offset_type Optional pending-below peg offset type.
#' @param pending_below_peg_offset_value Optional pending-below peg offset value.
#'
#' @return A parsed list.
#' @export
spot_place_order_list_otoco <- function(
    symbol,
    list_client_order_id = NULL,
    new_order_resp_type = NULL,
    self_trade_prevention_mode = NULL,
    working_type = c("LIMIT", "LIMIT_MAKER"),
    working_side = c("BUY", "SELL"),
    working_client_order_id = NULL,
    working_price,
    working_quantity,
    working_iceberg_qty = NULL,
    working_time_in_force = NULL,
    working_strategy_id = NULL,
    working_strategy_type = NULL,
    working_peg_price_type = NULL,
    working_peg_offset_type = NULL,
    working_peg_offset_value = NULL,
    pending_side = c("BUY", "SELL"),
    pending_quantity,
    pending_above_type = c("STOP_LOSS_LIMIT", "STOP_LOSS", "LIMIT_MAKER", "TAKE_PROFIT", "TAKE_PROFIT_LIMIT"),
    pending_above_client_order_id = NULL,
    pending_above_price = NULL,
    pending_above_stop_price = NULL,
    pending_above_trailing_delta = NULL,
    pending_above_iceberg_qty = NULL,
    pending_above_time_in_force = NULL,
    pending_above_strategy_id = NULL,
    pending_above_strategy_type = NULL,
    pending_above_peg_price_type = NULL,
    pending_above_peg_offset_type = NULL,
    pending_above_peg_offset_value = NULL,
    pending_below_type = c("STOP_LOSS", "STOP_LOSS_LIMIT", "TAKE_PROFIT", "TAKE_PROFIT_LIMIT"),
    pending_below_client_order_id = NULL,
    pending_below_price = NULL,
    pending_below_stop_price = NULL,
    pending_below_trailing_delta = NULL,
    pending_below_iceberg_qty = NULL,
    pending_below_time_in_force = NULL,
    pending_below_strategy_id = NULL,
    pending_below_strategy_type = NULL,
    pending_below_peg_price_type = NULL,
    pending_below_peg_offset_type = NULL,
    pending_below_peg_offset_value = NULL,
    config = config_spot()) {
  working_type <- match.arg(working_type)
  working_side <- match.arg(working_side)
  pending_side <- match.arg(pending_side)
  pending_above_type <- match.arg(pending_above_type)
  pending_below_type <- match.arg(pending_below_type)
  .validate_symbol(symbol)
  .validate_positive_number(working_price, "working_price")
  .validate_positive_number(working_quantity, "working_quantity")
  .validate_positive_number(pending_quantity, "pending_quantity")
  .spot_validate_working_order_list_leg(
    type = working_type,
    time_in_force = working_time_in_force,
    iceberg_qty = working_iceberg_qty,
    strategy_type = working_strategy_type,
    peg_price_type = working_peg_price_type,
    peg_offset_type = working_peg_offset_type,
    peg_offset_value = working_peg_offset_value,
    prefix = "working"
  )
  .spot_validate_order_list_leg(
    type = pending_above_type,
    price = pending_above_price,
    stop_price = pending_above_stop_price,
    trailing_delta = pending_above_trailing_delta,
    time_in_force = pending_above_time_in_force,
    iceberg_qty = pending_above_iceberg_qty,
    strategy_type = pending_above_strategy_type,
    peg_price_type = pending_above_peg_price_type,
    peg_offset_type = pending_above_peg_offset_type,
    peg_offset_value = pending_above_peg_offset_value,
    prefix = "pending_above"
  )
  .spot_validate_order_list_leg(
    type = pending_below_type,
    price = pending_below_price,
    stop_price = pending_below_stop_price,
    trailing_delta = pending_below_trailing_delta,
    time_in_force = pending_below_time_in_force,
    iceberg_qty = pending_below_iceberg_qty,
    strategy_type = pending_below_strategy_type,
    peg_price_type = pending_below_peg_price_type,
    peg_offset_type = pending_below_peg_offset_type,
    peg_offset_value = pending_below_peg_offset_value,
    prefix = "pending_below"
  )
  .spot_validate_optional_resp_type(new_order_resp_type)

  .request_signed(
    config,
    "/api/v3/orderList/otoco",
    params = list(
      symbol = symbol,
      listClientOrderId = list_client_order_id,
      newOrderRespType = new_order_resp_type,
      selfTradePreventionMode = self_trade_prevention_mode,
      workingType = working_type,
      workingSide = working_side,
      workingClientOrderId = working_client_order_id,
      workingPrice = working_price,
      workingQuantity = working_quantity,
      workingIcebergQty = working_iceberg_qty,
      workingTimeInForce = working_time_in_force,
      workingStrategyId = working_strategy_id,
      workingStrategyType = working_strategy_type,
      workingPegPriceType = working_peg_price_type,
      workingPegOffsetType = working_peg_offset_type,
      workingPegOffsetValue = working_peg_offset_value,
      pendingSide = pending_side,
      pendingQuantity = pending_quantity,
      pendingAboveType = pending_above_type,
      pendingAboveClientOrderId = pending_above_client_order_id,
      pendingAbovePrice = pending_above_price,
      pendingAboveStopPrice = pending_above_stop_price,
      pendingAboveTrailingDelta = pending_above_trailing_delta,
      pendingAboveIcebergQty = pending_above_iceberg_qty,
      pendingAboveTimeInForce = pending_above_time_in_force,
      pendingAboveStrategyId = pending_above_strategy_id,
      pendingAboveStrategyType = pending_above_strategy_type,
      pendingAbovePegPriceType = pending_above_peg_price_type,
      pendingAbovePegOffsetType = pending_above_peg_offset_type,
      pendingAbovePegOffsetValue = pending_above_peg_offset_value,
      pendingBelowType = pending_below_type,
      pendingBelowClientOrderId = pending_below_client_order_id,
      pendingBelowPrice = pending_below_price,
      pendingBelowStopPrice = pending_below_stop_price,
      pendingBelowTrailingDelta = pending_below_trailing_delta,
      pendingBelowIcebergQty = pending_below_iceberg_qty,
      pendingBelowTimeInForce = pending_below_time_in_force,
      pendingBelowStrategyId = pending_below_strategy_id,
      pendingBelowStrategyType = pending_below_strategy_type,
      pendingBelowPegPriceType = pending_below_peg_price_type,
      pendingBelowPegOffsetType = pending_below_peg_offset_type,
      pendingBelowPegOffsetValue = pending_below_peg_offset_value
    ),
    method = "POST"
  )
}

#' Cancel a Binance Spot order list
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param order_list_id Optional exchange order-list ID.
#' @param list_client_order_id Optional order-list client ID.
#' @param new_client_order_id Optional client order ID for the cancel request.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_cancel_order_list <- function(
    symbol,
    order_list_id = NULL,
    list_client_order_id = NULL,
    new_client_order_id = NULL,
    config = config_spot()) {
  .validate_symbol(symbol)
  .validate_required_any(order_list_id, list_client_order_id, arg_names = c("order_list_id", "list_client_order_id"))
  .validate_optional_scalar_numeric(order_list_id, "order_list_id")
  .validate_optional_scalar_character(list_client_order_id, "list_client_order_id")
  .validate_optional_scalar_character(new_client_order_id, "new_client_order_id")

  .request_signed(
    config,
    "/api/v3/orderList",
    params = list(
      symbol = symbol,
      orderListId = order_list_id,
      listClientOrderId = list_client_order_id,
      newClientOrderId = new_client_order_id
    ),
    method = "DELETE"
  )
}

#' Place a Binance Spot OPO order list
#'
#' @inheritParams spot_place_order_list_oto
#'
#' @return A parsed list.
#' @export
spot_place_order_list_opo <- function(
    symbol,
    list_client_order_id = NULL,
    new_order_resp_type = NULL,
    self_trade_prevention_mode = NULL,
    working_type = c("LIMIT", "LIMIT_MAKER"),
    working_side = c("BUY", "SELL"),
    working_client_order_id = NULL,
    working_price,
    working_quantity,
    working_iceberg_qty = NULL,
    working_time_in_force = NULL,
    working_strategy_id = NULL,
    working_strategy_type = NULL,
    working_peg_price_type = NULL,
    working_peg_offset_type = NULL,
    working_peg_offset_value = NULL,
    pending_type = c("LIMIT", "MARKET", "STOP_LOSS", "STOP_LOSS_LIMIT", "TAKE_PROFIT", "TAKE_PROFIT_LIMIT", "LIMIT_MAKER"),
    pending_side = c("BUY", "SELL"),
    pending_client_order_id = NULL,
    pending_price = NULL,
    pending_stop_price = NULL,
    pending_trailing_delta = NULL,
    pending_iceberg_qty = NULL,
    pending_time_in_force = NULL,
    pending_strategy_id = NULL,
    pending_strategy_type = NULL,
    pending_peg_price_type = NULL,
    pending_peg_offset_type = NULL,
    pending_peg_offset_value = NULL,
    config = config_spot()) {
  working_type <- match.arg(working_type)
  working_side <- match.arg(working_side)
  pending_type <- match.arg(pending_type)
  pending_side <- match.arg(pending_side)
  .validate_symbol(symbol)
  .validate_positive_number(working_price, "working_price")
  .validate_positive_number(working_quantity, "working_quantity")
  .spot_validate_working_order_list_leg(
    type = working_type,
    time_in_force = working_time_in_force,
    iceberg_qty = working_iceberg_qty,
    strategy_type = working_strategy_type,
    peg_price_type = working_peg_price_type,
    peg_offset_type = working_peg_offset_type,
    peg_offset_value = working_peg_offset_value,
    prefix = "working"
  )
  .spot_validate_order_list_leg(
    type = pending_type,
    price = pending_price,
    stop_price = pending_stop_price,
    trailing_delta = pending_trailing_delta,
    time_in_force = pending_time_in_force,
    iceberg_qty = pending_iceberg_qty,
    strategy_type = pending_strategy_type,
    peg_price_type = pending_peg_price_type,
    peg_offset_type = pending_peg_offset_type,
    peg_offset_value = pending_peg_offset_value,
    prefix = "pending",
    allow_market = TRUE
  )
  .spot_validate_optional_resp_type(new_order_resp_type)

  .request_signed(
    config,
    "/api/v3/orderList/opo",
    params = list(
      symbol = symbol,
      listClientOrderId = list_client_order_id,
      newOrderRespType = new_order_resp_type,
      selfTradePreventionMode = self_trade_prevention_mode,
      workingType = working_type,
      workingSide = working_side,
      workingClientOrderId = working_client_order_id,
      workingPrice = working_price,
      workingQuantity = working_quantity,
      workingIcebergQty = working_iceberg_qty,
      workingTimeInForce = working_time_in_force,
      workingStrategyId = working_strategy_id,
      workingStrategyType = working_strategy_type,
      workingPegPriceType = working_peg_price_type,
      workingPegOffsetType = working_peg_offset_type,
      workingPegOffsetValue = working_peg_offset_value,
      pendingType = pending_type,
      pendingSide = pending_side,
      pendingClientOrderId = pending_client_order_id,
      pendingPrice = pending_price,
      pendingStopPrice = pending_stop_price,
      pendingTrailingDelta = pending_trailing_delta,
      pendingIcebergQty = pending_iceberg_qty,
      pendingTimeInForce = pending_time_in_force,
      pendingStrategyId = pending_strategy_id,
      pendingStrategyType = pending_strategy_type,
      pendingPegPriceType = pending_peg_price_type,
      pendingPegOffsetType = pending_peg_offset_type,
      pendingPegOffsetValue = pending_peg_offset_value
    ),
    method = "POST"
  )
}

#' Place a Binance Spot OPOCO order list
#'
#' @inheritParams spot_place_order_list_otoco
#'
#' @return A parsed list.
#' @export
spot_place_order_list_opoco <- function(
    symbol,
    list_client_order_id = NULL,
    new_order_resp_type = NULL,
    self_trade_prevention_mode = NULL,
    working_type = c("LIMIT", "LIMIT_MAKER"),
    working_side = c("BUY", "SELL"),
    working_client_order_id = NULL,
    working_price,
    working_quantity,
    working_iceberg_qty = NULL,
    working_time_in_force = NULL,
    working_strategy_id = NULL,
    working_strategy_type = NULL,
    working_peg_price_type = NULL,
    working_peg_offset_type = NULL,
    working_peg_offset_value = NULL,
    pending_side = c("BUY", "SELL"),
    pending_above_type = c("STOP_LOSS_LIMIT", "STOP_LOSS", "LIMIT_MAKER", "TAKE_PROFIT", "TAKE_PROFIT_LIMIT"),
    pending_above_client_order_id = NULL,
    pending_above_price = NULL,
    pending_above_stop_price = NULL,
    pending_above_trailing_delta = NULL,
    pending_above_iceberg_qty = NULL,
    pending_above_time_in_force = NULL,
    pending_above_strategy_id = NULL,
    pending_above_strategy_type = NULL,
    pending_above_peg_price_type = NULL,
    pending_above_peg_offset_type = NULL,
    pending_above_peg_offset_value = NULL,
    pending_below_type = c("STOP_LOSS", "STOP_LOSS_LIMIT", "TAKE_PROFIT", "TAKE_PROFIT_LIMIT"),
    pending_below_client_order_id = NULL,
    pending_below_price = NULL,
    pending_below_stop_price = NULL,
    pending_below_trailing_delta = NULL,
    pending_below_iceberg_qty = NULL,
    pending_below_time_in_force = NULL,
    pending_below_strategy_id = NULL,
    pending_below_strategy_type = NULL,
    pending_below_peg_price_type = NULL,
    pending_below_peg_offset_type = NULL,
    pending_below_peg_offset_value = NULL,
    config = config_spot()) {
  working_type <- match.arg(working_type)
  working_side <- match.arg(working_side)
  pending_side <- match.arg(pending_side)
  pending_above_type <- match.arg(pending_above_type)
  pending_below_type <- match.arg(pending_below_type)
  .validate_symbol(symbol)
  .validate_positive_number(working_price, "working_price")
  .validate_positive_number(working_quantity, "working_quantity")
  .spot_validate_working_order_list_leg(
    type = working_type,
    time_in_force = working_time_in_force,
    iceberg_qty = working_iceberg_qty,
    strategy_type = working_strategy_type,
    peg_price_type = working_peg_price_type,
    peg_offset_type = working_peg_offset_type,
    peg_offset_value = working_peg_offset_value,
    prefix = "working"
  )
  .spot_validate_order_list_leg(
    type = pending_above_type,
    price = pending_above_price,
    stop_price = pending_above_stop_price,
    trailing_delta = pending_above_trailing_delta,
    time_in_force = pending_above_time_in_force,
    iceberg_qty = pending_above_iceberg_qty,
    strategy_type = pending_above_strategy_type,
    peg_price_type = pending_above_peg_price_type,
    peg_offset_type = pending_above_peg_offset_type,
    peg_offset_value = pending_above_peg_offset_value,
    prefix = "pending_above"
  )
  .spot_validate_order_list_leg(
    type = pending_below_type,
    price = pending_below_price,
    stop_price = pending_below_stop_price,
    trailing_delta = pending_below_trailing_delta,
    time_in_force = pending_below_time_in_force,
    iceberg_qty = pending_below_iceberg_qty,
    strategy_type = pending_below_strategy_type,
    peg_price_type = pending_below_peg_price_type,
    peg_offset_type = pending_below_peg_offset_type,
    peg_offset_value = pending_below_peg_offset_value,
    prefix = "pending_below"
  )
  .spot_validate_optional_resp_type(new_order_resp_type)

  .request_signed(
    config,
    "/api/v3/orderList/opoco",
    params = list(
      symbol = symbol,
      listClientOrderId = list_client_order_id,
      newOrderRespType = new_order_resp_type,
      selfTradePreventionMode = self_trade_prevention_mode,
      workingType = working_type,
      workingSide = working_side,
      workingClientOrderId = working_client_order_id,
      workingPrice = working_price,
      workingQuantity = working_quantity,
      workingIcebergQty = working_iceberg_qty,
      workingTimeInForce = working_time_in_force,
      workingStrategyId = working_strategy_id,
      workingStrategyType = working_strategy_type,
      workingPegPriceType = working_peg_price_type,
      workingPegOffsetType = working_peg_offset_type,
      workingPegOffsetValue = working_peg_offset_value,
      pendingSide = pending_side,
      pendingAboveType = pending_above_type,
      pendingAboveClientOrderId = pending_above_client_order_id,
      pendingAbovePrice = pending_above_price,
      pendingAboveStopPrice = pending_above_stop_price,
      pendingAboveTrailingDelta = pending_above_trailing_delta,
      pendingAboveIcebergQty = pending_above_iceberg_qty,
      pendingAboveTimeInForce = pending_above_time_in_force,
      pendingAboveStrategyId = pending_above_strategy_id,
      pendingAboveStrategyType = pending_above_strategy_type,
      pendingAbovePegPriceType = pending_above_peg_price_type,
      pendingAbovePegOffsetType = pending_above_peg_offset_type,
      pendingAbovePegOffsetValue = pending_above_peg_offset_value,
      pendingBelowType = pending_below_type,
      pendingBelowClientOrderId = pending_below_client_order_id,
      pendingBelowPrice = pending_below_price,
      pendingBelowStopPrice = pending_below_stop_price,
      pendingBelowTrailingDelta = pending_below_trailing_delta,
      pendingBelowIcebergQty = pending_below_iceberg_qty,
      pendingBelowTimeInForce = pending_below_time_in_force,
      pendingBelowStrategyId = pending_below_strategy_id,
      pendingBelowStrategyType = pending_below_strategy_type,
      pendingBelowPegPriceType = pending_below_peg_price_type,
      pendingBelowPegOffsetType = pending_below_peg_offset_type,
      pendingBelowPegOffsetValue = pending_below_peg_offset_value
    ),
    method = "POST"
  )
}

#' Place a Binance Spot SOR order
#'
#' @param symbol Trading pair symbol, for example `"BTCUSDT"`.
#' @param side One of `"BUY"` or `"SELL"`.
#' @param type SOR order type. One of `"LIMIT"` or `"MARKET"`.
#' @param quantity Order quantity.
#' @param time_in_force Optional time-in-force value.
#' @param price Optional limit price.
#' @param new_client_order_id Optional client order identifier.
#' @param strategy_id Optional numeric strategy identifier.
#' @param strategy_type Optional numeric strategy type. Must be at least `1000000`.
#' @param iceberg_qty Optional iceberg quantity.
#' @param new_order_resp_type Optional response type. One of `"ACK"`, `"RESULT"`, or `"FULL"`.
#' @param self_trade_prevention_mode Optional self-trade prevention mode.
#' @param config A spot configuration created by [config_spot()].
#'
#' @return A parsed list.
#' @export
spot_place_sor_order <- function(
    symbol,
    side = c("BUY", "SELL"),
    type = c("LIMIT", "MARKET"),
    quantity,
    time_in_force = NULL,
    price = NULL,
    new_client_order_id = NULL,
    strategy_id = NULL,
    strategy_type = NULL,
    iceberg_qty = NULL,
    new_order_resp_type = NULL,
    self_trade_prevention_mode = NULL,
    config = config_spot()) {
  side <- match.arg(side)
  type <- match.arg(type)
  .validate_symbol(symbol)
  .validate_positive_number(quantity, "quantity")
  .validate_optional_scalar_character(time_in_force, "time_in_force")
  .validate_optional_scalar_numeric(price, "price")
  .validate_optional_scalar_character(new_client_order_id, "new_client_order_id")
  .validate_optional_scalar_numeric(strategy_id, "strategy_id")
  .validate_optional_scalar_numeric(strategy_type, "strategy_type")
  .validate_optional_scalar_numeric(iceberg_qty, "iceberg_qty")
  .validate_optional_scalar_character(new_order_resp_type, "new_order_resp_type")
  .validate_optional_scalar_character(self_trade_prevention_mode, "self_trade_prevention_mode")

  .spot_validate_sor_order_request(
    type = type,
    time_in_force = time_in_force,
    price = price,
    strategy_type = strategy_type,
    iceberg_qty = iceberg_qty,
    new_order_resp_type = new_order_resp_type
  )

  .request_signed(
    config,
    "/api/v3/sor/order",
    params = list(
      symbol = symbol,
      side = side,
      type = type,
      quantity = quantity,
      timeInForce = time_in_force,
      price = price,
      newClientOrderId = new_client_order_id,
      strategyId = strategy_id,
      strategyType = strategy_type,
      icebergQty = iceberg_qty,
      newOrderRespType = new_order_resp_type,
      selfTradePreventionMode = self_trade_prevention_mode
    ),
    method = "POST"
  )
}

#' Test a Binance Spot SOR order
#'
#' @inheritParams spot_place_sor_order
#' @param compute_commission_rates Whether to return commission-rate details.
#'
#' @return A parsed list.
#' @export
spot_test_sor_order <- function(
    symbol,
    side = c("BUY", "SELL"),
    type = c("LIMIT", "MARKET"),
    quantity,
    time_in_force = NULL,
    price = NULL,
    new_client_order_id = NULL,
    strategy_id = NULL,
    strategy_type = NULL,
    iceberg_qty = NULL,
    new_order_resp_type = NULL,
    self_trade_prevention_mode = NULL,
    compute_commission_rates = FALSE,
    config = config_spot()) {
  .validate_scalar_logical(compute_commission_rates, "compute_commission_rates")

  side <- match.arg(side)
  type <- match.arg(type)
  .validate_symbol(symbol)
  .validate_positive_number(quantity, "quantity")
  .validate_optional_scalar_character(time_in_force, "time_in_force")
  .validate_optional_scalar_numeric(price, "price")
  .validate_optional_scalar_character(new_client_order_id, "new_client_order_id")
  .validate_optional_scalar_numeric(strategy_id, "strategy_id")
  .validate_optional_scalar_numeric(strategy_type, "strategy_type")
  .validate_optional_scalar_numeric(iceberg_qty, "iceberg_qty")
  .validate_optional_scalar_character(new_order_resp_type, "new_order_resp_type")
  .validate_optional_scalar_character(self_trade_prevention_mode, "self_trade_prevention_mode")

  .spot_validate_sor_order_request(
    type = type,
    time_in_force = time_in_force,
    price = price,
    strategy_type = strategy_type,
    iceberg_qty = iceberg_qty,
    new_order_resp_type = new_order_resp_type
  )

  .request_signed(
    config,
    "/api/v3/sor/order/test",
    params = list(
      symbol = symbol,
      side = side,
      type = type,
      quantity = quantity,
      timeInForce = time_in_force,
      price = price,
      newClientOrderId = new_client_order_id,
      strategyId = strategy_id,
      strategyType = strategy_type,
      icebergQty = iceberg_qty,
      newOrderRespType = new_order_resp_type,
      selfTradePreventionMode = self_trade_prevention_mode,
      computeCommissionRates = if (isTRUE(compute_commission_rates)) "true" else "false"
    ),
    method = "POST"
  )
}

#' @noRd
.spot_validate_order_request <- function(
    type,
    time_in_force,
    quantity,
    quote_order_qty,
    price,
    strategy_type,
    stop_price,
    trailing_delta,
    iceberg_qty,
    new_order_resp_type,
    peg_price_type,
    peg_offset_value,
    peg_offset_type) {
  if (!is.null(quantity) && quantity <= 0) {
    stop("`quantity` must be positive when provided.", call. = FALSE)
  }
  if (!is.null(quote_order_qty) && quote_order_qty <= 0) {
    stop("`quote_order_qty` must be positive when provided.", call. = FALSE)
  }
  if (!is.null(price) && price <= 0) {
    stop("`price` must be positive when provided.", call. = FALSE)
  }
  if (!is.null(stop_price) && stop_price <= 0) {
    stop("`stop_price` must be positive when provided.", call. = FALSE)
  }
  if (!is.null(trailing_delta) && trailing_delta <= 0) {
    stop("`trailing_delta` must be positive when provided.", call. = FALSE)
  }
  if (!is.null(iceberg_qty) && iceberg_qty <= 0) {
    stop("`iceberg_qty` must be positive when provided.", call. = FALSE)
  }
  if (!is.null(strategy_type) && strategy_type < 1000000) {
    stop("`strategy_type` must be at least 1000000 when provided.", call. = FALSE)
  }

  if (!is.null(new_order_resp_type)) {
    .validate_one_of(new_order_resp_type, c("ACK", "RESULT", "FULL"), "new_order_resp_type")
  }

  if (!is.null(time_in_force)) {
    .validate_one_of(time_in_force, c("GTC", "IOC", "FOK"), "time_in_force")
  }

  peg_allowed_types <- c("LIMIT", "LIMIT_MAKER", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT")
  if (!is.null(peg_price_type)) {
    .validate_one_of(peg_price_type, c("PRIMARY_PEG", "MARKET_PEG"), "peg_price_type")
    if (!(type %in% peg_allowed_types)) {
      stop(
        "`peg_price_type` must only be supplied for LIMIT, LIMIT_MAKER, STOP_LOSS_LIMIT, or TAKE_PROFIT_LIMIT orders.",
        call. = FALSE
      )
    }
  }
  if (!is.null(peg_offset_type)) {
    .validate_one_of(peg_offset_type, c("PRICE_LEVEL"), "peg_offset_type")
  }
  if (xor(is.null(peg_offset_type), is.null(peg_offset_value))) {
    stop("`peg_offset_type` and `peg_offset_value` must be supplied together.", call. = FALSE)
  }
  if (!is.null(peg_offset_value) && (peg_offset_value <= 0 || peg_offset_value != as.integer(peg_offset_value))) {
    stop("`peg_offset_value` must be a positive whole number when provided.", call. = FALSE)
  }
  if (!is.null(peg_offset_value) && is.null(peg_price_type)) {
    stop("`peg_offset_type` and `peg_offset_value` require `peg_price_type`.", call. = FALSE)
  }

  if (!is.null(iceberg_qty)) {
    iceberg_allowed_types <- c("LIMIT", "LIMIT_MAKER", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT")
    if (!(type %in% iceberg_allowed_types)) {
      stop(
        "`iceberg_qty` must only be supplied for LIMIT, LIMIT_MAKER, STOP_LOSS_LIMIT, or TAKE_PROFIT_LIMIT orders.",
        call. = FALSE
      )
    }
    if (is.null(time_in_force) || !identical(time_in_force, "GTC")) {
      stop("`iceberg_qty` requires `time_in_force = \"GTC\"`.", call. = FALSE)
    }
  }

  if (identical(type, "MARKET")) {
    .validate_required_any(quantity, quote_order_qty, arg_names = c("quantity", "quote_order_qty"))
    .validate_at_most_one(quantity, quote_order_qty, arg_names = c("quantity", "quote_order_qty"))
    if (!is.null(time_in_force)) {
      stop("`time_in_force` must not be supplied for MARKET orders.", call. = FALSE)
    }
    if (!is.null(price)) {
      stop("`price` must not be supplied for MARKET orders.", call. = FALSE)
    }
  }

  price_required <- type %in% c("LIMIT", "LIMIT_MAKER", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT") && is.null(peg_price_type)
  if (isTRUE(price_required) && is.null(price)) {
    stop(sprintf("`price` is required when `type = \"%s\"` unless `peg_price_type` is supplied.", type), call. = FALSE)
  }

  if (identical(type, "LIMIT")) {
    if (is.null(quantity)) {
      stop("`quantity` is required when `type = \"LIMIT\"`.", call. = FALSE)
    }
    if (is.null(time_in_force)) {
      stop("`time_in_force` is required when `type = \"LIMIT\"`.", call. = FALSE)
    }
    if (!is.null(quote_order_qty)) {
      stop("`quote_order_qty` must not be supplied when `type = \"LIMIT\"`.", call. = FALSE)
    }
    return(invisible(TRUE))
  }

  if (identical(type, "LIMIT_MAKER")) {
    if (is.null(quantity)) {
      stop("`quantity` is required when `type = \"LIMIT_MAKER\"`.", call. = FALSE)
    }
    if (!is.null(time_in_force)) {
      stop("`time_in_force` must not be supplied when `type = \"LIMIT_MAKER\"`.", call. = FALSE)
    }
    if (!is.null(quote_order_qty)) {
      stop("`quote_order_qty` must not be supplied when `type = \"LIMIT_MAKER\"`.", call. = FALSE)
    }
    if (!is.null(stop_price) || !is.null(trailing_delta)) {
      stop("`stop_price` and `trailing_delta` must not be supplied when `type = \"LIMIT_MAKER\"`.", call. = FALSE)
    }
    return(invisible(TRUE))
  }

  if (type %in% c("STOP_LOSS", "TAKE_PROFIT")) {
    if (is.null(quantity)) {
      stop(sprintf("`quantity` is required when `type = \"%s\"`.", type), call. = FALSE)
    }
    .validate_required_any(stop_price, trailing_delta, arg_names = c("stop_price", "trailing_delta"))
    if (!is.null(price)) {
      stop(sprintf("`price` must not be supplied when `type = \"%s\"`.", type), call. = FALSE)
    }
    if (!is.null(time_in_force)) {
      stop(sprintf("`time_in_force` must not be supplied when `type = \"%s\"`.", type), call. = FALSE)
    }
    if (!is.null(quote_order_qty)) {
      stop(sprintf("`quote_order_qty` must not be supplied when `type = \"%s\"`.", type), call. = FALSE)
    }
    return(invisible(TRUE))
  }

  if (type %in% c("STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT")) {
    if (is.null(quantity)) {
      stop(sprintf("`quantity` is required when `type = \"%s\"`.", type), call. = FALSE)
    }
    if (is.null(time_in_force)) {
      stop(sprintf("`time_in_force` is required when `type = \"%s\"`.", type), call. = FALSE)
    }
    .validate_required_any(stop_price, trailing_delta, arg_names = c("stop_price", "trailing_delta"))
    if (!is.null(quote_order_qty)) {
      stop(sprintf("`quote_order_qty` must not be supplied when `type = \"%s\"`.", type), call. = FALSE)
    }
    return(invisible(TRUE))
  }

  invisible(TRUE)
}

#' @noRd
.spot_validate_strategy_type <- function(x, arg) {
  if (is.null(x)) {
    return(invisible(x))
  }
  if (x < 1000000) {
    stop(sprintf("`%s` must be at least 1000000 when provided.", arg), call. = FALSE)
  }
  invisible(x)
}

#' @noRd
.spot_validate_positive_optional <- function(x, arg) {
  if (is.null(x)) {
    return(invisible(x))
  }
  if (!is.numeric(x) || length(x) != 1L || is.na(x) || x <= 0) {
    stop(sprintf("`%s` must be positive when provided.", arg), call. = FALSE)
  }
  invisible(x)
}

#' @noRd
.spot_validate_optional_resp_type <- function(x) {
  if (is.null(x)) {
    return(invisible(x))
  }
  .validate_one_of(x, c("ACK", "RESULT", "FULL"), "new_order_resp_type")
}

#' @noRd
.spot_validate_optional_tif <- function(x, arg) {
  if (is.null(x)) {
    return(invisible(x))
  }
  .validate_one_of(x, c("GTC", "IOC", "FOK"), arg)
}

#' @noRd
.spot_validate_peg_fields <- function(peg_price_type, peg_offset_type, peg_offset_value, prefix, allowed_types, type) {
  peg_price_arg <- paste0(prefix, "_peg_price_type")
  peg_offset_type_arg <- paste0(prefix, "_peg_offset_type")
  peg_offset_value_arg <- paste0(prefix, "_peg_offset_value")
  if (!is.null(peg_price_type)) {
    .validate_one_of(peg_price_type, c("PRIMARY_PEG", "MARKET_PEG"), peg_price_arg)
    if (!(type %in% allowed_types)) {
      stop(
        sprintf("`%s` must only be supplied for supported pegged order types.", peg_price_arg),
        call. = FALSE
      )
    }
  }
  if (!is.null(peg_offset_type)) {
    .validate_one_of(peg_offset_type, c("PRICE_LEVEL"), peg_offset_type_arg)
  }
  if (xor(is.null(peg_offset_type), is.null(peg_offset_value))) {
    stop(sprintf("`%s` and `%s` must be supplied together.", peg_offset_type_arg, peg_offset_value_arg), call. = FALSE)
  }
  if (!is.null(peg_offset_value)) {
    if (!is.numeric(peg_offset_value) || length(peg_offset_value) != 1L || is.na(peg_offset_value) ||
        peg_offset_value <= 0 || peg_offset_value != as.integer(peg_offset_value)) {
      stop(sprintf("`%s` must be a positive whole number when provided.", peg_offset_value_arg), call. = FALSE)
    }
    if (is.null(peg_price_type)) {
      stop(sprintf("`%s` requires `%s`.", peg_offset_value_arg, peg_price_arg), call. = FALSE)
    }
  }
  invisible(TRUE)
}

#' @noRd
.spot_validate_order_list_leg <- function(
    type,
    price,
    stop_price,
    trailing_delta,
    time_in_force,
    iceberg_qty,
    strategy_type,
    peg_price_type,
    peg_offset_type,
    peg_offset_value,
    prefix,
    allow_market = FALSE) {
  allowed_types <- c("STOP_LOSS", "STOP_LOSS_LIMIT", "LIMIT_MAKER", "TAKE_PROFIT", "TAKE_PROFIT_LIMIT")
  if (isTRUE(allow_market)) {
    allowed_types <- c("LIMIT", "MARKET", allowed_types)
  }
  if (!(type %in% allowed_types)) {
    stop(sprintf("Unsupported `%s_type` value `%s`.", prefix, type), call. = FALSE)
  }

  .validate_optional_scalar_numeric(price, paste0(prefix, "_price"))
  .validate_optional_scalar_numeric(stop_price, paste0(prefix, "_stop_price"))
  .validate_optional_scalar_numeric(trailing_delta, paste0(prefix, "_trailing_delta"))
  .validate_optional_scalar_character(time_in_force, paste0(prefix, "_time_in_force"))
  .validate_optional_scalar_numeric(iceberg_qty, paste0(prefix, "_iceberg_qty"))
  .spot_validate_strategy_type(strategy_type, paste0(prefix, "_strategy_type"))
  .spot_validate_positive_optional(price, paste0(prefix, "_price"))
  .spot_validate_positive_optional(stop_price, paste0(prefix, "_stop_price"))
  .spot_validate_positive_optional(trailing_delta, paste0(prefix, "_trailing_delta"))
  .spot_validate_positive_optional(iceberg_qty, paste0(prefix, "_iceberg_qty"))
  .spot_validate_optional_tif(time_in_force, paste0(prefix, "_time_in_force"))

  .spot_validate_peg_fields(
    peg_price_type = peg_price_type,
    peg_offset_type = peg_offset_type,
    peg_offset_value = peg_offset_value,
    prefix = prefix,
    allowed_types = c("LIMIT_MAKER", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT"),
    type = type
  )

  if (identical(type, "LIMIT")) {
    if (is.null(price)) {
      stop(sprintf("`%s_price` is required when `%s_type = \"LIMIT\"`.", prefix, prefix), call. = FALSE)
    }
    if (is.null(time_in_force)) {
      stop(sprintf("`%s_time_in_force` is required when `%s_type = \"LIMIT\"`.", prefix, prefix), call. = FALSE)
    }
  }

  if (identical(type, "MARKET")) {
    if (!is.null(price)) {
      stop(sprintf("`%s_price` must not be supplied when `%s_type = \"MARKET\"`.", prefix, prefix), call. = FALSE)
    }
    if (!is.null(time_in_force)) {
      stop(sprintf("`%s_time_in_force` must not be supplied when `%s_type = \"MARKET\"`.", prefix, prefix), call. = FALSE)
    }
  }

  if (type %in% c("STOP_LOSS", "TAKE_PROFIT")) {
    .validate_required_any(stop_price, trailing_delta, arg_names = c(paste0(prefix, "_stop_price"), paste0(prefix, "_trailing_delta")))
    if (!is.null(price)) {
      stop(sprintf("`%s_price` must not be supplied when `%s_type = \"%s\"`.", prefix, prefix, type), call. = FALSE)
    }
    if (!is.null(time_in_force)) {
      stop(sprintf("`%s_time_in_force` must not be supplied when `%s_type = \"%s\"`.", prefix, prefix, type), call. = FALSE)
    }
  }

  if (type %in% c("STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT")) {
    if (is.null(price)) {
      stop(sprintf("`%s_price` is required when `%s_type = \"%s\"`.", prefix, prefix, type), call. = FALSE)
    }
    .validate_required_any(stop_price, trailing_delta, arg_names = c(paste0(prefix, "_stop_price"), paste0(prefix, "_trailing_delta")))
    if (is.null(time_in_force)) {
      stop(sprintf("`%s_time_in_force` is required when `%s_type = \"%s\"`.", prefix, prefix, type), call. = FALSE)
    }
  }

  if (identical(type, "LIMIT_MAKER") && is.null(price)) {
    stop(sprintf("`%s_price` is required when `%s_type = \"LIMIT_MAKER\"`.", prefix, prefix), call. = FALSE)
  }

  if (!is.null(iceberg_qty)) {
    if (!(type %in% c("LIMIT_MAKER", "STOP_LOSS_LIMIT", "TAKE_PROFIT_LIMIT", "LIMIT"))) {
      stop(sprintf("`%s_iceberg_qty` is not supported for `%s_type = \"%s\"`.", prefix, prefix, type), call. = FALSE)
    }
    if (!identical(time_in_force, "GTC") && !identical(type, "LIMIT_MAKER")) {
      stop(sprintf("`%s_iceberg_qty` requires `%s_time_in_force = \"GTC\"`.", prefix, prefix), call. = FALSE)
    }
  }

  invisible(TRUE)
}

#' @noRd
.spot_validate_working_order_list_leg <- function(
    type,
    time_in_force,
    iceberg_qty,
    strategy_type,
    peg_price_type,
    peg_offset_type,
    peg_offset_value,
    prefix) {
  .spot_validate_strategy_type(strategy_type, paste0(prefix, "_strategy_type"))
  .spot_validate_positive_optional(iceberg_qty, paste0(prefix, "_iceberg_qty"))
  .spot_validate_optional_tif(time_in_force, paste0(prefix, "_time_in_force"))
  .spot_validate_peg_fields(
    peg_price_type = peg_price_type,
    peg_offset_type = peg_offset_type,
    peg_offset_value = peg_offset_value,
    prefix = prefix,
    allowed_types = c("LIMIT", "LIMIT_MAKER"),
    type = type
  )

  if (identical(type, "LIMIT") && is.null(time_in_force)) {
    stop(sprintf("`%s_time_in_force` is required when `%s_type = \"LIMIT\"`.", prefix, prefix), call. = FALSE)
  }
  if (!is.null(iceberg_qty) && !identical(time_in_force, "GTC") && !identical(type, "LIMIT_MAKER")) {
    stop(sprintf("`%s_iceberg_qty` requires `%s_time_in_force = \"GTC\"`.", prefix, prefix), call. = FALSE)
  }
  invisible(TRUE)
}

#' @noRd
.spot_validate_sor_order_request <- function(
    type,
    time_in_force,
    price,
    strategy_type,
    iceberg_qty,
    new_order_resp_type) {
  if (!is.null(strategy_type) && strategy_type < 1000000) {
    stop("`strategy_type` must be at least 1000000 when provided.", call. = FALSE)
  }
  .spot_validate_optional_resp_type(new_order_resp_type)

  if (identical(type, "LIMIT")) {
    if (is.null(price)) {
      stop("`price` is required when `type = \"LIMIT\"`.", call. = FALSE)
    }
    if (is.null(time_in_force)) {
      stop("`time_in_force` is required when `type = \"LIMIT\"`.", call. = FALSE)
    }
    .validate_one_of(time_in_force, c("GTC", "IOC", "FOK"), "time_in_force")
  } else {
    if (!is.null(time_in_force)) {
      stop("`time_in_force` must not be supplied when `type = \"MARKET\"`.", call. = FALSE)
    }
    if (!is.null(price)) {
      stop("`price` must not be supplied when `type = \"MARKET\"`.", call. = FALSE)
    }
  }

  if (!is.null(iceberg_qty)) {
    if (iceberg_qty <= 0) {
      stop("`iceberg_qty` must be positive when provided.", call. = FALSE)
    }
    if (!identical(type, "LIMIT")) {
      stop("`iceberg_qty` must only be supplied when `type = \"LIMIT\"`.", call. = FALSE)
    }
    if (!identical(time_in_force, "GTC")) {
      stop("`iceberg_qty` requires `time_in_force = \"GTC\"`.", call. = FALSE)
    }
  }

  invisible(TRUE)
}

test_that("spot place order validates market quantity selectors", {
  cfg <- config_spot(api_key = "k", secret_key = "s")

  expect_error(
    spot_place_order(
      symbol = "BTCUSDT",
      side = "BUY",
      type = "MARKET",
      config = cfg
    ),
    "Provide at least one of: quantity, quote_order_qty"
  )

  expect_error(
    spot_place_order(
      symbol = "BTCUSDT",
      side = "BUY",
      type = "MARKET",
      quantity = 1,
      quote_order_qty = 100,
      config = cfg
    ),
    "Only one of quantity, quote_order_qty may be supplied"
  )

  expect_error(
    spot_place_order(
      symbol = "BTCUSDT",
      side = "BUY",
      type = "MARKET",
      quantity = 1,
      time_in_force = "GTC",
      config = cfg
    ),
    "time_in_force"
  )
})

test_that("spot place order validates conditional limit and stop order fields", {
  cfg <- config_spot(api_key = "k", secret_key = "s")

  expect_error(
    spot_place_order(
      symbol = "BTCUSDT",
      side = "BUY",
      type = "LIMIT",
      quantity = 1,
      price = 100,
      config = cfg
    ),
    "time_in_force"
  )

  expect_error(
    spot_place_order(
      symbol = "BTCUSDT",
      side = "SELL",
      type = "STOP_LOSS",
      quantity = 1,
      config = cfg
    ),
    "Provide at least one of: stop_price, trailing_delta"
  )

  expect_error(
    spot_place_order(
      symbol = "BTCUSDT",
      side = "SELL",
      type = "LIMIT_MAKER",
      quantity = 1,
      price = 100,
      stop_price = 90,
      config = cfg
    ),
    "stop_price"
  )
})

test_that("spot place order validates iceberg and pegged order constraints", {
  cfg <- config_spot(api_key = "k", secret_key = "s")

  expect_error(
    spot_place_order(
      symbol = "BTCUSDT",
      side = "BUY",
      type = "LIMIT",
      quantity = 1,
      price = 100,
      time_in_force = "IOC",
      iceberg_qty = 0.1,
      config = cfg
    ),
    "iceberg_qty"
  )

  expect_error(
    spot_place_order(
      symbol = "BTCUSDT",
      side = "BUY",
      type = "MARKET",
      quantity = 1,
      peg_price_type = "PRIMARY_PEG",
      config = cfg
    ),
    "peg_price_type"
  )

  expect_error(
    spot_place_order(
      symbol = "BTCUSDT",
      side = "BUY",
      type = "LIMIT",
      quantity = 1,
      time_in_force = "GTC",
      peg_price_type = "PRIMARY_PEG",
      peg_offset_value = 2,
      config = cfg
    ),
    "peg_offset_type"
  )
})

test_that("spot place order forwards signed params for valid requests", {
  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "DELETE")) {
      list(path = path, params = params, method = method)
    },
    .package = "binxr"
  )

  out <- spot_place_order(
    symbol = "BTCUSDT",
    side = "BUY",
    type = "LIMIT",
    quantity = 1,
    price = 100,
    time_in_force = "GTC",
    iceberg_qty = 0.2,
    new_order_resp_type = "FULL",
    config = config_spot(api_key = "k", secret_key = "s")
  )

  expect_identical(out$path, "/api/v3/order")
  expect_identical(out$method, "POST")
  expect_identical(out$params$symbol, "BTCUSDT")
  expect_identical(out$params$timeInForce, "GTC")
  expect_identical(out$params$icebergQty, 0.2)
  expect_identical(out$params$newOrderRespType, "FULL")
})

test_that("spot test order includes computeCommissionRates", {
  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "DELETE")) {
      list(path = path, params = params, method = method)
    },
    .package = "binxr"
  )

  out <- spot_test_order(
    symbol = "BTCUSDT",
    side = "BUY",
    type = "MARKET",
    quantity = 1,
    compute_commission_rates = TRUE,
    config = config_spot(api_key = "k", secret_key = "s")
  )

  expect_identical(out$path, "/api/v3/order/test")
  expect_identical(out$params$computeCommissionRates, "true")
})

test_that("spot cancel order validates selectors and restrictions", {
  cfg <- config_spot(api_key = "k", secret_key = "s")

  expect_error(
    spot_cancel_order(symbol = "BTCUSDT", config = cfg),
    "Provide at least one of: order_id, orig_client_order_id"
  )

  expect_error(
    spot_cancel_order(
      symbol = "BTCUSDT",
      order_id = 1,
      cancel_restrictions = "INVALID",
      config = cfg
    ),
    "cancel_restrictions"
  )
})

test_that("spot cancel order and cancel all orders call signed endpoints", {
  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "DELETE")) {
      list(path = path, params = params, method = method)
    },
    .package = "binxr"
  )

  cancel_out <- spot_cancel_order(
    symbol = "BTCUSDT",
    order_id = 10,
    new_client_order_id = "cancel-1",
    cancel_restrictions = "ONLY_NEW",
    config = config_spot(api_key = "k", secret_key = "s")
  )
  cancel_all_out <- spot_cancel_all_orders(
    symbol = "BTCUSDT",
    config = config_spot(api_key = "k", secret_key = "s")
  )

  expect_identical(cancel_out$path, "/api/v3/order")
  expect_identical(cancel_out$method, "DELETE")
  expect_identical(cancel_out$params$orderId, 10)
  expect_identical(cancel_out$params$cancelRestrictions, "ONLY_NEW")

  expect_identical(cancel_all_out$path, "/api/v3/openOrders")
  expect_identical(cancel_all_out$method, "DELETE")
  expect_identical(cancel_all_out$params$symbol, "BTCUSDT")
})

test_that("spot cancel-replace validates cancel selectors and forwards params", {
  cfg <- config_spot(api_key = "k", secret_key = "s")

  expect_error(
    spot_cancel_replace_order(
      symbol = "BTCUSDT",
      side = "BUY",
      type = "MARKET",
      cancel_replace_mode = "STOP_ON_FAILURE",
      quantity = 1,
      config = cfg
    ),
    "Provide at least one of: cancel_order_id, cancel_orig_client_order_id"
  )

  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "PUT", "DELETE")) {
      list(path = path, params = params, method = method)
    },
    .package = "binxr"
  )

  out <- spot_cancel_replace_order(
    symbol = "BTCUSDT",
    side = "BUY",
    type = "LIMIT",
    cancel_replace_mode = "ALLOW_FAILURE",
    cancel_order_id = 123,
    quantity = 1,
    price = 100,
    time_in_force = "GTC",
    order_rate_limit_exceeded_mode = "CANCEL_ONLY",
    config = cfg
  )

  expect_identical(out$path, "/api/v3/order/cancelReplace")
  expect_identical(out$method, "POST")
  expect_identical(out$params$cancelOrderId, 123)
  expect_identical(out$params$cancelReplaceMode, "ALLOW_FAILURE")
  expect_identical(out$params$orderRateLimitExceededMode, "CANCEL_ONLY")
})

test_that("spot amend keep priority validates selectors and uses PUT", {
  cfg <- config_spot(api_key = "k", secret_key = "s")

  expect_error(
    spot_amend_order_keep_priority(
      symbol = "BTCUSDT",
      new_qty = 1,
      config = cfg
    ),
    "Provide at least one of: order_id, orig_client_order_id"
  )

  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "PUT", "DELETE")) {
      list(path = path, params = params, method = method)
    },
    .package = "binxr"
  )

  out <- spot_amend_order_keep_priority(
    symbol = "BTCUSDT",
    order_id = 10,
    new_client_order_id = "amended-1",
    new_qty = 0.5,
    config = cfg
  )

  expect_identical(out$path, "/api/v3/order/amend/keepPriority")
  expect_identical(out$method, "PUT")
  expect_identical(out$params$newQty, 0.5)
})

test_that("spot order list oco validates conditional fields", {
  cfg <- config_spot(api_key = "k", secret_key = "s")

  expect_error(
    spot_place_order_list_oco(
      symbol = "BTCUSDT",
      side = "SELL",
      quantity = 1,
      above_type = "STOP_LOSS_LIMIT",
      below_type = "STOP_LOSS",
      below_stop_price = 90,
      config = cfg
    ),
    "above_price"
  )

  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "PUT", "DELETE")) {
      list(path = path, params = params, method = method)
    },
    .package = "binxr"
  )

  out <- spot_place_order_list_oco(
    symbol = "BTCUSDT",
    side = "SELL",
    quantity = 1,
    above_type = "TAKE_PROFIT_LIMIT",
    above_price = 110,
    above_stop_price = 108,
    above_time_in_force = "GTC",
    below_type = "STOP_LOSS_LIMIT",
    below_price = 90,
    below_stop_price = 92,
    below_time_in_force = "GTC",
    config = cfg
  )

  expect_identical(out$path, "/api/v3/orderList/oco")
  expect_identical(out$params$aboveType, "TAKE_PROFIT_LIMIT")
  expect_identical(out$params$belowType, "STOP_LOSS_LIMIT")
})

test_that("spot order list oto validates working and pending requirements", {
  cfg <- config_spot(api_key = "k", secret_key = "s")

  expect_error(
    spot_place_order_list_oto(
      symbol = "BTCUSDT",
      working_type = "LIMIT",
      working_side = "BUY",
      working_price = 100,
      working_quantity = 1,
      pending_type = "LIMIT",
      pending_side = "SELL",
      pending_quantity = 1,
      config = cfg
    ),
    "working_time_in_force"
  )

  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "PUT", "DELETE")) {
      list(path = path, params = params, method = method)
    },
    .package = "binxr"
  )

  out <- spot_place_order_list_oto(
    symbol = "BTCUSDT",
    working_type = "LIMIT",
    working_side = "BUY",
    working_price = 100,
    working_quantity = 1,
    working_time_in_force = "GTC",
    pending_type = "STOP_LOSS_LIMIT",
    pending_side = "SELL",
    pending_price = 90,
    pending_stop_price = 92,
    pending_quantity = 1,
    pending_time_in_force = "GTC",
    config = cfg
  )

  expect_identical(out$path, "/api/v3/orderList/oto")
  expect_identical(out$params$workingType, "LIMIT")
  expect_identical(out$params$pendingType, "STOP_LOSS_LIMIT")
})

test_that("spot order list otoco and cancel order list call signed endpoints", {
  cfg <- config_spot(api_key = "k", secret_key = "s")

  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "PUT", "DELETE")) {
      list(path = path, params = params, method = method)
    },
    .package = "binxr"
  )

  otoco_out <- spot_place_order_list_otoco(
    symbol = "BTCUSDT",
    working_type = "LIMIT",
    working_side = "BUY",
    working_price = 100,
    working_quantity = 1,
    working_time_in_force = "GTC",
    pending_side = "SELL",
    pending_quantity = 1,
    pending_above_type = "TAKE_PROFIT_LIMIT",
    pending_above_price = 110,
    pending_above_stop_price = 108,
    pending_above_time_in_force = "GTC",
    pending_below_type = "STOP_LOSS_LIMIT",
    pending_below_price = 90,
    pending_below_stop_price = 92,
    pending_below_time_in_force = "GTC",
    config = cfg
  )
  cancel_out <- spot_cancel_order_list(
    symbol = "BTCUSDT",
    order_list_id = 99,
    config = cfg
  )

  expect_identical(otoco_out$path, "/api/v3/orderList/otoco")
  expect_identical(otoco_out$params$pendingAboveType, "TAKE_PROFIT_LIMIT")
  expect_identical(cancel_out$path, "/api/v3/orderList")
  expect_identical(cancel_out$method, "DELETE")
  expect_identical(cancel_out$params$orderListId, 99)
})

test_that("spot opo and opoco order lists call signed endpoints", {
  cfg <- config_spot(api_key = "k", secret_key = "s")

  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "PUT", "DELETE")) {
      list(path = path, params = params, method = method)
    },
    .package = "binxr"
  )

  opo_out <- spot_place_order_list_opo(
    symbol = "BTCUSDT",
    working_type = "LIMIT",
    working_side = "BUY",
    working_price = 100,
    working_quantity = 1,
    working_time_in_force = "GTC",
    pending_type = "MARKET",
    pending_side = "SELL",
    config = cfg
  )
  opoco_out <- spot_place_order_list_opoco(
    symbol = "BTCUSDT",
    working_type = "LIMIT",
    working_side = "BUY",
    working_price = 100,
    working_quantity = 1,
    working_time_in_force = "GTC",
    pending_side = "SELL",
    pending_above_type = "TAKE_PROFIT_LIMIT",
    pending_above_price = 110,
    pending_above_stop_price = 108,
    pending_above_time_in_force = "GTC",
    pending_below_type = "STOP_LOSS_LIMIT",
    pending_below_price = 90,
    pending_below_stop_price = 92,
    pending_below_time_in_force = "GTC",
    config = cfg
  )

  expect_identical(opo_out$path, "/api/v3/orderList/opo")
  expect_identical(opo_out$params$pendingType, "MARKET")
  expect_identical(opoco_out$path, "/api/v3/orderList/opoco")
  expect_identical(opoco_out$params$pendingAboveType, "TAKE_PROFIT_LIMIT")
})

test_that("spot sor order validates type-specific fields and forwards params", {
  cfg <- config_spot(api_key = "k", secret_key = "s")

  expect_error(
    spot_place_sor_order(
      symbol = "BTCUSDT",
      side = "BUY",
      type = "LIMIT",
      quantity = 1,
      config = cfg
    ),
    "price"
  )

  expect_error(
    spot_place_sor_order(
      symbol = "BTCUSDT",
      side = "BUY",
      type = "MARKET",
      quantity = 1,
      price = 100,
      config = cfg
    ),
    "price"
  )

  local_mocked_bindings(
    .request_signed = function(config, path, params = NULL, method = c("GET", "POST", "PUT", "DELETE")) {
      list(path = path, params = params, method = method)
    },
    .package = "binxr"
  )

  place_out <- spot_place_sor_order(
    symbol = "BTCUSDT",
    side = "BUY",
    type = "LIMIT",
    quantity = 1,
    price = 100,
    time_in_force = "GTC",
    iceberg_qty = 0.1,
    config = cfg
  )
  test_out <- spot_test_sor_order(
    symbol = "BTCUSDT",
    side = "BUY",
    type = "MARKET",
    quantity = 1,
    compute_commission_rates = TRUE,
    config = cfg
  )

  expect_identical(place_out$path, "/api/v3/sor/order")
  expect_identical(place_out$params$icebergQty, 0.1)
  expect_identical(test_out$path, "/api/v3/sor/order/test")
  expect_identical(test_out$params$computeCommissionRates, "true")
})

#' binxr: Binance REST API Client for R
#'
#' Client helpers for the Binance Spot, Futures, and Options REST APIs.
#' The package provides configuration constructors, signed and unsigned
#' request helpers, market data retrieval, account access, and order management
#' functions. Tabular endpoint responses are returned as `data.table` objects
#' by default.
#'
#' Network examples and live trading calls are not run during package checks.
#' Authenticated endpoints require user-supplied Binance API credentials.
#'
#' @keywords internal
"_PACKAGE"

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-trade.R
\name{spot_amend_order_keep_priority}
\alias{spot_amend_order_keep_priority}
\title{Amend a Binance Spot order while keeping priority}
\usage{
spot_amend_order_keep_priority(
  symbol,
  order_id = NULL,
  orig_client_order_id = NULL,
  new_client_order_id = NULL,
  new_qty,
  config = config_spot()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"BTCUSDT"}.}

\item{order_id}{Optional exchange order ID.}

\item{orig_client_order_id}{Optional client order ID.}

\item{new_client_order_id}{Optional new client order ID after amendment.}

\item{new_qty}{New order quantity. Must be positive.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list.
}
\description{
Amend a Binance Spot order while keeping priority
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-trade.R
\name{futures_countdown_cancel_all}
\alias{futures_countdown_cancel_all}
\title{Set Binance Futures countdown cancel-all}
\usage{
futures_countdown_cancel_all(symbol, countdown_time, config = config_futures())
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"ETHUSDT"}.}

\item{countdown_time}{Countdown time in milliseconds.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}
}
\value{
A parsed list.
}
\description{
Set Binance Futures countdown cancel-all
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-trade.R
\name{futures_get_force_orders}
\alias{futures_get_force_orders}
\title{Get Binance Futures force orders}
\usage{
futures_get_force_orders(
  symbol = NULL,
  auto_close_type = NULL,
  startTime = NULL,
  endTime = NULL,
  limit = 50,
  json_list = FALSE,
  config = config_futures()
)
}
\arguments{
\item{symbol}{Optional trading pair symbol.}

\item{auto_close_type}{Optional auto-close type.}

\item{startTime}{Optional start time in milliseconds since Unix epoch.}

\item{endTime}{Optional end time in milliseconds since Unix epoch.}

\item{limit}{Maximum number of rows to return. Must not exceed \code{100}.}

\item{json_list}{If \code{TRUE}, return the parsed list instead of a \code{data.table}.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}
}
\value{
A \code{data.table} by default, or a parsed list when \code{json_list = TRUE}.
}
\description{
Get Binance Futures force orders
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-market.R
\name{spot_get_average_price}
\alias{spot_get_average_price}
\title{Get Binance Spot average price}
\usage{
spot_get_average_price(symbol, config = config_spot())
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"BTCUSDT"}.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list.
}
\description{
Get Binance Spot average price
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-market.R
\name{spot_get_order_book}
\alias{spot_get_order_book}
\title{Get Binance Spot order book}
\usage{
spot_get_order_book(
  symbol,
  limit = NULL,
  symbol_status = NULL,
  config = config_spot()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"BTCUSDT"}.}

\item{limit}{Optional order book depth limit.}

\item{symbol_status}{Optional trading status filter.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list.
}
\description{
Get Binance Spot order book
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-account.R
\name{spot_get_unfilled_order_count}
\alias{spot_get_unfilled_order_count}
\title{Get Binance Spot unfilled order counts}
\usage{
spot_get_unfilled_order_count(json_list = FALSE, config = config_spot())
}
\arguments{
\item{json_list}{If \code{TRUE}, return the parsed list instead of a \code{data.table}.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A \code{data.table} by default, or a parsed list when \code{json_list = TRUE}.
}
\description{
Get Binance Spot unfilled order counts
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-general.R
\name{spot_ping}
\alias{spot_ping}
\title{Test Binance Spot API connectivity}
\usage{
spot_ping(config = config_spot())
}
\arguments{
\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
An empty parsed list on success.
}
\description{
Test Binance Spot API connectivity
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-options-trade.R
\name{options_get_commission}
\alias{options_get_commission}
\title{Get Binance Options user commission}
\usage{
options_get_commission(config = config_options())
}
\arguments{
\item{config}{An options configuration created by \code{\link[=config_options]{config_options()}}.}
}
\value{
A parsed list.
}
\description{
Get Binance Options user commission
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-trade.R
\name{spot_place_sor_order}
\alias{spot_place_sor_order}
\title{Place a Binance Spot SOR order}
\usage{
spot_place_sor_order(
  symbol,
  side = c("BUY", "SELL"),
  type = c("LIMIT", "MARKET"),
  quantity,
  time_in_force = NULL,
  price = NULL,
  new_client_order_id = NULL,
  strategy_id = NULL,
  strategy_type = NULL,
  iceberg_qty = NULL,
  new_order_resp_type = NULL,
  self_trade_prevention_mode = NULL,
  config = config_spot()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"BTCUSDT"}.}

\item{side}{One of \code{"BUY"} or \code{"SELL"}.}

\item{type}{SOR order type. One of \code{"LIMIT"} or \code{"MARKET"}.}

\item{quantity}{Order quantity.}

\item{time_in_force}{Optional time-in-force value.}

\item{price}{Optional limit price.}

\item{new_client_order_id}{Optional client order identifier.}

\item{strategy_id}{Optional numeric strategy identifier.}

\item{strategy_type}{Optional numeric strategy type. Must be at least \code{1000000}.}

\item{iceberg_qty}{Optional iceberg quantity.}

\item{new_order_resp_type}{Optional response type. One of \code{"ACK"}, \code{"RESULT"}, or \code{"FULL"}.}

\item{self_trade_prevention_mode}{Optional self-trade prevention mode.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list.
}
\description{
Place a Binance Spot SOR order
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-trade.R
\name{futures_place_order}
\alias{futures_place_order}
\alias{place_fapi_trade_order}
\title{Place a Binance Futures order}
\usage{
futures_place_order(
  symbol,
  side = c("BUY", "SELL"),
  type = c("LIMIT", "MARKET", "STOP", "STOP_MARKET", "TAKE_PROFIT", "TAKE_PROFIT_MARKET",
    "TRAILING_STOP_MARKET"),
  quantity,
  price = NULL,
  time_in_force = NULL,
  position_side = c("BOTH", "LONG", "SHORT"),
  reduce_only = FALSE,
  stop_price = NULL,
  working_type = NULL,
  new_client_order_id = NULL,
  config = config_futures()
)

place_fapi_trade_order(
  symbol,
  side = c("BUY", "SELL"),
  type = c("LIMIT", "MARKET", "STOP", "STOP_MARKET", "TAKE_PROFIT", "TAKE_PROFIT_MARKET",
    "TRAILING_STOP_MARKET"),
  quantity,
  price = NULL,
  timeInForce = c("GTC", "IOC", "FOK", "GTX"),
  positionSide = c("BOTH", "LONG", "SHORT"),
  reduceOnly = FALSE,
  stopPrice = NULL,
  workingType = c("CONTRACT_PRICE", "MARK_PRICE"),
  newClientOrderId = NULL,
  config = binxr_config_futures()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"ETHUSDT"}.}

\item{side}{One of \code{"BUY"} or \code{"SELL"}.}

\item{type}{Order type.}

\item{quantity}{Order quantity.}

\item{price}{Optional limit price.}

\item{time_in_force}{Optional time-in-force value. Required for \code{LIMIT}.}

\item{position_side}{One of \code{"BOTH"}, \code{"LONG"}, or \code{"SHORT"}.}

\item{reduce_only}{Whether the order is reduce-only.}

\item{stop_price}{Optional trigger price for conditional orders.}

\item{working_type}{Optional trigger source. Must only be supplied with
\code{stop_price}.}

\item{new_client_order_id}{Optional client order identifier.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}

\item{timeInForce}{Legacy alias for \code{time_in_force}.}

\item{positionSide}{Legacy alias for \code{position_side}.}

\item{reduceOnly}{Legacy alias for \code{reduce_only}.}

\item{stopPrice}{Legacy alias for \code{stop_price}.}

\item{workingType}{Legacy alias for \code{working_type}.}

\item{newClientOrderId}{Legacy alias for \code{new_client_order_id}.}
}
\value{
A parsed list.
}
\description{
Place a Binance Futures order
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-market.R
\name{spot_get_rolling_window_ticker}
\alias{spot_get_rolling_window_ticker}
\title{Get Binance Spot rolling window ticker statistics}
\usage{
spot_get_rolling_window_ticker(
  symbol = NULL,
  symbols = NULL,
  windowSize = NULL,
  type = c("FULL", "MINI"),
  symbol_status = NULL,
  json_list = FALSE,
  config = config_spot()
)
}
\arguments{
\item{symbol}{Optional trading pair symbol.}

\item{symbols}{Optional character vector of trading pair symbols.}

\item{windowSize}{Optional rolling window size such as \code{"1m"}, \code{"4h"}, or \code{"7d"}.}

\item{type}{One of \code{"FULL"} or \code{"MINI"}.}

\item{symbol_status}{Optional trading status filter.}

\item{json_list}{If \code{TRUE}, return the parsed list instead of converting
array responses to \code{data.table}.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list for single-symbol requests, or a \code{data.table} for
multi-symbol requests unless \code{json_list = TRUE}.
}
\description{
Get Binance Spot rolling window ticker statistics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-options-market.R
\name{options_get_24hr_ticker}
\alias{options_get_24hr_ticker}
\title{Get Binance Options 24hr ticker statistics}
\usage{
options_get_24hr_ticker(
  symbol = NULL,
  json_list = FALSE,
  config = config_options()
)
}
\arguments{
\item{symbol}{Optional option symbol.}

\item{json_list}{If \code{TRUE}, return the parsed list instead of a \code{data.table}.}

\item{config}{An options configuration created by \code{\link[=config_options]{config_options()}}.}
}
\value{
A \code{data.table} by default, or a parsed list when \code{json_list = TRUE}.
}
\description{
Get Binance Options 24hr ticker statistics
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/helpers-exchange-info.R
\name{round_price_qty}
\alias{round_price_qty}
\alias{util_round_price_qty}
\title{Round price and quantity to exchange increments}
\usage{
round_price_qty(
  exchangeInfo,
  symbol,
  price = NULL,
  quantity = NULL,
  round_dir = c("down", "up")
)

util_round_price_qty(
  exchangeInfo,
  symbol,
  price = NULL,
  quantity = NULL,
  round_dir = c("down", "up")
)
}
\arguments{
\item{exchangeInfo}{Exchange info from \code{\link[=futures_get_exchange_info]{futures_get_exchange_info()}}.}

\item{symbol}{Trading pair symbol, for example \code{"ETHUSDT"}.}

\item{price}{Optional price to round using the symbol tick size.}

\item{quantity}{Optional quantity to round using the symbol step size.}

\item{round_dir}{Direction: \code{"down"} or \code{"up"}.}
}
\value{
A list with elements \code{price} and \code{quantity}.
}
\description{
Round price and quantity to exchange increments
}
\examples{
exchange_info <- list(
  symbols = list(
    ETHUSDT = list(
      filters = list(
        list(filterType = "PRICE_FILTER", tickSize = "0.01"),
        list(filterType = "LOT_SIZE", stepSize = "0.001")
      )
    )
  )
)
round_price_qty(exchange_info, "ETHUSDT", price = 1800.123, quantity = 0.12345)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-trade.R
\name{futures_set_multi_assets_mode}
\alias{futures_set_multi_assets_mode}
\title{Set Binance Futures multi-assets mode}
\usage{
futures_set_multi_assets_mode(
  multi_assets_margin = TRUE,
  config = config_futures()
)
}
\arguments{
\item{multi_assets_margin}{\code{TRUE} to enable multi-assets mode, \code{FALSE} to disable it.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}
}
\value{
A parsed list.
}
\description{
Set Binance Futures multi-assets mode
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-options-trade.R
\name{options_get_order}
\alias{options_get_order}
\title{Get a Binance Options order}
\usage{
options_get_order(
  symbol,
  order_id = NULL,
  client_order_id = NULL,
  config = config_options()
)
}
\arguments{
\item{symbol}{Option symbol, for example \code{"BTC-200730-9000-C"}.}

\item{order_id}{Optional exchange order ID.}

\item{client_order_id}{Optional client order ID.}

\item{config}{An options configuration created by \code{\link[=config_options]{config_options()}}.}
}
\value{
A parsed list.
}
\description{
Get a Binance Options order
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-market.R
\name{spot_get_klines}
\alias{spot_get_klines}
\title{Get Binance Spot klines}
\usage{
spot_get_klines(
  symbol,
  interval,
  startTime = NULL,
  endTime = NULL,
  timeZone = NULL,
  limit = 500,
  json_list = FALSE,
  config = config_spot()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"BTCUSDT"}.}

\item{interval}{Kline interval string.}

\item{startTime}{Optional start time in milliseconds since Unix epoch.}

\item{endTime}{Optional end time in milliseconds since Unix epoch.}

\item{timeZone}{Optional timezone string accepted by Binance.}

\item{limit}{Maximum number of rows to return.}

\item{json_list}{If \code{TRUE}, return the parsed list instead of a \code{data.table}.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A \code{data.table} by default, or a parsed list when \code{json_list = TRUE}.
}
\description{
Get Binance Spot klines
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-options-trade.R
\name{options_cancel_all_orders}
\alias{options_cancel_all_orders}
\title{Cancel all Binance Options orders for a symbol}
\usage{
options_cancel_all_orders(symbol, config = config_options())
}
\arguments{
\item{symbol}{Option symbol, for example \code{"BTC-200730-9000-C"}.}

\item{config}{An options configuration created by \code{\link[=config_options]{config_options()}}.}
}
\value{
A parsed list.
}
\description{
Cancel all Binance Options orders for a symbol
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-trade.R
\name{futures_test_order}
\alias{futures_test_order}
\title{Test a Binance Futures order}
\usage{
futures_test_order(
  symbol,
  side = c("BUY", "SELL"),
  type = c("LIMIT", "MARKET", "STOP", "STOP_MARKET", "TAKE_PROFIT", "TAKE_PROFIT_MARKET",
    "TRAILING_STOP_MARKET"),
  quantity,
  price = NULL,
  time_in_force = NULL,
  position_side = c("BOTH", "LONG", "SHORT"),
  reduce_only = FALSE,
  stop_price = NULL,
  working_type = NULL,
  new_client_order_id = NULL,
  config = config_futures()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"ETHUSDT"}.}

\item{side}{One of \code{"BUY"} or \code{"SELL"}.}

\item{type}{Order type.}

\item{quantity}{Order quantity.}

\item{price}{Optional limit price.}

\item{time_in_force}{Optional time-in-force value. Required for \code{LIMIT}.}

\item{position_side}{One of \code{"BOTH"}, \code{"LONG"}, or \code{"SHORT"}.}

\item{reduce_only}{Whether the order is reduce-only.}

\item{stop_price}{Optional trigger price for conditional orders.}

\item{working_type}{Optional trigger source. Must only be supplied with
\code{stop_price}.}

\item{new_client_order_id}{Optional client order identifier.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}
}
\value{
A parsed list.
}
\description{
Test a Binance Futures order
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-account.R
\name{spot_get_account}
\alias{spot_get_account}
\title{Get Binance Spot account information}
\usage{
spot_get_account(omit_zero_balances = FALSE, config = config_spot())
}
\arguments{
\item{omit_zero_balances}{Whether to omit zero balances from the response.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list.
}
\description{
Get Binance Spot account information
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-trade.R
\name{futures_set_margin_type}
\alias{futures_set_margin_type}
\alias{set_fapi_account_margin_type}
\title{Set Binance Futures margin type}
\usage{
futures_set_margin_type(
  symbol,
  margin_type = c("CROSSED", "ISOLATED"),
  config = config_futures()
)

set_fapi_account_margin_type(
  symbol,
  marginType = c("CROSSED", "ISOLATED"),
  config = binxr_config_futures()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"ETHUSDT"}.}

\item{margin_type}{One of \code{"CROSSED"} or \code{"ISOLATED"}.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}

\item{marginType}{Legacy alias for \code{margin_type}.}
}
\value{
A parsed list.
}
\description{
Set Binance Futures margin type
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-trade.R
\name{spot_cancel_all_orders}
\alias{spot_cancel_all_orders}
\title{Cancel all open Binance Spot orders for a symbol}
\usage{
spot_cancel_all_orders(symbol, config = config_spot())
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"BTCUSDT"}.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list by default.
}
\description{
Cancel all open Binance Spot orders for a symbol
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-general.R
\name{spot_get_execution_rules}
\alias{spot_get_execution_rules}
\title{Get Binance Spot execution rules}
\usage{
spot_get_execution_rules(
  symbol = NULL,
  symbols = NULL,
  symbol_status = NULL,
  config = config_spot()
)
}
\arguments{
\item{symbol}{Optional trading pair symbol.}

\item{symbols}{Optional character vector of trading pair symbols.}

\item{symbol_status}{Optional trading status filter.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list.
}
\description{
Get Binance Spot execution rules
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-account.R
\name{futures_get_account}
\alias{futures_get_account}
\alias{get_fapi_account}
\title{Get Binance Futures account info}
\usage{
futures_get_account(config = config_futures())

get_fapi_account(config = binxr_config_futures())
}
\arguments{
\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}
}
\value{
A parsed list.
}
\description{
Get Binance Futures account info
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-trade.R
\name{futures_cancel_all_orders}
\alias{futures_cancel_all_orders}
\alias{cancel_fapi_trade_orders_all}
\title{Cancel all open Binance Futures orders for a symbol}
\usage{
futures_cancel_all_orders(symbol, config = config_futures())

cancel_fapi_trade_orders_all(symbol, config = binxr_config_futures())
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"ETHUSDT"}.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}
}
\value{
A parsed list.
}
\description{
Cancel all open Binance Futures orders for a symbol
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-market.R
\name{futures_ping}
\alias{futures_ping}
\title{Test Binance Futures connectivity}
\usage{
futures_ping(config = config_futures())
}
\arguments{
\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}
}
\value{
A parsed list.
}
\description{
Test Binance Futures connectivity
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-market.R
\name{futures_get_open_interest}
\alias{futures_get_open_interest}
\title{Get Binance Futures open interest}
\usage{
futures_get_open_interest(symbol, config = config_futures())
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"ETHUSDT"}.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}
}
\value{
A parsed list.
}
\description{
Get Binance Futures open interest
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-trade.R
\name{spot_cancel_replace_order}
\alias{spot_cancel_replace_order}
\title{Cancel and replace a Binance Spot order}
\usage{
spot_cancel_replace_order(
  symbol,
  side = c("BUY", "SELL"),
  type = c("LIMIT", "MARKET", "STOP_LOSS", "STOP_LOSS_LIMIT", "TAKE_PROFIT",
    "TAKE_PROFIT_LIMIT", "LIMIT_MAKER"),
  cancel_replace_mode = c("STOP_ON_FAILURE", "ALLOW_FAILURE"),
  time_in_force = NULL,
  quantity = NULL,
  quote_order_qty = NULL,
  price = NULL,
  cancel_new_client_order_id = NULL,
  cancel_orig_client_order_id = NULL,
  cancel_order_id = NULL,
  new_client_order_id = NULL,
  strategy_id = NULL,
  strategy_type = NULL,
  stop_price = NULL,
  trailing_delta = NULL,
  iceberg_qty = NULL,
  new_order_resp_type = NULL,
  self_trade_prevention_mode = NULL,
  cancel_restrictions = NULL,
  order_rate_limit_exceeded_mode = NULL,
  peg_price_type = NULL,
  peg_offset_value = NULL,
  peg_offset_type = NULL,
  config = config_spot()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"BTCUSDT"}.}

\item{side}{One of \code{"BUY"} or \code{"SELL"}.}

\item{type}{Order type.}

\item{cancel_replace_mode}{One of \code{"STOP_ON_FAILURE"} or \code{"ALLOW_FAILURE"}.}

\item{time_in_force}{Optional time-in-force value.}

\item{quantity}{Optional order quantity.}

\item{quote_order_qty}{Optional quote-order quantity for \code{MARKET} orders.}

\item{price}{Optional limit price.}

\item{cancel_new_client_order_id}{Optional client order ID for the cancel request.}

\item{cancel_orig_client_order_id}{Optional original client order ID to cancel.}

\item{cancel_order_id}{Optional exchange order ID to cancel.}

\item{new_client_order_id}{Optional client order identifier.}

\item{strategy_id}{Optional numeric strategy identifier.}

\item{strategy_type}{Optional numeric strategy type. Must be at least \code{1000000}.}

\item{stop_price}{Optional trigger price.}

\item{trailing_delta}{Optional trailing delta value.}

\item{iceberg_qty}{Optional iceberg quantity.}

\item{new_order_resp_type}{Optional response type. One of \code{"ACK"}, \code{"RESULT"}, or \code{"FULL"}.}

\item{self_trade_prevention_mode}{Optional self-trade prevention mode.}

\item{cancel_restrictions}{Optional cancel restriction. One of \code{"ONLY_NEW"} or \code{"ONLY_PARTIALLY_FILLED"}.}

\item{order_rate_limit_exceeded_mode}{Optional order-rate-limit behavior. One of \code{"DO_NOTHING"} or \code{"CANCEL_ONLY"}.}

\item{peg_price_type}{Optional pegged price type. One of \code{"PRIMARY_PEG"} or \code{"MARKET_PEG"}.}

\item{peg_offset_value}{Optional peg offset value.}

\item{peg_offset_type}{Optional peg offset type. Only \code{"PRICE_LEVEL"} is supported.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list.
}
\description{
Cancel and replace a Binance Spot order
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-options-trade.R
\name{options_place_order}
\alias{options_place_order}
\title{Place a Binance Options order}
\usage{
options_place_order(
  symbol,
  side = c("BUY", "SELL"),
  type = c("LIMIT"),
  quantity,
  price,
  time_in_force = c("GTC", "IOC", "FOK"),
  reduce_only = FALSE,
  post_only = FALSE,
  new_order_resp_type = c("ACK", "RESULT"),
  client_order_id = NULL,
  is_mmp = FALSE,
  self_trade_prevention_mode = NULL,
  config = config_options()
)
}
\arguments{
\item{symbol}{Option symbol, for example \code{"BTC-200730-9000-C"}.}

\item{side}{One of \code{"BUY"} or \code{"SELL"}.}

\item{type}{Order type. Only \code{"LIMIT"} is currently supported.}

\item{quantity}{Order quantity.}

\item{price}{Order price.}

\item{time_in_force}{Time in force. Default is \code{"GTC"}.}

\item{reduce_only}{Whether the order is reduce-only.}

\item{post_only}{Whether the order is post-only.}

\item{new_order_resp_type}{Response type. One of \code{"ACK"} or \code{"RESULT"}.}

\item{client_order_id}{Optional client order ID.}

\item{is_mmp}{Whether the order is an MMP order.}

\item{self_trade_prevention_mode}{Optional STP mode.}

\item{config}{An options configuration created by \code{\link[=config_options]{config_options()}}.}
}
\value{
A parsed list.
}
\description{
Place a Binance Options order
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/binxr-package.R
\docType{package}
\name{binxr-package}
\alias{binxr}
\alias{binxr-package}
\title{binxr: Binance REST API Client for R}
\description{
Client helpers for the Binance Spot, Futures, and Options REST APIs.
The package provides configuration constructors, signed and unsigned
request helpers, market data retrieval, account access, and order management
functions. Tabular endpoint responses are returned as \code{data.table} objects
by default.
}
\details{
Network examples and live trading calls are not run during package checks.
Authenticated endpoints require user-supplied Binance API credentials.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/OliverLDS/binxr}
  \item \url{https://oliverlds.github.io/binxr/}
  \item Report bugs at \url{https://github.com/OliverLDS/binxr/issues}
}

}
\author{
\strong{Maintainer}: Oliver Zhou \email{oliver.yxzhou@gmail.com}

Authors:
\itemize{
  \item Lily Li \email{lilyli.qc.hk@gmail.com}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-market.R
\name{futures_get_book_ticker}
\alias{futures_get_book_ticker}
\title{Get Binance Futures order book ticker}
\usage{
futures_get_book_ticker(symbol = NULL, config = config_futures())
}
\arguments{
\item{symbol}{Optional trading pair symbol.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}
}
\value{
A parsed list.
}
\description{
Get Binance Futures order book ticker
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-market.R
\name{futures_get_continuous_klines}
\alias{futures_get_continuous_klines}
\title{Get Binance Futures continuous contract klines}
\usage{
futures_get_continuous_klines(
  pair,
  contract_type = c("PERPETUAL", "CURRENT_MONTH", "NEXT_MONTH", "CURRENT_QUARTER",
    "NEXT_QUARTER"),
  interval,
  startTime = NULL,
  endTime = NULL,
  limit = 500,
  json_list = FALSE,
  config = config_futures()
)
}
\arguments{
\item{pair}{Futures pair, for example \code{"BTCUSDT"}.}

\item{contract_type}{One of \code{"PERPETUAL"}, \code{"CURRENT_MONTH"},
\code{"NEXT_MONTH"}, \code{"CURRENT_QUARTER"}, or \code{"NEXT_QUARTER"}.}

\item{interval}{Kline interval string.}

\item{startTime}{Optional start time in milliseconds since Unix epoch.}

\item{endTime}{Optional end time in milliseconds since Unix epoch.}

\item{limit}{Maximum number of rows to return. Must not exceed \code{1500}.}

\item{json_list}{If \code{TRUE}, return the parsed list instead of a \code{data.table}.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}
}
\value{
A \code{data.table} by default, or a parsed list when \code{json_list = TRUE}.
}
\description{
Get Binance Futures continuous contract klines
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-trade.R
\name{futures_get_open_orders}
\alias{futures_get_open_orders}
\alias{get_fapi_trade_open_orders}
\title{Get Binance Futures open orders}
\usage{
futures_get_open_orders(
  symbol = NULL,
  json_list = FALSE,
  config = config_futures()
)

get_fapi_trade_open_orders(symbol = NULL, config = binxr_config_futures())
}
\arguments{
\item{symbol}{Optional trading pair symbol. If \code{NULL}, returns all open
orders.}

\item{json_list}{If \code{TRUE}, return the parsed list instead of a \code{data.table}.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}
}
\value{
A \code{data.table} by default, or a parsed list when \code{json_list = TRUE}.
}
\description{
Get Binance Futures open orders
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-trade.R
\name{spot_place_order_list_otoco}
\alias{spot_place_order_list_otoco}
\title{Place a Binance Spot OTOCO order list}
\usage{
spot_place_order_list_otoco(
  symbol,
  list_client_order_id = NULL,
  new_order_resp_type = NULL,
  self_trade_prevention_mode = NULL,
  working_type = c("LIMIT", "LIMIT_MAKER"),
  working_side = c("BUY", "SELL"),
  working_client_order_id = NULL,
  working_price,
  working_quantity,
  working_iceberg_qty = NULL,
  working_time_in_force = NULL,
  working_strategy_id = NULL,
  working_strategy_type = NULL,
  working_peg_price_type = NULL,
  working_peg_offset_type = NULL,
  working_peg_offset_value = NULL,
  pending_side = c("BUY", "SELL"),
  pending_quantity,
  pending_above_type = c("STOP_LOSS_LIMIT", "STOP_LOSS", "LIMIT_MAKER", "TAKE_PROFIT",
    "TAKE_PROFIT_LIMIT"),
  pending_above_client_order_id = NULL,
  pending_above_price = NULL,
  pending_above_stop_price = NULL,
  pending_above_trailing_delta = NULL,
  pending_above_iceberg_qty = NULL,
  pending_above_time_in_force = NULL,
  pending_above_strategy_id = NULL,
  pending_above_strategy_type = NULL,
  pending_above_peg_price_type = NULL,
  pending_above_peg_offset_type = NULL,
  pending_above_peg_offset_value = NULL,
  pending_below_type = c("STOP_LOSS", "STOP_LOSS_LIMIT", "TAKE_PROFIT",
    "TAKE_PROFIT_LIMIT"),
  pending_below_client_order_id = NULL,
  pending_below_price = NULL,
  pending_below_stop_price = NULL,
  pending_below_trailing_delta = NULL,
  pending_below_iceberg_qty = NULL,
  pending_below_time_in_force = NULL,
  pending_below_strategy_id = NULL,
  pending_below_strategy_type = NULL,
  pending_below_peg_price_type = NULL,
  pending_below_peg_offset_type = NULL,
  pending_below_peg_offset_value = NULL,
  config = config_spot()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"BTCUSDT"}.}

\item{list_client_order_id}{Optional order-list client ID.}

\item{new_order_resp_type}{Optional response type. One of \code{"ACK"}, \code{"RESULT"}, or \code{"FULL"}.}

\item{self_trade_prevention_mode}{Optional self-trade prevention mode.}

\item{working_type}{Working leg type. One of \code{"LIMIT"} or \code{"LIMIT_MAKER"}.}

\item{working_side}{Working leg side. One of \code{"BUY"} or \code{"SELL"}.}

\item{working_client_order_id}{Optional working-leg client ID.}

\item{working_price}{Working leg price.}

\item{working_quantity}{Working leg quantity.}

\item{working_iceberg_qty}{Optional working-leg iceberg quantity.}

\item{working_time_in_force}{Optional working-leg time-in-force.}

\item{working_strategy_id}{Optional working-leg strategy ID.}

\item{working_strategy_type}{Optional working-leg strategy type.}

\item{working_peg_price_type}{Optional working-leg peg price type.}

\item{working_peg_offset_type}{Optional working-leg peg offset type.}

\item{working_peg_offset_value}{Optional working-leg peg offset value.}

\item{pending_side}{Shared side for the pending OCO legs.}

\item{pending_quantity}{Pending leg quantity.}

\item{pending_above_type}{Type for the pending-above leg.}

\item{pending_above_client_order_id}{Optional pending-above client ID.}

\item{pending_above_price}{Optional pending-above price.}

\item{pending_above_stop_price}{Optional pending-above stop price.}

\item{pending_above_trailing_delta}{Optional pending-above trailing delta.}

\item{pending_above_iceberg_qty}{Optional pending-above iceberg quantity.}

\item{pending_above_time_in_force}{Optional pending-above time-in-force.}

\item{pending_above_strategy_id}{Optional pending-above strategy ID.}

\item{pending_above_strategy_type}{Optional pending-above strategy type.}

\item{pending_above_peg_price_type}{Optional pending-above peg price type.}

\item{pending_above_peg_offset_type}{Optional pending-above peg offset type.}

\item{pending_above_peg_offset_value}{Optional pending-above peg offset value.}

\item{pending_below_type}{Type for the pending-below leg.}

\item{pending_below_client_order_id}{Optional pending-below client ID.}

\item{pending_below_price}{Optional pending-below price.}

\item{pending_below_stop_price}{Optional pending-below stop price.}

\item{pending_below_trailing_delta}{Optional pending-below trailing delta.}

\item{pending_below_iceberg_qty}{Optional pending-below iceberg quantity.}

\item{pending_below_time_in_force}{Optional pending-below time-in-force.}

\item{pending_below_strategy_id}{Optional pending-below strategy ID.}

\item{pending_below_strategy_type}{Optional pending-below strategy type.}

\item{pending_below_peg_price_type}{Optional pending-below peg price type.}

\item{pending_below_peg_offset_type}{Optional pending-below peg offset type.}

\item{pending_below_peg_offset_value}{Optional pending-below peg offset value.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list.
}
\description{
Place a Binance Spot OTOCO order list
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-account.R
\name{spot_get_order_list}
\alias{spot_get_order_list}
\title{Get a Binance Spot order list}
\usage{
spot_get_order_list(
  order_list_id = NULL,
  orig_client_order_id = NULL,
  config = config_spot()
)
}
\arguments{
\item{order_list_id}{Optional exchange order-list ID.}

\item{orig_client_order_id}{Optional order-list client ID.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list.
}
\description{
Get a Binance Spot order list
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-options-trade.R
\name{options_cancel_all_orders_by_underlying}
\alias{options_cancel_all_orders_by_underlying}
\title{Cancel all Binance Options orders by underlying}
\usage{
options_cancel_all_orders_by_underlying(underlying, config = config_options())
}
\arguments{
\item{underlying}{Underlying symbol, for example \code{"BTCUSDT"}.}

\item{config}{An options configuration created by \code{\link[=config_options]{config_options()}}.}
}
\value{
A parsed list.
}
\description{
Cancel all Binance Options orders by underlying
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-trade.R
\name{spot_place_order_list_oco}
\alias{spot_place_order_list_oco}
\title{Place a Binance Spot OCO order list}
\usage{
spot_place_order_list_oco(
  symbol,
  side = c("BUY", "SELL"),
  quantity,
  above_type = c("STOP_LOSS_LIMIT", "STOP_LOSS", "LIMIT_MAKER", "TAKE_PROFIT",
    "TAKE_PROFIT_LIMIT"),
  above_client_order_id = NULL,
  above_iceberg_qty = NULL,
  above_price = NULL,
  above_stop_price = NULL,
  above_trailing_delta = NULL,
  above_time_in_force = NULL,
  above_strategy_id = NULL,
  above_strategy_type = NULL,
  above_peg_price_type = NULL,
  above_peg_offset_type = NULL,
  above_peg_offset_value = NULL,
  below_type = c("STOP_LOSS", "STOP_LOSS_LIMIT", "TAKE_PROFIT", "TAKE_PROFIT_LIMIT"),
  below_client_order_id = NULL,
  below_iceberg_qty = NULL,
  below_price = NULL,
  below_stop_price = NULL,
  below_trailing_delta = NULL,
  below_time_in_force = NULL,
  below_strategy_id = NULL,
  below_strategy_type = NULL,
  below_peg_price_type = NULL,
  below_peg_offset_type = NULL,
  below_peg_offset_value = NULL,
  list_client_order_id = NULL,
  new_order_resp_type = NULL,
  self_trade_prevention_mode = NULL,
  config = config_spot()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"BTCUSDT"}.}

\item{side}{One of \code{"BUY"} or \code{"SELL"}.}

\item{quantity}{Order quantity shared by both legs.}

\item{above_type}{Type for the above leg.}

\item{above_client_order_id}{Optional above-leg client ID.}

\item{above_iceberg_qty}{Optional above-leg iceberg quantity.}

\item{above_price}{Optional above-leg limit price.}

\item{above_stop_price}{Optional above-leg stop price.}

\item{above_trailing_delta}{Optional above-leg trailing delta.}

\item{above_time_in_force}{Optional above-leg time-in-force.}

\item{above_strategy_id}{Optional above-leg strategy ID.}

\item{above_strategy_type}{Optional above-leg strategy type.}

\item{above_peg_price_type}{Optional above-leg peg price type.}

\item{above_peg_offset_type}{Optional above-leg peg offset type.}

\item{above_peg_offset_value}{Optional above-leg peg offset value.}

\item{below_type}{Type for the below leg.}

\item{below_client_order_id}{Optional below-leg client ID.}

\item{below_iceberg_qty}{Optional below-leg iceberg quantity.}

\item{below_price}{Optional below-leg limit price.}

\item{below_stop_price}{Optional below-leg stop price.}

\item{below_trailing_delta}{Optional below-leg trailing delta.}

\item{below_time_in_force}{Optional below-leg time-in-force.}

\item{below_strategy_id}{Optional below-leg strategy ID.}

\item{below_strategy_type}{Optional below-leg strategy type.}

\item{below_peg_price_type}{Optional below-leg peg price type.}

\item{below_peg_offset_type}{Optional below-leg peg offset type.}

\item{below_peg_offset_value}{Optional below-leg peg offset value.}

\item{list_client_order_id}{Optional order-list client ID.}

\item{new_order_resp_type}{Optional response type. One of \code{"ACK"}, \code{"RESULT"}, or \code{"FULL"}.}

\item{self_trade_prevention_mode}{Optional self-trade prevention mode.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list.
}
\description{
Place a Binance Spot OCO order list
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{config_options}
\alias{config_options}
\title{Create a Binance options configuration}
\usage{
config_options(
  api_key = Sys.getenv("BINX_API_KEY", unset = ""),
  secret_key = Sys.getenv("BINX_SECRET_KEY", unset = ""),
  base_url = "https://eapi.binance.com",
  recv_window = 10000L,
  timeout = 30,
  verbose = FALSE
)
}
\arguments{
\item{api_key}{character API key. Leave \code{NULL} for unsigned/public requests.}

\item{secret_key}{character Secret key. Leave \code{NULL} for unsigned/public requests.}

\item{base_url}{character Base URL for the Binance Options API.}

\item{recv_window}{integer recvWindow in milliseconds used for signed requests.}

\item{timeout}{numeric request timeout in seconds.}

\item{verbose}{logical whether to enable verbose request debugging in future helpers.}
}
\value{
A \code{binxr_config} list.
}
\description{
Create a Binance options configuration
}
\examples{
cfg <- config_options(api_key = NULL, secret_key = NULL)
cfg$product
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-market.R
\name{futures_get_order_book}
\alias{futures_get_order_book}
\title{Get Binance Futures order book}
\usage{
futures_get_order_book(symbol, limit = NULL, config = config_futures())
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"ETHUSDT"}.}

\item{limit}{Optional depth limit. One of \code{5}, \code{10}, \code{20}, \code{50}, \code{100},
\code{500}, or \code{1000}.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}
}
\value{
A parsed list.
}
\description{
Get Binance Futures order book
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-trade.R
\name{spot_place_order_list_opo}
\alias{spot_place_order_list_opo}
\title{Place a Binance Spot OPO order list}
\usage{
spot_place_order_list_opo(
  symbol,
  list_client_order_id = NULL,
  new_order_resp_type = NULL,
  self_trade_prevention_mode = NULL,
  working_type = c("LIMIT", "LIMIT_MAKER"),
  working_side = c("BUY", "SELL"),
  working_client_order_id = NULL,
  working_price,
  working_quantity,
  working_iceberg_qty = NULL,
  working_time_in_force = NULL,
  working_strategy_id = NULL,
  working_strategy_type = NULL,
  working_peg_price_type = NULL,
  working_peg_offset_type = NULL,
  working_peg_offset_value = NULL,
  pending_type = c("LIMIT", "MARKET", "STOP_LOSS", "STOP_LOSS_LIMIT", "TAKE_PROFIT",
    "TAKE_PROFIT_LIMIT", "LIMIT_MAKER"),
  pending_side = c("BUY", "SELL"),
  pending_client_order_id = NULL,
  pending_price = NULL,
  pending_stop_price = NULL,
  pending_trailing_delta = NULL,
  pending_iceberg_qty = NULL,
  pending_time_in_force = NULL,
  pending_strategy_id = NULL,
  pending_strategy_type = NULL,
  pending_peg_price_type = NULL,
  pending_peg_offset_type = NULL,
  pending_peg_offset_value = NULL,
  config = config_spot()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"BTCUSDT"}.}

\item{list_client_order_id}{Optional order-list client ID.}

\item{new_order_resp_type}{Optional response type. One of \code{"ACK"}, \code{"RESULT"}, or \code{"FULL"}.}

\item{self_trade_prevention_mode}{Optional self-trade prevention mode.}

\item{working_type}{Working leg type. One of \code{"LIMIT"} or \code{"LIMIT_MAKER"}.}

\item{working_side}{Working leg side. One of \code{"BUY"} or \code{"SELL"}.}

\item{working_client_order_id}{Optional working-leg client ID.}

\item{working_price}{Working leg price.}

\item{working_quantity}{Working leg quantity.}

\item{working_iceberg_qty}{Optional working-leg iceberg quantity.}

\item{working_time_in_force}{Optional working-leg time-in-force.}

\item{working_strategy_id}{Optional working-leg strategy ID.}

\item{working_strategy_type}{Optional working-leg strategy type.}

\item{working_peg_price_type}{Optional working-leg peg price type.}

\item{working_peg_offset_type}{Optional working-leg peg offset type.}

\item{working_peg_offset_value}{Optional working-leg peg offset value.}

\item{pending_type}{Pending leg order type.}

\item{pending_side}{Pending leg side. One of \code{"BUY"} or \code{"SELL"}.}

\item{pending_client_order_id}{Optional pending-leg client ID.}

\item{pending_price}{Optional pending-leg price.}

\item{pending_stop_price}{Optional pending-leg stop price.}

\item{pending_trailing_delta}{Optional pending-leg trailing delta.}

\item{pending_iceberg_qty}{Optional pending-leg iceberg quantity.}

\item{pending_time_in_force}{Optional pending-leg time-in-force.}

\item{pending_strategy_id}{Optional pending-leg strategy ID.}

\item{pending_strategy_type}{Optional pending-leg strategy type.}

\item{pending_peg_price_type}{Optional pending-leg peg price type.}

\item{pending_peg_offset_type}{Optional pending-leg peg offset type.}

\item{pending_peg_offset_value}{Optional pending-leg peg offset value.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list.
}
\description{
Place a Binance Spot OPO order list
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parsers.R
\name{coerce_account_frames}
\alias{coerce_account_frames}
\title{(Optional) Coerce account to data.frames}
\usage{
coerce_account_frames(account)
}
\arguments{
\item{account}{get_account() result}
}
\value{
list(assets=..., positions=...)
}
\description{
(Optional) Coerce account to data.frames
}
\examples{
account <- list(
  assets = list(list(asset = "USDT", walletBalance = "10")),
  positions = list(list(symbol = "ETHUSDT", positionAmt = "0"))
)
coerce_account_frames(account)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-trade.R
\name{spot_cancel_order}
\alias{spot_cancel_order}
\title{Cancel a Binance Spot order}
\usage{
spot_cancel_order(
  symbol,
  order_id = NULL,
  orig_client_order_id = NULL,
  new_client_order_id = NULL,
  cancel_restrictions = NULL,
  config = config_spot()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"BTCUSDT"}.}

\item{order_id}{Optional exchange order ID.}

\item{orig_client_order_id}{Optional client order ID.}

\item{new_client_order_id}{Optional client order ID for the cancel request.}

\item{cancel_restrictions}{Optional cancel restriction. One of \code{"ONLY_NEW"} or \code{"ONLY_PARTIALLY_FILLED"}.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list.
}
\description{
Cancel a Binance Spot order
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-general.R
\name{spot_get_server_time}
\alias{spot_get_server_time}
\title{Get Binance Spot server time}
\usage{
spot_get_server_time(config = config_spot())
}
\arguments{
\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A \code{POSIXct} timestamp.
}
\description{
Get Binance Spot server time
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-account.R
\name{futures_get_position_risk}
\alias{futures_get_position_risk}
\alias{get_fapi_account_position_risk}
\title{Get Binance Futures position risk}
\usage{
futures_get_position_risk(json_list = FALSE, config = config_futures())

get_fapi_account_position_risk(config = binxr_config_futures())
}
\arguments{
\item{json_list}{If \code{TRUE}, return the parsed list instead of a \code{data.table}.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}
}
\value{
A \code{data.table} by default, or a parsed list when \code{json_list = TRUE}.
}
\description{
Get Binance Futures position risk
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-market.R
\name{futures_get_ticker_price}
\alias{futures_get_ticker_price}
\title{Get Binance Futures ticker price}
\usage{
futures_get_ticker_price(
  symbol = NULL,
  version = c("v2", "deprecated"),
  config = config_futures()
)
}
\arguments{
\item{symbol}{Optional trading pair symbol.}

\item{version}{One of \code{"v2"} or \code{"deprecated"}.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}
}
\value{
A parsed list.
}
\description{
Get Binance Futures ticker price
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-market.R
\name{spot_get_historical_trades}
\alias{spot_get_historical_trades}
\title{Get Binance Spot historical trades}
\usage{
spot_get_historical_trades(
  symbol,
  limit = 500,
  fromId = NULL,
  json_list = FALSE,
  config = config_spot()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"BTCUSDT"}.}

\item{limit}{Maximum number of rows to return.}

\item{fromId}{Optional trade identifier to fetch from.}

\item{json_list}{If \code{TRUE}, return the parsed list instead of a \code{data.table}.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A \code{data.table} by default, or a parsed list when \code{json_list = TRUE}.
}
\description{
Get Binance Spot historical trades
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-general.R
\name{spot_get_exchange_info}
\alias{spot_get_exchange_info}
\title{Get Binance Spot exchange information}
\usage{
spot_get_exchange_info(
  symbol = NULL,
  symbols = NULL,
  permissions = NULL,
  show_permission_sets = TRUE,
  symbol_status = NULL,
  config = config_spot()
)
}
\arguments{
\item{symbol}{Optional trading pair symbol.}

\item{symbols}{Optional character vector of trading pair symbols.}

\item{permissions}{Optional character vector of permission strings.}

\item{show_permission_sets}{Logical flag forwarded as \code{showPermissionSets}.}

\item{symbol_status}{Optional trading status filter.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list.
}
\description{
Get Binance Spot exchange information
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{binxr_config_options}
\alias{binxr_config_options}
\title{Backward-compatible options config constructor}
\usage{
binxr_config_options(
  api_key = Sys.getenv("BINX_API_KEY", unset = ""),
  secret_key = Sys.getenv("BINX_SECRET_KEY", unset = ""),
  base = "https://eapi.binance.com",
  recvWindow = 10000L
)
}
\arguments{
\item{api_key}{character API key. Leave \code{NULL} for unsigned/public requests.}

\item{secret_key}{character Secret key. Leave \code{NULL} for unsigned/public requests.}

\item{base}{Legacy alias for \code{base_url}.}

\item{recvWindow}{Legacy alias for \code{recv_window}.}
}
\value{
A \code{binxr_config} list.
}
\description{
Backward-compatible options config constructor
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-spot-trade.R
\name{spot_place_oco_order}
\alias{spot_place_oco_order}
\title{Place a deprecated Binance Spot OCO order}
\usage{
spot_place_oco_order(
  symbol,
  side = c("BUY", "SELL"),
  quantity,
  price,
  stop_price,
  list_client_order_id = NULL,
  limit_client_order_id = NULL,
  limit_strategy_id = NULL,
  limit_strategy_type = NULL,
  limit_iceberg_qty = NULL,
  trailing_delta = NULL,
  stop_client_order_id = NULL,
  stop_strategy_id = NULL,
  stop_strategy_type = NULL,
  stop_limit_price = NULL,
  stop_iceberg_qty = NULL,
  stop_limit_time_in_force = NULL,
  new_order_resp_type = NULL,
  self_trade_prevention_mode = NULL,
  config = config_spot()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"BTCUSDT"}.}

\item{side}{One of \code{"BUY"} or \code{"SELL"}.}

\item{quantity}{Order quantity shared by both legs.}

\item{price}{Limit price for the maker leg.}

\item{stop_price}{Stop trigger price.}

\item{list_client_order_id}{Optional order-list client ID.}

\item{limit_client_order_id}{Optional limit-leg client ID.}

\item{limit_strategy_id}{Optional limit-leg strategy ID.}

\item{limit_strategy_type}{Optional limit-leg strategy type.}

\item{limit_iceberg_qty}{Optional limit-leg iceberg quantity.}

\item{trailing_delta}{Optional shared trailing delta.}

\item{stop_client_order_id}{Optional stop-leg client ID.}

\item{stop_strategy_id}{Optional stop-leg strategy ID.}

\item{stop_strategy_type}{Optional stop-leg strategy type.}

\item{stop_limit_price}{Optional stop-limit price.}

\item{stop_iceberg_qty}{Optional stop-leg iceberg quantity.}

\item{stop_limit_time_in_force}{Optional stop-limit time-in-force.}

\item{new_order_resp_type}{Optional response type. One of \code{"ACK"}, \code{"RESULT"}, or \code{"FULL"}.}

\item{self_trade_prevention_mode}{Optional self-trade prevention mode.}

\item{config}{A spot configuration created by \code{\link[=config_spot]{config_spot()}}.}
}
\value{
A parsed list.
}
\description{
Place a deprecated Binance Spot OCO order
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-trade.R
\name{futures_set_position_mode}
\alias{futures_set_position_mode}
\alias{set_fapi_account_position_side}
\title{Set Binance Futures position mode}
\usage{
futures_set_position_mode(dual_side_position = TRUE, config = config_futures())

set_fapi_account_position_side(
  dualSidePosition = TRUE,
  config = binxr_config_futures()
)
}
\arguments{
\item{dual_side_position}{\code{TRUE} for hedge mode, \code{FALSE} for one-way mode.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}

\item{dualSidePosition}{Legacy alias for \code{dual_side_position}.}
}
\value{
A parsed list.
}
\description{
Set Binance Futures position mode
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-futures-trade.R
\name{futures_cancel_order}
\alias{futures_cancel_order}
\alias{cancel_fapi_trade_order}
\title{Cancel a Binance Futures order}
\usage{
futures_cancel_order(
  symbol,
  order_id = NULL,
  orig_client_order_id = NULL,
  config = config_futures()
)

cancel_fapi_trade_order(
  symbol,
  orderId = NULL,
  origClientOrderId = NULL,
  config = binxr_config_futures()
)
}
\arguments{
\item{symbol}{Trading pair symbol, for example \code{"ETHUSDT"}.}

\item{order_id}{Optional exchange order ID.}

\item{orig_client_order_id}{Optional client order ID.}

\item{config}{A futures configuration created by \code{\link[=config_futures]{config_futures()}}.}

\item{orderId}{Legacy alias for \code{order_id}.}

\item{origClientOrderId}{Legacy alias for \code{orig_client_order_id}.}
}
\value{
A parsed list.
}
\description{
Cancel a Binance Futures order
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/endpoints-options-market.R
\name{options_get_order_book}
\alias{options_get_order_book}
\title{Get Binance Options order book}
\usage{
options_get_order_book(symbol, limit = NULL, config = config_options())
}
\arguments{
\item{symbol}{Option symbol, for example \code{"BTC-200730-9000-C"}.}

\item{limit}{Optional depth limit. One of \code{10}, \code{20}, \code{50}, \code{100}, \code{500}, or
\code{1000}.}

\item{config}{An options configuration created by \code{\link[=config_options]{config_options()}}.}
}
\value{
A parsed list.
}
\description{
Get Binance Options order book
}
